/*
* A simple ray tracer in a sequence of sphere movements
* Code based on the program of the 
* http://www.purplealienplanet.com/sites/default/files/raytrace_part1.c
*
* @author Ana Moretón Fernández
* @version v0.1
*
* (c) 2017, Grupo Trasgo, Universidad de Valladolid
*/


#include <stdio.h>
#include <hitmap.h>
#include <stdlib.h>
#include <stdbool.h> /* Needed for boolean datatype */
#include <math.h>

#define min(a,b) (((a) < (b)) ? (a) : (b))
#define max(a,b) (((a) > (b)) ? (a) : (b))

/* The vector structure */
typedef struct{
      float x,y,z;
}vector;

/* The sphere */
typedef struct{
        vector pos;
        float  radius;
	int material;
}sphere; 

/* The ray */
typedef struct{
        vector start;
        vector dir;
}ray;

/* Colour */
typedef struct{
	float red, green, blue;
}colour;

/* Material Definition */
typedef struct{
	colour diffuse;
	float reflection;
}material;

/* Lightsource definition */
typedef struct{
	vector pos;
	colour intensity;
}light;

/* Subtract two vectors and return the resulting vector */
vector vectorSub(vector *v1, vector *v2){
	vector result = {v1->x - v2->x, v1->y - v2->y, v1->z - v2->z };
	return result;
}

/* Multiply two vectors and return the resulting scalar (dot product) */
float vectorDot(vector *v1, vector *v2){
	return v1->x * v2->x + v1->y * v2->y + v1->z * v2->z;
}

/* Calculate Vector x Scalar and return resulting Vector*/ 
vector vectorScale(float c, vector *v){
        vector result = {v->x * c, v->y * c, v->z * c };
        return result;
}

/* Add two vectors and return the resulting vector */
vector vectorAdd(vector *v1, vector *v2){
        vector result = {v1->x + v2->x, v1->y + v2->y, v1->z + v2->z };
        return result;
}

/* Check if the ray and sphere intersect */
bool intersectRaySphere(ray *r, sphere *s, float *t){
	
	bool retval = false;

	/* A = d.d, the vector dot product of the direction */
	float A = vectorDot(&r->dir, &r->dir); 
	
	/* We need a vector representing the distance between the start of 
	 * the ray and the position of the circle.
	 * This is the term (p0 - c) 
	 */
	vector dist = vectorSub(&r->start, &s->pos);
	
	/* 2d.(p0 - c) */  
	float B = 2 * vectorDot(&r->dir, &dist);
	
	/* (p0 - c).(p0 - c) - r^2 */
	float C = vectorDot(&dist, &dist) - (s->radius * s->radius);
	
	/* Solving the discriminant */
	float discr = B * B - 4 * A * C;
	
	/* If the discriminant is negative, there are no real roots.
	 * Return false in that case as the ray misses the sphere.
	 * Return true in all other cases (can be one or two intersections)
	 * t represents the distance between the start of the ray and
	 * the point on the sphere where it intersects.
	 */
	if(discr < 0)
		retval = false;
	else{
		float sqrtdiscr = sqrtf(discr);
		float t0 = (-B + sqrtdiscr)/(2);
		float t1 = (-B - sqrtdiscr)/(2);
		
		/* We want the closest one */
		if(t0 > t1)
			t0 = t1;

		/* Verify t1 larger than 0 and less than the original t */
		if((t0 > 0.001f) && (t0 < *t)){
			*t = t0;
			retval = true;
		}else
			retval = false;
	}

return retval;
}

/* Output data as PPM file */
void saveppm(char *filename, unsigned char *img, int width, int height){
	/* FILE pointer */
	FILE *f;

	/* Open file for writing */
	f = fopen(filename, "w");

	/* PPM header info, including the size of the image */
	fprintf(f, "P6 %d %d %d\n", width, height, 255);

	/* Write the image data to the file - remember 3 byte per pixel */
	fwrite(img, 3, width*height, f);

	/* Make sure you close the file */
	fclose(f);
}

/* Perform computation of each pixel */
int compute (int x, int y, float *red_g, float *green_g, float *blue_g, sphere sphe, light lights[3], material materials[3]){

			ray r;
			float red = *red_g;
			float green = *green_g;
			float blue = *blue_g;

			int level = 0;
			float coef = 1.0;
			
			r.start.x = (float)x;
			r.start.y = (float)y;
			r.start.z = -2000;
		
			r.dir.x = 0;
			r.dir.y = 0;
			r.dir.z = 1;

			int sal=1;
			
			do{
				/* Find closest intersection */
				float t = 20000.0f;
				
				if(!intersectRaySphere(&r, &sphe, &t)){
					if( (red==0) && (blue==0) && (green==0)){
						sal=0;
					}
						break;
						//return 0; 
				}
				sal=1;
				vector scaled = vectorScale(t, &r.dir);
				vector newStart = vectorAdd(&r.start, &scaled);
				
				/* Find the normal for this new vector at the point of intersection */
				vector n = vectorSub(&newStart, &sphe.pos);
				float temp = vectorDot(&n, &n);
				if(temp == 0) {
					sal =0;
					break;
				 }
				
				temp = 1.0f / sqrtf(temp);
				n = vectorScale(temp, &n);

				/* Find the material to determine the colour */
				material currentMat = materials[sphe.material];
				
				/* Find the value of the light at this point */
				unsigned int j;
				for(j=0; j < 3; j++){
					light currentLight = lights[j];
					vector dist = vectorSub(&currentLight.pos, &newStart);
					if(vectorDot(&n, &dist) <= 0.0f) continue;
					float ti = sqrtf(vectorDot(&dist,&dist));
					if(ti <= 0.0f) continue;
					
					ray lightRay;
					lightRay.start = newStart;
					lightRay.dir = vectorScale((1/ti), &dist);
					
					/* Lambert diffusion */
					float lambert = vectorDot(&lightRay.dir, &n) * coef; 
					red += lambert * currentLight.intensity.red * currentMat.diffuse.red;
					green += lambert * currentLight.intensity.green * currentMat.diffuse.green;
					blue += lambert * currentLight.intensity.blue * currentMat.diffuse.blue;
				}
				/* Iterate over the reflection */
				coef *= currentMat.reflection;
				
				/* The reflected ray start and direction */
				r.start = newStart;
				float reflect = 2.0f * vectorDot(&r.dir, &n);
				vector tmp = vectorScale(reflect, &n);
				r.dir = vectorSub(&r.dir, &tmp);

				level++;

			}while((coef > 0.0f) && (level < 15));

		*red_g =red;
		*green_g=green;
		*blue_g=blue;
		//printf("SAL: %d\n",sal);
		return sal; 

}

/* Declare type */
hit_tileNewType( float );

/* Hit Clocks */
HitClock        mainClock;
HitClock        initClock;
HitClock        sequentialClock;



/* Main program*/
int main(int argc, char *argv[]){


        hit_comInit( &argc, &argv );
	
	if ( argc != 3 ) {
                fprintf(stderr, "\nUsage: %s <numRows> <numColumns> \n", argv[0]);
                exit(EXIT_FAILURE);
        }

	/* Width and height of out image */
 	int WIDTH = atoi( argv[1] );
	int HEIGHT = atoi( argv[2] );
	
	/* Initial position of the sphere */
	int pos_sphere[] ={
		WIDTH/8 ,WIDTH/8, 10, WIDTH/9
	};

	/* Vector movement */
	float vector_movement[] = {
		1,1
	};


	
	material materials[3];
	materials[0].diffuse.red = (float)1;
	materials[0].diffuse.green = (float)0.5;
	materials[0].diffuse.blue = (float)0.5;
	materials[0].reflection = (float)0.2;
	
	materials[1].diffuse.red = (float)0;
	materials[1].diffuse.green = (float)1;
	materials[1].diffuse.blue = (float)0;
	materials[1].reflection = (float)0.5;
	
	materials[2].diffuse.red = (float)0;
	materials[2].diffuse.green = (float)0;
	materials[2].diffuse.blue = (float)1;
	materials[2].reflection = (float)0.9;
	
	/* Sphere */
	sphere sphe; 
	sphe.pos.x = (float)pos_sphere[0];
	sphe.pos.y = (float)pos_sphere[1];
	sphe.pos.z = (float)pos_sphere[2];
	sphe.radius = (float)pos_sphere[3];
	sphe.material = 0;


	/* lights */
	light lights[3];
	
	lights[0].pos.x = 0;
	lights[0].pos.y = 240;
	lights[0].pos.z = -100;
	lights[0].intensity.red = 1;
	lights[0].intensity.green = 1;
	lights[0].intensity.blue = 1;
	
	lights[1].pos.x = 3200;
	lights[1].pos.y = 3000;
	lights[1].pos.z = -1000;
	lights[1].intensity.red = (float)0.6;
	lights[1].intensity.green = (float)0.7;
	lights[1].intensity.blue = (float)1;

	lights[2].pos.x = 600;
	lights[2].pos.y = 0;
	lights[2].pos.z = -100;
	lights[2].intensity.red = (float)0.3;
	lights[2].intensity.green = (float)0.5;
	lights[2].intensity.blue = (float)1;




	int t;
	int x,y;

	/* 0. INIT CLOCKS */
        hit_clockSynchronizeAll();
        hit_clockStart( mainClock );
        hit_clockStart( initClock );
        hit_clockReset( sequentialClock );


	/* Data structure to keep the image */

    	/* 1. CREATE VIRTUAL TOPOLOGY */
    	HitTopology topo = hit_topology( plug_topArray2DComplete );

        /* 2. DECLARE FULL MATRIX WITHOUT MEMORY */
        HitTile_float image;
        hit_tileDomain( &image, float, 2, WIDTH, HEIGHT );

    	/* 3. COMPUTE PARTITION */
        HitShape parallelShape = hit_tileShape( image );
        HitLayout matLayout = hit_layout( plug_layBlocks, topo, parallelShape );
	HitShape parShape = hit_layShape(matLayout);

	/* DATA STRUCTURES */
	HitTile_float image_green, image_blue, image_red;
	hit_tileDomainShapeAlloc(&image_red, float,parShape); 
	hit_tileDomainShapeAlloc(&image_green, float,parShape); 
	hit_tileDomainShapeAlloc(&image_blue, float,parShape); 

	hit_tileForDimDomain(image_red, 1, y){
		hit_tileForDimDomain(image_red, 0, x){
			hit_tileElemAt( image_red, 2, x, y ) = 0; 	
			hit_tileElemAt( image_blue, 2, x, y ) = 0; 	
			hit_tileElemAt( image_green, 2, x, y ) = 0; 	
		}
	}

        hit_clockStart( sequentialClock );

#define element(a,x,y) hit_tileElemAtNoStride( a, 2, x, y )

	int x2,y2; 
	int movement=20;

	// Iterate through movements
	for(t=0;t<movement-1;t++){
				
		// Position the sphere in the t movement
		sphe.pos.x = vector_movement[0] * (float)pos_sphere[0]+ vector_movement[1]*(float)t;
		sphe.pos.y = vector_movement[0] * (float)pos_sphere[1]+ vector_movement[1]*(float)t;
		sphe.pos.z = vector_movement[0] * (float)pos_sphere[2]+ vector_movement[1]*(float)t;


		// Iterate over every 'pixel' of our screen
		// x, y for screen 'resolution' 
		hit_tileForDimDomain(image_red, 1, y){
			hit_tileForDimDomain(image_red, 0, x){

				x2=x+ hit_layDimBegin( matLayout, 0 ) ;
				y2=y+ hit_layDimBegin( matLayout, 1 ) ;
				float red = 0, green = 0, blue = 0;
				int sal = compute(x2, y2, &red, &green, &blue, sphe, lights, materials); 
			
				// Obtaining the boundary box of the sphere
				if(sal==1){
					hit_tileElemAtNoStride( image_red, 2, x, y ) = (float)0.5*red;
					hit_tileElemAtNoStride( image_blue, 2, x, y ) = (float)0.5*blue; 	
					hit_tileElemAtNoStride( image_green, 2, x, y ) = (float)0.5*green;
				}

			}
		}
		

	}

	/* Final iteration */
	// Position the sphere in the t movement
	sphe.pos.x = vector_movement[0] * (float)pos_sphere[0]+ vector_movement[1]*(float)t;
	sphe.pos.y = vector_movement[0] * (float)pos_sphere[1]+ vector_movement[1]*(float)t;
	sphe.pos.z = vector_movement[0] * (float)pos_sphere[2]+ vector_movement[1]*(float)t;

	// Iterate over every 'pixel' of our screen
	// x, y for screen 'resolution' 
	hit_tileForDimDomain(image_red, 1, y){
		hit_tileForDimDomain(image_red, 0, x){

			x2=x+ hit_layDimBegin( matLayout, 0 ) ;
			y2=y+ hit_layDimBegin( matLayout, 1 ) ;
			float red = 0, green = 0, blue = 0;
			int sal = compute(x2, y2, &red, &green, &blue, sphe, lights, materials); 
		
			// Obtaining the boundary box of the sphere
			if(sal==1){
				hit_tileElemAtNoStride( image_red, 2, x, y ) = red;
				hit_tileElemAtNoStride( image_blue, 2, x, y ) = blue; 	
				hit_tileElemAtNoStride( image_green, 2, x, y ) = green;
			}
		}
	}


	/* CLOSE TIMING */
	hit_clockStop( mainClock );
	hit_clockStop( sequentialClock );
        hit_clockReduce( matLayout, mainClock );
        hit_clockReduce( matLayout, sequentialClock );

	/* Final images */
	if( hit_layNumActives(matLayout) ==1){ 
		unsigned char img[3*WIDTH*HEIGHT];
		for(y=0;y<HEIGHT;y++){
			for(x=0;x<WIDTH;x++){
				img[(x + y*WIDTH)*3 + 0] = (unsigned char)min(hit_tileElemAt( image_red, 2, x, y )*255.0f, 255.0f);
				img[(x + y*WIDTH)*3 + 1] = (unsigned char)min(hit_tileElemAt( image_green, 2, x, y )*255.0f, 255.0f);
				img[(x + y*WIDTH)*3 + 2] = (unsigned char)min(hit_tileElemAt( image_blue, 2, x, y )*255.0f, 255.0f);
			}
		}
		saveppm("image_hit.ppm", img, WIDTH, HEIGHT);
	}

	{
		double sum=0,sumTotal=0;
		hit_tileForDimDomain(image_red, 1, y){
			hit_tileForDimDomain(image_red, 0, x){
				sum=(sum+ sqrt( hit_tileElemAt( image_red, 2, x, y ) +	
					       hit_tileElemAt( image_green, 2, x, y ) +
					       hit_tileElemAt( image_blue, 2, x, y )));
			}
		}
		// printf("\n Acumulated sum: %lf\n",sum);
		// Reduction
		 MPI_Reduce( &sum, &sumTotal, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );

		// Print result
		if (hit_Rank==0){
			printf("Clock main: %lf\n", mainClock.max );
			printf("Clock seq: %lf\n", sequentialClock.max );

			printf("\n ----------------------- NORM ----------------------- \n");
			printf("\n Acumulated sum: %lf",sumTotal);

    			double resultado=sqrt( sumTotal );
    			printf("\n Result: %lf \n",resultado);
	
    			printf("\n ---------------------------------------------------- \n");
		}
	}

	/* 6. FREE OTHER RESOURCES */
	hit_tileFree(image_red);
	hit_tileFree(image_blue);
	hit_tileFree(image_green);
        hit_layFree( matLayout );
        hit_topFree( topo );
        hit_comFinalize();
	return 0; 
}
