/*
* A simple ray tracer in a sequence of sphere movements
* Code based on the program of the 
* http://www.purplealienplanet.com/sites/default/files/raytrace_part1.c
*
* @author Ana Moretón Fernández
* @version v0.1
*
* (c) 2017, Grupo Trasgo, Universidad de Valladolid
*/


#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h> /* Needed for boolean datatype */
#include <math.h>

#define min(a,b) (((a) < (b)) ? (a) : (b))


/* The vector structure */
typedef struct{
      float x,y,z;
}vector;

/* The sphere */
typedef struct{
        vector pos;
        float  radius;
	int material;
}sphere; 

/* The ray */
typedef struct{
        vector start;
        vector dir;
}ray;

/* Colour */
typedef struct{
	float red, green, blue;
}colour;

/* Material Definition */
typedef struct{
	colour diffuse;
	float reflection;
}material;

/* Lightsource definition */
typedef struct{
	vector pos;
	colour intensity;
}light;

/* Subtract two vectors and return the resulting vector */
vector vectorSub(vector *v1, vector *v2){
	vector result = {v1->x - v2->x, v1->y - v2->y, v1->z - v2->z };
	return result;
}

/* Multiply two vectors and return the resulting scalar (dot product) */
float vectorDot(vector *v1, vector *v2){
	return v1->x * v2->x + v1->y * v2->y + v1->z * v2->z;
}

/* Calculate Vector x Scalar and return resulting Vector*/ 
vector vectorScale(float c, vector *v){
        vector result = {v->x * c, v->y * c, v->z * c };
        return result;
}

/* Add two vectors and return the resulting vector */
vector vectorAdd(vector *v1, vector *v2){
        vector result = {v1->x + v2->x, v1->y + v2->y, v1->z + v2->z };
        return result;
}

/* Check if the ray and sphere intersect */
bool intersectRaySphere(ray *r, sphere *s, float *t){
	
	bool retval = false;

	/* A = d.d, the vector dot product of the direction */
	float A = vectorDot(&r->dir, &r->dir); 
	
	/* We need a vector representing the distance between the start of 
	 * the ray and the position of the circle.
	 * This is the term (p0 - c) 
	 */
	vector dist = vectorSub(&r->start, &s->pos);
	
	/* 2d.(p0 - c) */  
	float B = 2 * vectorDot(&r->dir, &dist);
	
	/* (p0 - c).(p0 - c) - r^2 */
	float C = vectorDot(&dist, &dist) - (s->radius * s->radius);
	
	/* Solving the discriminant */
	float discr = B * B - 4 * A * C;
	
	/* If the discriminant is negative, there are no real roots.
	 * Return false in that case as the ray misses the sphere.
	 * Return true in all other cases (can be one or two intersections)
	 * t represents the distance between the start of the ray and
	 * the point on the sphere where it intersects.
	 */
	if(discr < 0)
		retval = false;
	else{
		float sqrtdiscr = sqrtf(discr);
		float t0 = (-B + sqrtdiscr)/(2);
		float t1 = (-B - sqrtdiscr)/(2);
		
		/* We want the closest one */
		if(t0 > t1)
			t0 = t1;

		/* Verify t1 larger than 0 and less than the original t */
		if((t0 > 0.001f) && (t0 < *t)){
			*t = t0;
			retval = true;
		}else
			retval = false;
	}

return retval;
}

/* Output data as PPM file */
void saveppm(char *filename, unsigned char *img, int width, int height){
	/* FILE pointer */
	FILE *f;

	/* Open file for writing */
	f = fopen(filename, "w");

	/* PPM header info, including the size of the image */
	fprintf(f, "P6 %d %d %d\n", width, height, 255);

	/* Write the image data to the file - remember 3 byte per pixel */
	fwrite(img, 3, width*height, f);

	/* Make sure you close the file */
	fclose(f);
}

/* Perform computation of each pixel */
void compute (int x, int y, float *red_g, float *green_g, float *blue_g, sphere sphe, light lights[3], material materials[3]){

			ray r;
			float red = *red_g;
			float green = *green_g;
			float blue = *blue_g;

			int level = 0;
			float coef = 1.0;
			
			r.start.x = (float)x;
			r.start.y = (float)y;
			r.start.z = -2000;
			
			r.dir.x = 0;
			r.dir.y = 0;
			r.dir.z = 1;
			
			do{
				/* Find closest intersection */
				float t = 20000.0f;
				
				unsigned int i;
				if(!intersectRaySphere(&r, &sphe, &t))
						break;
				
				vector scaled = vectorScale(t, &r.dir);
				vector newStart = vectorAdd(&r.start, &scaled);
				
				/* Find the normal for this new vector at the point of intersection */
				vector n = vectorSub(&newStart, &sphe.pos);
				float temp = vectorDot(&n, &n);
				if(temp == 0) break;
				
				temp = 1.0f / sqrtf(temp);
				n = vectorScale(temp, &n);

				/* Find the material to determine the colour */
				material currentMat = materials[sphe.material];
				
				/* Find the value of the light at this point */
				unsigned int j;
				for(j=0; j < 3; j++){
					light currentLight = lights[j];
					vector dist = vectorSub(&currentLight.pos, &newStart);
					if(vectorDot(&n, &dist) <= 0.0f) continue;
					float ti = sqrtf(vectorDot(&dist,&dist));
					if(ti <= 0.0f) continue;
					
					ray lightRay;
					lightRay.start = newStart;
					lightRay.dir = vectorScale((1/ti), &dist);
					
					/* Lambert diffusion */
					float lambert = vectorDot(&lightRay.dir, &n) * coef; 
					red += lambert * currentLight.intensity.red * currentMat.diffuse.red;
					green += lambert * currentLight.intensity.green * currentMat.diffuse.green;
					blue += lambert * currentLight.intensity.blue * currentMat.diffuse.blue;
				}
				/* Iterate over the reflection */
				coef *= currentMat.reflection;
				
				/* The reflected ray start and direction */
				r.start = newStart;
				float reflect = 2.0f * vectorDot(&r.dir, &n);
				vector tmp = vectorScale(reflect, &n);
				r.dir = vectorSub(&r.dir, &tmp);

				level++;

			}while((coef > 0.0f) && (level < 15));

		*red_g =red;
		*green_g=green;
		*blue_g=blue;

}

/* Main program*/
int main(int argc, char *argv[]){
	
	if ( argc != 3 ) {
                fprintf(stderr, "\nUsage: %s <numRows> <numColumns> \n", argv[0]);
                exit(EXIT_FAILURE);
        }

	/* Width and height of out image */
 	int WIDTH = atoi( argv[1] );
	int HEIGHT = atoi( argv[2] );
	
	/* Initial position of the sphere */
	int pos_sphere[] ={
		100,100, 10, 100
	};

	/* Vector movement */
	int vector_movement[] = {
		1,1
	};


	
	material materials[3];
	materials[0].diffuse.red = 1;
	materials[0].diffuse.green = 0;
	materials[0].diffuse.blue = 0;
	materials[0].reflection = 0.2;
	
	materials[1].diffuse.red = 0;
	materials[1].diffuse.green = 1;
	materials[1].diffuse.blue = 0;
	materials[1].reflection = 0.5;
	
	materials[2].diffuse.red = 0;
	materials[2].diffuse.green = 0;
	materials[2].diffuse.blue = 1;
	materials[2].reflection = 0.9;
	
	/* Sphere */
	sphere sphere; 
	sphere.pos.x = pos_sphere[0];
	sphere.pos.y = pos_sphere[1];
	sphere.pos.z = pos_sphere[2];
	sphere.radius = pos_sphere[3];
	sphere.material = 0;


	/* lights */
	light lights[3];
	
	lights[0].pos.x = 0;
	lights[0].pos.y = 240;
	lights[0].pos.z = -100;
	lights[0].intensity.red = 1;
	lights[0].intensity.green = 1;
	lights[0].intensity.blue = 1;
	
	lights[1].pos.x = 3200;
	lights[1].pos.y = 3000;
	lights[1].pos.z = -1000;
	lights[1].intensity.red = 0.6;
	lights[1].intensity.green = 0.7;
	lights[1].intensity.blue = 1;

	lights[2].pos.x = 600;
	lights[2].pos.y = 0;
	lights[2].pos.z = -100;
	lights[2].intensity.red = 0.3;
	lights[2].intensity.green = 0.5;
	lights[2].intensity.blue = 1;


	int movement=200;

	int t;
	int x,y;
	int min_x=0, min_y=0, max_x=WIDTH, max_y=HEIGHT;




	/* Data structure to keep the image */
	float *image_red, *image_green, *image_blue ;
	image_red=(float*)malloc(WIDTH*HEIGHT*sizeof(float));
	image_blue=(float*)malloc(WIDTH*HEIGHT*sizeof(float));
	image_green=(float*)malloc(WIDTH*HEIGHT*sizeof(float));

	if(image_red==NULL || image_green==NULL || image_blue==NULL) {
		printf("\n Error");
		return 0;
	}
	for(y=0;y<HEIGHT;y++){
		for(x=0;x<WIDTH;x++){
			image_red[x + y*WIDTH]=-1;
			image_blue[x + y*WIDTH]=-1;
			image_green[x + y*WIDTH]=-1;
		}
	}




	// Iterate through movements
	for(t=0;t<movement;t++){
	
		/* Position the sphere in the t movement*/
		sphere.pos.x = vector_movement[0] * pos_sphere[0]+ vector_movement[1]*t;
		sphere.pos.y = vector_movement[0] * pos_sphere[1]+ vector_movement[1]*t;
		sphere.pos.z = vector_movement[0] * pos_sphere[2]+ vector_movement[1]*t;



		/* Iterate over every 'pixel' of our screen
		 */
		/* x, y for screen 'resolution' */
		for(y=0;y<HEIGHT;y++){
			for(x=0;x<WIDTH;x++){

			float red = 0;
			float green = 0;
			float blue = 0;
			compute(x, y, &red, &green, &blue, sphere, lights, materials); 
			image_red[(x + y*WIDTH)] =red;
			image_blue[(x + y*WIDTH)] =blue;
			image_green[(x + y*WIDTH)] =green;

			}

		}
	}


	/* Final images */ 
	unsigned char img[3*WIDTH*HEIGHT];
	for(y=0;y<HEIGHT;y++){
		for(x=0;x<WIDTH;x++){
			img[(x + y*WIDTH)*3 + 0] = (unsigned char)min(image_red[(x + y*WIDTH)]*255.0f, 255.0f);
			img[(x + y*WIDTH)*3 + 1] = (unsigned char)min(image_green[(x + y*WIDTH)]*255.0f, 255.0f);
			img[(x + y*WIDTH)*3 + 2] = (unsigned char)min(image_blue[(x + y*WIDTH)]*255.0f, 255.0f);
		}
	}
	saveppm("image.ppm", img, WIDTH, HEIGHT);
}
