/* 
 * FFT: Sequential Cooley-Tukey algorithm
 *
 * 		Obtained from Wikipedia (last visit Mar 18th, 2013):
 *		http://en.wikipedia.org/wiki/Cooley%E2%80%93Tukey_FFT_algorithm	
 *
 * Sequential implementation using Hitmap library
 * (c) Arturo Gonzalez-Escribano 2013
 *
 *
X0,...,N-1 <- ditfft2(x, N, s):             DFT of (x0, xs, x2s, ..., x(N-1)s):
    if N = 1 then
        X0 <- x0                                      trivial size-1 DFT base case
    else
        X0,...,N/2-1 <- ditfft2(x, N/2, 2s)             DFT of (x0, x2s, x4s, ...)
        XN/2,...,N-1 <- ditfft2(x+s, N/2, 2s)           DFT of (xs, xs+2s, xs+4s, ...)
        for k = 0 to N/2-1                           combine DFTs of two halves into full DFT:
            t <- Xk
            Xk <- t + exp(-2PI*i k/N) Xk+N/2
            Xk+N/2 <- t - exp(-2PI*i k/N) Xk+N/2
        endfor
    endif
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <complex.h>
#include <hitmap.h>

/* CLOCK */
HitClock	mainClock;
HitClock	setupClock;
HitClock	fftClock;

/* DECLARE NEW TYPES */
typedef double complex Complex;
hit_tileNewType( Complex );

/* GLOBAL DECLARATION OF PI, INITIALIZED IN main */
double PI;


/* FUNCTION: INITIALIZE LOCAL TILE */
void initV( HitTile_Complex tileV, int numElems ) {
	int i;
	int cont = 0;
	for ( i=0; i<numElems/2 - hit_tileDimBegin( tileV, 0 ); i++ ) hit_tileElemAt( tileV, 1, i ) = 1;
	for ( ; i<hit_tileDimCard( tileV ); i++ ) hit_tileElemAt( tileV, 1, i ) = 0;
}

/* FUNCTION: SHOW THE CONTENT OF A COMPLEX TILE */
void show(const char * s, HitTile_Complex buf ) {
	int i;
	printf("%s", s);
	for (i = 0; i < hit_tileCard( buf); i++) {
		Complex t = hit_tileElemAt( buf, 1, i );
		if (!cimag( t ))
			printf("%g ", creal( t ) );
		else
			printf("(%g, %g) ", creal( t ), cimag( t ));
	}
	printf("\n");
}

/* FUNCTION: RECURSIVE DIT FFT USING EXTRA PARAMETERS FOR TRACKING COORDINATES
 * BETTER PERFORMANCE DUE TO SKIPPING THE CREATION OF NEW TILES FOR SUBSELECTIONs */
void ditfft2( HitTile_Complex V, int beginV, HitTile_Complex OutV, int beginOutV, int N, int s ) {

	if ( N == 1 ) {
#ifdef DEBUG
		printf("Copying: %d to %d\n", hit_tileDimBegin( V, 0 ), hit_tileDimBegin( OutV, 0 ) );
#endif
		hit_tileElemAt( OutV, 1, beginOutV ) = hit_tileElemAt( V, 1, beginV );
	}
	else {
#ifdef DEBUG
		printf("Size: %d\n", N );
#endif

		ditfft2( V, beginV, OutV, beginOutV, N/2, 2*s );
		ditfft2( V, beginV+s, OutV, beginOutV+N/2, N/2, 2*s );

		int k;
		for ( k=0; k < N/2; k++ ) {
#ifdef DEBUG
			printf("Computing: %d to %d\n", beginOutV+k, beginOutV+k+N/2 );
#endif
			Complex u = hit_tileElemAt( OutV, 1, beginOutV+k );
			Complex t = cexp( -2 * PI * I * k/N ) * hit_tileElemAt( OutV, 1, beginOutV+k+N/2 );
			hit_tileElemAt( OutV, 1, beginOutV+k ) = u + t ;
			hit_tileElemAt( OutV, 1, beginOutV+k+N/2 ) = u - t;
		}
	}
}

/* FUNCTION: RECURSIVE DIT FFT WITH SELECTIONS IN TILE COORDINATES */
void ditfft2Select( HitTile_Complex V, HitLayout wholeV, HitTile_Complex OutV, HitLayout wholeOutV ) {

	/* 1.a. STOP RECURSION FOR NON-ACTIVE PROCESSOR */
	if ( ! hit_layActive( wholeV ) ) return;

	/* 1.b. ONE ACTIVE PROCESS IN THE GROUP: ALLOCATE OutV BEFORE DATA IS POTENTIALLY PROCESSED */
/*
	if ( ! hit_tileAllocated( OutV ) && hit_layNumActives( wholeV ) == 1 ) {
		hit_tileAlloc( &OutV );
	}
*/

	/* 1.c. STOP RECURSION FOR ONE DATA ELEMENT IN THE WHOLE LAYOUT */
	/*		I'M ACTIVE, I HAVE THE ELEMENT */
	HitShape fullShpV = hit_layFullShape( wholeV );
	int N = hit_shapeCard( fullShpV );
	if ( N == 1 ) {
#ifdef DEBUG
		printf("Copying: %d to %d\n", hit_tileDimBegin( V, 0 ), hit_tileDimBegin( OutV, 0 ) );
#endif
		hit_tileElemAt( OutV, 1, 0 ) = hit_tileElemAt( V, 1, 0 );
		return;
	}

#ifdef DEBUG
	printf("Size: %d\n", N );
#endif

	/* 2. GENERATE SHAPES FOR GLOBAL SUBPARTS */
	HitShape shpVeven = hit_shape( 1, hit_sig( 0, N-2, 2 ) );
	HitShape shpVodd = hit_shape( 1, hit_sig( 1, N-1, 2 ) );
	HitShape shpOutVleft = hit_shape( 1, hit_sig( 0, N/2-1, 1 ) );
	HitShape shpOutVright = hit_shape( 1, hit_sig( N/2, N-1, 1 ) );

	/* 3. SELECT LOCAL SUBPARTS */
	HitTile_Complex Vodd, Veven, OutVleft, OutVright;
	hit_tileSelect( &Veven, &V, hit_shapeIntersect( shpEven, hit_shape( V ) ) );
	hit_tileSelect( &Vodd, &V, hit_shapeIntersect( shpOdd, hit_shape( V )  ) );
	hit_tileSelect( &OutVleft, &OutV, hit_shapeIntersect( shpOutVleft, hit_shape( OutV ) ) );
	hit_tileSelect( &OutVright, &OutV, hit_shapeIntersect( shpOutVright, hit_shape( OutV ) ) );

	/* 4. OPTIMIZATION FOR ONLY ONE ACTIVE PROCESS IN THE GROUP: SEQUENTIAL RECURSION */
	/* ???? */
	/*
	if ( hit_layNumActives( wholeV ) == 1 ) {
		parallelQuickSort( &left, whole );
		parallelQuickSort( &right, whole );
		return;
	}
	*/

	/* 4. SPLIT WORK IN TWO SIMILARLY WEIGHTED GROUPS OF PROCESSORS */
	double weights[2] = { (double)N/2, (double)N/2 };
	HitShape twoElements = hit_shape( 1, hit_sigStd( 2 ) );
	HitTopology activesTopo = hit_layActivesTopology( wholeV );
	HitLayout groups = hit_layout( plug_layIndependentLB, activesTopo, twoElements, weights );
	hit_topFree( activesTopo );

	/* 5. REDISTRIBUTE DATA IN PROCESSORS OF THE CORRESPONDING GROUP */
	/* 5.a. COMPUTE LAYOUTS FOR THE DATA REDISTRIBUTION */
	HitTopology gt0 = hit_layGroupTopo( groups, hit_lgr_elementGroup( groups, 0 ) );
	HitTopology gt1 = hit_layGroupTopo( groups, hit_lgr_elementGroup( groups, 1 ) );
	HitLayout evenLayout = hit_layout( plug_layBlocks, gt0, shpVeven );
	HitLayout oddLayout = hit_layout( plug_layBlocks, gt1, shpVodd );
	HitLayout leftLayout = hit_layout( plug_layBlocks, gt0, shpOutVleft );
	HitLayout rightLayout = hit_layout( plug_layBlocks, gt1, shpOutVright );

	/* 5.b. ALLOCATE BUFFERS FOR LOCAL PARTS */
	HitTile_double newVeven, newVodd;
	hit_tileDomainShapeAlloc( &newVeven, double, hit_layShape( evenLayout ) );
	hit_tileDomainShapeAlloc( &newVodd, double, hit_layShape( oddLayout ) );
	/* TO OPTIMIZE */
	HitTile_double newOutVleft, newOutVright;
	hit_tileDomainShapeAlloc( &newOutVleft, double, hit_layShape( leftLayout ) );
	hit_tileDomainShapeAlloc( &newOutVright, double, hit_layShape( rightLayout ) );

	/* 5.c. DATA REDISTRIBUTION */
	hit_patternDoOnce( hit_patternRedistribute( wholeV, &Veven, &newVeven, HIT_DOUBLE, 1 ) );   
	hit_patternDoOnce( hit_patternRedistribute( wholeV, &Vodd, &newVodd, HIT_DOUBLE, 1 ) );
	hit_tileFree( &Veven );
	hit_tileFree( &Vodd );
	/* TO OPTIMIZE */
	hit_patternDoOnce( hit_patternRedistribute( wholeOutV, &OutVleft, &newOutVleft, HIT_DOUBLE, 1 ) );   
	hit_patternDoOnce( hit_patternRedistribute( wholeOutV, &OutVright, &newOutVright, HIT_DOUBLE, 1 ) );
	hit_tileFree( &OutVleft );
	hit_tileFree( &OutVright );


	/* 4. SPLIT WORK IN GROUPS */
	ditfft2Select( newVeven, evenLayout, newOutVleft, leftLayout );
	ditfft2Select( newVodd, oddLayout, newOutVright, rightLayout );

	/* COMPUTATION */
	/* TRAVERSE OutVleft */
	/* BRING THE CORRESPONDING PART OF OutVright */

	/* TRAVERSE OutVright */
	/* BRING THE CORRESPONDING PART OF OutVleft */

	int k;
	for ( k=0; k < N/2; k++ ) {
#ifdef DEBUG
		printf("Computing: %d to %d\n", hit_tileDimBegin( OutV, 0 )+k, 
										hit_tileDimBegin( OutV, 0 )+k+N/2 );
#endif
		Complex u = hit_tileElemAt( OutV, 1, k );
		Complex t = cexp( -2 * PI * I * k/N ) * hit_tileElemAt( OutV, 1, k+N/2 );
		hit_tileElemAt( OutV, 1, k ) = u + t ;
		hit_tileElemAt( OutV, 1, k+N/2 ) = u - t;
	}
}


/* MAIN */
int main(int argc, char *argv[]) {
	
#ifdef DEBUG
	setbuf( stderr, NULL );
	setbuf( stdout, NULL );
#endif

	/* 0. READ ARGUMENT */
	if ( argc < 2 ) {
		fprintf(stderr, "Usage: %s <numElemsLog2>\n", argv[0] );
		exit(-1);
	}
	int numElemsLog2 = atoi( argv[1] );
	if ( numElemsLog2 < 1 || numElemsLog2 > 32 ) {
		fprintf(stderr, "Usage: %s <numElemsLog2>  ( parameter should be in the range [ 1, 32 ] )\n", argv[0] );
		exit(-1);
	}

	/* 1. INIT CLOCKS */
	hit_comInit(&argc, &argv);
	hit_clockSynchronizeAll();
	hit_clockStart( mainClock );
	hit_clockStart( setupClock );

	/* CREATE COMM. TYPE FOR Complex */


    /* 2. CREATE VIRTUAL TOPOLOGY */
    HitTopology topo = hit_topology( plug_topPlain );

	/* 3. DECLARE FULL VECTORs WITHOUT MEMORY */
	int numElems = 1 << numElemsLog2;
	HitTile_Complex V, OutV;
	hit_tileDomain( &V, Complex, 1, numElems );
	hit_tileDomain( &OutV, Complex, 1, numElems );

    /* 4. COMPUTE ORIGINAL PARTITION */
	HitLayout lay = hit_layout(plug_layBlocks, topo, hit_tileShape( V ) );

	/* 5. CREATE AND ALLOCATE LOCAL TILE */
	HitTile_double tileV, tileOutV;
	hit_tileSelect( &tileV, &V, hit_layShape(lay) );
	hit_tileSelect( &tileOutV, &OutV, hit_layShape(lay) );
	hit_tileAlloc( &tileV );
	/* TO OPTIMIZE */
	hit_tileAlloc( &tileOutV );

	/* 6. INITIALIZE LOCAL TILE */
	initV( tileV, numElems );
	PI = atan( 1.0 ) * 4;
	hit_clockStop( setupClock );
	//show("Data: ", V);

	/* 7. START RECURSIVE FFT */
	hit_clockStart( fftClock );
	ditfft2Select( lay, tileV, tileOutV, 1 );

	/* 8. CLOCK RESULTS */
	hit_clockStop( fftClock );
	hit_clockStop( mainClock );

	hit_clockReduce( lay, mainClock );
	hit_clockReduce( lay, setupClock );
	hit_clockReduce( lay, fftClock );
	hit_clockPrintMax( mainClock );
	hit_clockPrintMax( setupClock );
	hit_clockPrintMax( fftClock );

#ifdef WRITE_DATA
	/* 9. WRITE RESULT VECTOR */
	hit_tileTextFileWrite( &tileOutV, "V.result.dtxt", HIT_FILE_ARRAY, HIT_FILE_COMPLEX_DOUBLE, 14, 12 );
#endif

	/* 10. FREE RESOURCES */
	hit_tileFree( tileV );
	hit_tileFree( tileOutV );
	hit_layFree( lay );
	hit_topFree( topo );

	hit_comFinalize();
	return 0;
}
