/* 
 * FFT: Sequential Cooley-Tukey algorithm
 *
 * 		Obtained from Wikipedia (last visit Mar 18th, 2013):
 *		http://en.wikipedia.org/wiki/Cooley%E2%80%93Tukey_FFT_algorithm	
 *
 * Sequential implementation using Hitmap library
 * (c) Arturo Gonzalez-Escribano 2013
 *
 *
X0,...,N-1 <- ditfft2(x, N, s):             DFT of (x0, xs, x2s, ..., x(N-1)s):
    if N = 1 then
        X0 <- x0                                      trivial size-1 DFT base case
    else
        X0,...,N/2-1 <- ditfft2(x, N/2, 2s)             DFT of (x0, x2s, x4s, ...)
        XN/2,...,N-1 <- ditfft2(x+s, N/2, 2s)           DFT of (xs, xs+2s, xs+4s, ...)
        for k = 0 to N/2-1                           combine DFTs of two halves into full DFT:
            t <- Xk
            Xk <- t + exp(-2PI*i k/N) Xk+N/2
            Xk+N/2 <- t - exp(-2PI*i k/N) Xk+N/2
        endfor
    endif
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <complex.h>
#include <hitmap.h>

/* CLOCK */
HitClock	total;
HitClock	setup;
HitClock	fft;

/* DECLARE NEW TYPES */
typedef double complex Complex;
hit_tileNewType( Complex );

/* GLOBAL DECLARATION OF PI, INITIALIZED IN main */
double PI;


/* FUNCTION: SHOW THE CONTENT OF A COMPLEX TILE */
void show(const char * s, HitTile_Complex buf ) {
	int i;
	printf("%s", s);
	for (i = 0; i < hit_tileCard( buf); i++) {
		Complex t = hit_tileElemAt( buf, 1, i );
		if (!cimag( t ))
			printf("%g ", creal( t ) );
		else
			printf("(%g, %g) ", creal( t ), cimag( t ));
	}
	printf("\n");
}

/* FUNCTION: RECURSIVE DIT FFT USING EXTRA PARAMETERS FOR TRACKING COORDINATES
 * BETTER PERFORMANCE DUE TO SKIPPING THE CREATION OF NEW TILES FOR SUBSELECTIONs */
void ditfft2( HitTile_Complex V, int beginV, HitTile_Complex OutV, int beginOutV, int N, int s ) {

	if ( N == 1 ) {
#ifdef DEBUG
		printf("Copying: %d to %d\n", hit_tileDimBegin( V, 0 ), hit_tileDimBegin( OutV, 0 ) );
#endif
		hit_tileElemAt( OutV, 1, beginOutV ) = hit_tileElemAt( V, 1, beginV );
	}
	else {
#ifdef DEBUG
		printf("Size: %d\n", N );
#endif

		ditfft2( V, beginV, OutV, beginOutV, N/2, 2*s );
		ditfft2( V, beginV+s, OutV, beginOutV+N/2, N/2, 2*s );

		int k;
		for ( k=0; k < N/2; k++ ) {
#ifdef DEBUG
			printf("Computing: %d to %d\n", beginOutV+k, beginOutV+k+N/2 );
#endif
			Complex u = hit_tileElemAt( OutV, 1, beginOutV+k );
			Complex t = cexp( -2 * PI * I * k/N ) * hit_tileElemAt( OutV, 1, beginOutV+k+N/2 );
			hit_tileElemAt( OutV, 1, beginOutV+k ) = u + t ;
			hit_tileElemAt( OutV, 1, beginOutV+k+N/2 ) = u - t;
		}
	}
}

/* FUNCTION: RECURSIVE DIT FFT WITH SELECTIONS IN TILE COORDINATES */
void ditfft2Select( HitTile_Complex V, HitTile_Complex OutV ) {

	int N = hit_tileCard( V );
	if ( N == 1 ) {
#ifdef DEBUG
		printf("Copying: %d to %d\n", hit_tileDimBegin( V, 0 ), hit_tileDimBegin( OutV, 0 ) );
#endif
		hit_tileElemAt( OutV, 1, 0 ) = hit_tileElemAt( V, 1, 0 );
	}
	else {
#ifdef DEBUG
		printf("Size: %d\n", N );
#endif

		HitTile_Complex Vodd, Veven, OutVleft, OutVright;

		hit_tileSelect( &Veven, &V, hit_shape( 1, hit_sig( 0, N-2, 2 ) ) );
		hit_tileSelect( &Vodd, &V, hit_shape( 1, hit_sig( 1, N-1, 2 ) ) );

		hit_tileSelect( &OutVleft, &OutV, hit_shape( 1, hit_sig( 0, N/2-1, 1 ) ) );
		hit_tileSelect( &OutVright, &OutV, hit_shape( 1, hit_sig( N/2, N-1, 1 ) ) );

		ditfft2Select( Veven, OutVleft );
		ditfft2Select( Vodd, OutVright );

		int k;
		for ( k=0; k < N/2; k++ ) {
#ifdef DEBUG
			printf("Computing: %d to %d\n", hit_tileDimBegin( OutV, 0 )+k, 
											hit_tileDimBegin( OutV, 0 )+k+N/2 );
#endif
			Complex u = hit_tileElemAt( OutV, 1, k );
			Complex t = cexp( -2 * PI * I * k/N ) * hit_tileElemAt( OutV, 1, k+N/2 );
			hit_tileElemAt( OutV, 1, k ) = u + t ;
			hit_tileElemAt( OutV, 1, k+N/2 ) = u - t;
		}
	}
}

/* FUNCTION: RECURSIVE DIT FFT WITH SELECTIONS IN ARRAY COORDINATES */
void ditfft2SelectArrayCoords( HitTile_Complex V, HitTile_Complex OutV, int s ) {

	int N = hit_tileCard( V );
	if ( N <= 1 ) {
#ifdef DEBUG
		printf("Copying: %d to %d\n", hit_tileDimBegin( V, 0 ), hit_tileDimBegin( OutV, 0 ) );
#endif
		hit_tileElemAt( OutV, 1, 0 ) = hit_tileElemAt( V, 1, 0 );
	}
	else {
#ifdef DEBUG
		printf("Size: %d\n", N );
#endif

		HitTile_Complex Vodd, Veven, OutVleft, OutVright;

		int beginV = hit_tileDimBegin( V, 0 );
		int endV = hit_tileDimEnd( V, 0 );
		int beginOutV = hit_tileDimBegin( OutV, 0 );

		hit_tileSelectArrayCoords( &Veven, &V, hit_shape( 1, hit_sig( beginV, endV, 2*s ) ) );
		hit_tileSelectArrayCoords( &Vodd, &V, hit_shape( 1, hit_sig( beginV+s, endV, 2*s ) ) );
		hit_tileSelectArrayCoords( &OutVleft, &OutV, hit_shape( 1, hit_sig( beginOutV, beginOutV+ N/2-1, 1 ) ) );
		hit_tileSelectArrayCoords( &OutVright, &OutV, hit_shape( 1, hit_sig(  beginOutV+N/2,  beginOutV+N-1, 1 ) ) );

		ditfft2SelectArrayCoords( Veven, OutVleft, 2*s );
		ditfft2SelectArrayCoords( Vodd, OutVright, 2*s );

		int k;
		for ( k=0; k < N/2; k++ ) {
#ifdef DEBUG
			printf("Computing: %d to %d\n", hit_tileDimBegin( OutV, 0 )+k, 
											hit_tileDimBegin( OutV, 0 )+k+N/2 );
#endif
			Complex u = hit_tileElemAt( OutV, 1, k );
			Complex t = cexp( -2 * PI * I * k/N ) * hit_tileElemAt( OutV, 1, k+N/2 );
			hit_tileElemAt( OutV, 1, k ) = u + t ;
			hit_tileElemAt( OutV, 1, k+N/2 ) = u - t;
		}
	}
}

/* MAIN */
int main(int argc, char *argv[]) {
	HitTile_Complex V, OutV;
	
#ifdef DEBUG
	setbuf( stderr, NULL );
	setbuf( stdout, NULL );
#endif

	if ( argc < 2 ) {
		fprintf(stderr, "Usage: %s <numElemsLog2>\n", argv[0] );
		exit(-1);
	}
	int numElemsLog2 = atoi( argv[1] );
	if ( numElemsLog2 < 1 || numElemsLog2 > 32 ) {
		fprintf(stderr, "Usage: %s <numElemsLog2>  ( parameter should be in the range [ 1, 32 ] )\n", argv[0] );
		exit(-1);
	}

	// Hitmap Com INITIALIZATION NEEDED ONLY FOR THE CLOCKS (THEY INTERNALLY USE MPI_Wtime)
	hit_comInit(&argc, &argv);
	hit_clockStart( total );
	hit_clockStart( setup );

	int numElems = 1 << numElemsLog2;

	hit_tileDomainAlloc( &V, Complex, 1, numElems );
	hit_tileDomainAlloc( &OutV, Complex, 1, numElems );

	int i;
	for ( i=0; i<numElems/2; i++ ) hit_tileElemAt( V, 1, i ) = 1;
	for ( i=numElems/2; i<numElems; i++ ) hit_tileElemAt( V, 1, i ) = 0;

	PI = atan( 1.0 ) * 4;

	printf("Num. Elements: %d\n", numElems);
	//show("Data: ", V);
	
	hit_clockStop( setup );
	hit_clockStart( fft );

#ifdef WITHOUT_SELECT
	ditfft2( V, 0, OutV, 0, numElems, 1 );
#else
	#ifdef ARRAY_COORDS
	ditfft2SelectArrayCoords( V, OutV, 1 );
	#else
	ditfft2Select( V, OutV );
	#endif
#endif

	hit_clockStop( fft );
	hit_clockStop( total );

	hit_clockPrint( setup );
	hit_clockPrint( fft );
	hit_clockPrint( total );

	hit_comFinalize();

#ifdef WRITE_RESULT
	show("\nFFT : ", OutV);
#endif
 
	return 0;
}
