
/* 
 * FFT: Sequential Cooley-Tukey algorithm
 *
 * 		Obtained from Wikipedia (last visit Mar 18th, 2013):
 *		http://en.wikipedia.org/wiki/Cooley%E2%80%93Tukey_FFT_algorithm	
 *
 * Direct C Implementation 
 * (c) Arturo Gonzalez-Escribano, 2013
 *
X0,...,N-1 <- ditfft2(x, N, s):             DFT of (x0, xs, x2s, ..., x(N-1)s):
    if N = 1 then
        X0 <- x0                                      trivial size-1 DFT base case
    else
        X0,...,N/2-1 <- ditfft2(x, N/2, 2s)             DFT of (x0, x2s, x4s, ...)
        XN/2,...,N-1 <- ditfft2(x+s, N/2, 2s)           DFT of (xs, xs+2s, xs+4s, ...)
        for k = 0 to N/2-1                           combine DFTs of two halves into full DFT:
            t <- Xk
            Xk <- t + exp(-2PI*i k/N) Xk+N/2
            Xk+N/2 <- t - exp(-2PI*i k/N) Xk+N/2
        endfor
    endif
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <complex.h>

double PI;

void ditfft2( double complex V[], int beginV, double complex OutV[], int beginOutV, int N, int s ) {
	if ( N == 1 ) {
#ifdef DEBUG
		printf("Copying: %d to %d\n", beginV, beginOutV );
#endif
		OutV[beginOutV] = V[beginV];
	}
	else {
		ditfft2( V, beginV, OutV, beginOutV, N/2, 2*s );
		ditfft2( V, beginV+s, OutV, beginOutV+N/2, N/2, 2*s );

		int k;
		for ( k=0; k < N/2; k++ ) {
#ifdef DEBUG
			printf("Computing: %d to %d\n", beginOutV+k, beginOutV+k+N/2 );
#endif
			double complex t = OutV[ beginOutV+k ];
			double complex f = cexp( -2 * PI * I * k/N ) * OutV[ beginOutV+k+N/2 ];
			OutV[ beginOutV+k ] = t + f ;
			OutV[ beginOutV+k+N/2 ] = t - f;
		}
	}
}

int main(int argc, char *argv[]) {

	if ( argc < 2 ) {
		fprintf(stderr, "Usage: %s <numElemsLog2>\n", argv[0] );
		exit(-1);
	}
	int numElemsLog2 = atoi( argv[1] );
	if ( numElemsLog2 < 1 || numElemsLog2 > 32 ) {
		fprintf(stderr, "Usage: %s <numElemsLog2>  ( parameter should be in the range [ 1, 32 ] )\n", argv[0] );
		exit(-1);
	}

	int numElems = 1 << numElemsLog2;

	double complex *V, *OutV;
	V = (double complex *)malloc( numElems * sizeof(double complex) );
	OutV = (double complex *)malloc( numElems * sizeof(double complex) );

	int k;
	for ( k=0; k<numElems/2; k++ ) V[k] = 1;
	for ( k=numElems/2; k<numElems; k++ ) V[k] = 0;

	PI = atan( 1.0 ) * 4;

	void show(const char * s, complex double buf[] ) {
		int i;
		printf("%s", s);
		for (i = 0; i < numElems; i++)
			if (!cimag(buf[i]))
				printf("%g ", creal(buf[i]));
			else
				printf("(%g, %g) ", creal(buf[i]), cimag(buf[i]));
			printf("\n");
	}

	printf("Num. Elements: %d\n", numElems);
	// show("Data: ", V);
	ditfft2( V, 0, OutV, 0, numElems, 1 );
#ifdef WRITE_RESULT
	show("\nFFT : ", OutV);
#endif
 
	return 0;
}
