/**
 * Mandelbrot set benchmark
 *
 * @file mandelbrot.c
 * @author Javier Fresno
 * @date 15-3-2012
 */

/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2007-2013, Javier Fresno Bausela
 *
 * All Rights Reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <hitmap.h>
#include <hit_com.h>
#include <hit_worker.h>
#include <string.h>


#ifndef GRAIN
#define GRAIN 16
#endif

// Maximum iterations
int maxIter = 1000;

// Maximum absolute value
static const double maxAbs = 2.0;

// area of the mandelbrot set
static const double leftMargin	= -1.5;
static const double topMargin	=  1.0;
static const double width		=  2.2;
static const double height		=  2.0;

// resolution in rows and columns
int rows = 32;
int cols = 32;
double xRatio;
double yRatio;


hit_tileNewType(int);


/*
 * Complex represents complex numbers. 
 */
typedef struct{
	double real;
	double imag;
} Complex;

/**
 * ComplexSet represents a set of complex numbers.
 */
typedef struct{
	int row, col;
	Complex elems[GRAIN*GRAIN];
} ComplexSet;


/**
 * ComplexSet represents a set of results.
 */
typedef struct{
	int row;
	int col;
	int elems[GRAIN*GRAIN];
} ResultSet;


// Declare de HitTask_* types
hit_taskNewType(ComplexSet);
hit_taskNewType(ResultSet);


// Abs
// actually, this function calculates abs(this)^2. 
// but we don't need the square root.
inline double complex_abs(Complex c){
	return c.real * c.real + c.imag * c.imag;
}

// Adds c2 to c
inline void complex_add(Complex * c, Complex c2){
	c->real += c2.real;
	c->imag += c2.imag;
}

// Square root of c
inline void complex_square(Complex * c){
	double real = c->real;
	double imag = c->imag;
	c->real = real * real - imag * imag;
	c->imag = 2 * real * imag;
}



/**
 * Function for the Initial skeleton. 
 * Calculates complex numbers depending on the resolution.
 */
void init(HitWorker * worker){


	HitTask_ComplexSet task;
	hit_taskAlloc(&task,ComplexSet);

	// Loops for the block
	int i;
	for(i=0; i<rows; i+=GRAIN){
		int j;
		for(j=0; j<cols; j+=GRAIN){


			hit_taskData(task).row = i;
			hit_taskData(task).col = j;

			// Loops for the block elements
			int ii;
			for(ii=0; ii<GRAIN; ii++){
				int jj;
				for(jj=0; jj<GRAIN; jj++){

					double real = leftMargin + (jj+j) * xRatio;
					double imag = topMargin - (ii+i) * yRatio;

					hit_taskData(task).elems[ii*GRAIN+jj].real = real;
					hit_taskData(task).elems[ii*GRAIN+jj].imag = imag;

				}
			}

			hit_workerSendTask(worker, &task, 0);

		}

	}


	hit_taskFree(&task);
}


/*
 * Function of the worker of the farm
 * Iterates the mandelbrot sequence (z_n+1 -> z_n² + c, z_0 = 0).
 */
void iterate(HitWorker * worker){

	HitTask_ComplexSet intask;
	HitTask_ResultSet rtask;
	hit_taskAlloc(&rtask,ResultSet);

	while(hit_workerGetTask(worker,&intask)){

		hit_taskData(rtask).row = hit_taskData(intask).row;
		hit_taskData(rtask).col = hit_taskData(intask).col;

		// Loop for each element in the set
		int e;
		for(e=0; e < (GRAIN*GRAIN); e++){

			Complex elem =  hit_taskData(intask).elems[e];
			Complex tmp = {elem.real, elem.imag};
	
			int iterations = 0;
			double cabs = complex_abs(elem);
	
			while(cabs <= maxAbs && iterations < maxIter){
				complex_square(&tmp);
				complex_add(&tmp,elem);
				cabs = complex_abs(tmp);
				iterations++;
			}

			if(iterations == maxIter){
				hit_taskData(rtask).elems[e] = 1;
			} else {
				hit_taskData(rtask).elems[e] = 0;
			}

		}

		rtask.weight = 43;
		hit_workerSendTask(worker, &rtask, 0);
		hit_taskFree(&intask);
	}

	hit_taskFree(&rtask);
}


void dumpShape(HitShape sh){

	int i;
	printf("[%d:%d:%d",hit_shapeSig(sh,0).begin,hit_shapeSig(sh,0).end,hit_shapeSig(sh,0).stride);
	for(i=1;i<hit_shapeDims(sh);i++) printf(",%d:%d:%d",hit_shapeSig(sh,i).begin,hit_shapeSig(sh,i).end,hit_shapeSig(sh,i).stride);
	printf("] \t cards: [%d",hit_sigCard(hit_shapeSig(sh,0)));
	for(i=1;i<hit_shapeDims(sh);i++) printf(",%d",hit_sigCard(hit_shapeSig(sh,i)));
	printf("]\n");
}




void fin(HitWorker * worker){

	HitShape shape;
	HitTile_int tile;

	shape = hit_shape(2,hit_sig(0,rows-1,1),hit_sig(0,cols-1,1));
	//dumpShape(shape);

	hit_tileDomainShapeAlloc(&tile,int,shape);


	HitTask_ResultSet rtask;
	while(hit_workerGetTask(worker,&rtask)){

		//printf("Reading here ===================\n");
		ResultSet r = hit_taskData(rtask);

		int i;
		for(i=0; i<GRAIN; i++){
			int j;
			for(j=0; j<GRAIN; j++){


				hit_tileElemAt(tile, 2, r.row+i, r.col+j) =
				hit_taskData(rtask).elems[i*GRAIN+j];

			}
		}

	}

	int num = 0;
	int i,j;
	for (i = 0; i < rows; i++) {
		for (j = 0; j < cols; j++) {
			if(hit_tileElemAt(tile, 2, i, j) == 1) num++;
		}
	}
	
	printf("Count %d\n",num);

}




/**
 * Main function
 */
int main(int argc, char ** argv) {



	hit_comInit(&argc,&argv);
	hit_workerContextInit();

	double t1, t2;
	t1 = MPI_Wtime();

	int buffer_size = 0;
	int threshold_send = 0;
	int enable_ws = 1;
	int threshold_ws_request = 0;
	int threshold_ws_send = 0;


	if(argc < 3){
		printf("Usage: %s [rows] [cols] [iter] {bsize} {thsend} {ws} {th.ws.req} {th.ws.send}\n",argv[0]);
		return -1;
	} else {
		rows = atoi(argv[1]);
		cols = atoi(argv[2]);
		maxIter = atoi(argv[3]);

		if(argc>4){
			buffer_size = atoi(argv[4]);
			threshold_ws_request = buffer_size / 10;
			threshold_ws_send = threshold_ws_request * 2;
		}

		if(argc>5) threshold_send = atoi(argv[5]);
		if(argc>6) enable_ws = atoi(argv[6]);
		if(argc>7) threshold_ws_request = atoi(argv[7]);
		if(argc>8) threshold_ws_send = atoi(argv[8]);
	}

	if( hit_NProcs < 2 ){
		printf("I need 3 procs at least\n");
		return -1;
	}


	if(buffer_size != 0) HIT_WORKER_CONFIG.buffer_in = HIT_WORKER_CONFIG.buffer_out = buffer_size;
	if(threshold_send !=0) HIT_WORKER_CONFIG.threshold_send  = threshold_send;
	if(threshold_ws_request !=0) HIT_WORKER_CONFIG.threshold_ws_request = threshold_ws_request;
	if(threshold_ws_send !=0) HIT_WORKER_CONFIG.threshold_ws_send = threshold_ws_send;


	if(hit_Rank == 0){
		printf("======= Config =======\n");
		printf("Problem size: %dx%d\n",rows,cols);
		printf("Grain %d\n",GRAIN);
		printf("Max iterations: %d\n",maxIter);
		printf("Buffer size: I:%d O:%d\n",HIT_WORKER_CONFIG.buffer_in,HIT_WORKER_CONFIG.buffer_out);
		printf("Default task weight: %f\n",( HIT_WORKER_CONFIG.threshold_send / (double) HIT_WORKER_CONFIG.buffer_out));
		printf("Threshold send: %f\n",HIT_WORKER_CONFIG.threshold_send);
		printf("Enable WS: %d\n",enable_ws);
		printf("Threshold ws request: %f\n",HIT_WORKER_CONFIG.threshold_ws_request);
		printf("Threshold ws send: %f\n",HIT_WORKER_CONFIG.threshold_ws_send);
		printf("======================\n");
	}


	int nWorkers = hit_NProcs - 2;

	if(hit_Rank == 0) printf("nWorkers = %d\n", nWorkers);

	xRatio     = width/cols;
	yRatio     = height/rows;

	HitType HitComplex, HitComplexSet;
	hit_comTypeStruct(&HitComplex,Complex,2, real,1,HIT_DOUBLE, imag,1,HIT_DOUBLE);
	hit_comTypeStruct(&HitComplexSet,ComplexSet,3, row,1,HIT_INT, col,1,HIT_INT, elems,GRAIN*GRAIN,HitComplex);

	HitType HitResultSet;
	hit_comTypeStruct(&HitResultSet,ResultSet,3, row,1,HIT_INT, col,1,HIT_INT, elems,GRAIN*GRAIN, MPI_INT);


	HitWorker w1;
	hit_workerCreate(&w1, init, 0,1, HitComplexSet);

	HitWorker w2 ;
	hit_workerCreate(&w2, iterate, 1,1, HitComplexSet,HitResultSet);

	HitWorker farm;
	hit_workerCreateFarm(&farm,nWorkers,&w2);
	farm.enable_ws = enable_ws;

	HitWorker w3;
	hit_workerCreate(&w3, fin, 1,0, HitResultSet);

	HitWorker pipe;
	hit_workerOpSuccession(&pipe,3,&w1,&farm,&w3);
	hit_workerExecute(&pipe);

	MPI_Barrier(MPI_COMM_WORLD);
	t2 = MPI_Wtime();
	if(hit_Rank == 0) printf("time: %f\n",t2-t1);

	hit_comFinalize();

	return 0;
}


