/**
 * Mandelbrot set benchmark (without variable grain, just one
 * element in each communication)
 *
 * @file mandelbrot.c
 * @author Javier Fresno
 * @date 25-3-2012
 */

/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2007-2013, Javier Fresno Bausela
 *
 * All Rights Reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <hitmap.h>
#include <hit_com.h>
#include <hit_worker.h>
#include <string.h>


double leftMargin = -1.5;
double topMargin  =  1.0;
double width      =  2.2;
double height     =  2.0;
double xRatio;
double yRatio ;

int    maxIter = 1000;
double maxAbs  = 2.0;

int rows;
int cols;

hit_tileNewType(int);

typedef struct{
	double real;
	double imag;
	int i;
	int j;
} Complex;

typedef struct{
	int i;
	int j;
	int result;
} Result;



hit_taskNewType(Complex);
hit_taskNewType(Result);



void complex_init(Complex * c, int i, int j, double real, double imag){
	c->i = i;
	c->j = j;
	
	c->real = real;
	c->imag = imag;
}

double complex_abs(Complex c){
	return c.real * c.real + c.imag * c.imag;
}

void complex_add(Complex * c, Complex c2){
	c->real += c2.real;
	c->imag += c2.imag;
}

void complex_square(Complex * c){
	double real = c->real;
	double imag = c->imag;
	c->real = real * real - imag * imag;
	c->imag = 2 * real * imag;
}




void init(HitWorker * worker){

	HitTask_Complex task;
	hit_taskAlloc(&task,Complex);

	int i;
	for(i=0; i<rows; i++){
		
		int j;
		for(j=0; j<cols; j++){


			//double real = leftMargin + i * xRatio;
			//double imag = topMargin  - j * yRatio;

			double real = leftMargin + j * xRatio;
			double imag = topMargin  - i * yRatio;

			complex_init(&hit_taskData(task), i, j, real, imag);

			//printf("init (%d,%d)\n",i,j);

			hit_workerSendTask(worker, &task,0);

		}

	}

	hit_taskFree(&task);
}



void iterate(HitWorker * worker){

	int num = 0;

	HitTask_Complex intask;
	HitTask_Result rtask;
	hit_taskAlloc(&rtask,Result);

	while(hit_workerGetTask(worker,&intask)){

		Complex c = hit_taskData(intask);

		//printf("iterate [%d,%d]\n",c.i,c.j);
		num++;

		int iterations = 0;
		int maxIterations = maxIter;
		double maxAbsolute = maxAbs;
		double cabs = complex_abs(c);
		Complex tmp = c;

		while (cabs <= maxAbsolute && iterations < maxIterations) {
			complex_square(&tmp);
			complex_add(&tmp,c);
			cabs = complex_abs(tmp);
			iterations++;
		}

		Result r;
		r.i = c.i;
		r.j = c.j;

		if (iterations == maxIterations)
			r.result = 1;
		else
			r.result = 0;

		hit_taskData(rtask) = r;
		hit_workerSendTask(worker, &rtask,0);
		hit_taskFree(&intask);
	}

	printf("Num iterations %d\n",num);
	hit_taskFree(&rtask);

}






void fin(HitWorker * worker){

	HitShape shape;
	HitTile_int tile;

	shape = hit_shape(2,hit_sig(0,rows-1,1),hit_sig(0,cols-1,1));
	hit_tileDomainShapeAlloc(&tile,int,shape);


	HitTask_Result rtask;
	while(hit_workerGetTask(worker,&rtask)){

		Result r = hit_taskData(rtask);
		hit_tileElemAt(tile, 2, r.i, r.j) = r.result;
		//printf("result[%d,%d] = %d\n",r.i,r.j,r.result);
	}

	int num = 0;

	//printf("rows %d, cols %d\n",rows,cols);

	int i,j;
	for (i = 0; i < rows; i++) {
		for (j = 0; j < cols; j++) {
			if(hit_tileElemAt(tile, 2, i, j) == 1) num++;
		}
	}
	
	/*
	for (i = 0; i < rows; i++) {
		for (j = 0; j < cols; j++) {
			printf("[%d,%d]:%d", i,j,hit_tileElemAt(tile, 2, i, j));
		}
		printf("\n");
	}
	*/

	printf("Count %d\n",num);

}




/**
 * Main function
 */
int main(int argc, char ** argv) {

	hit_comInit(&argc,&argv);
	hit_workerContextInit();

	double t1, t2;
	t1 = MPI_Wtime();

	//if(hit_Rank == 1){ stdout = fopen("newout.txt", "w"); stderr = stdout; }

	//printf("Rank %d\n",hit_Rank);

	int buffer_size = 0;
	int threshold_send = 0;
	int enable_ws = 1;
	int threshold_ws_request = 0;
	int threshold_ws_send = 0;


	if(argc < 3){
		printf("Usage: %s [rows] [cols] [iter] {bsize} {thsend} {ws} {th.ws.req} {th.ws.send}\n",argv[0]);
		return -1;
	} else {
		rows = atoi(argv[1]);
		cols = atoi(argv[2]);
		maxIter = atoi(argv[3]);

		if(argc>4) buffer_size = atoi(argv[4]);
		if(argc>5) threshold_send = atoi(argv[5]); // @todo They are double now
		if(argc>6) enable_ws = atoi(argv[6]);
		if(argc>7) threshold_ws_request = atoi(argv[7]);
		if(argc>8) threshold_ws_send = atoi(argv[8]);
	}

	if( hit_NProcs < 3 ){
		printf("I need 3 procs at least\n");
		return -1;
	}


	if(buffer_size != 0) HIT_WORKER_CONFIG.buffer_in = HIT_WORKER_CONFIG.buffer_out = buffer_size;
	if(threshold_send !=0) HIT_WORKER_CONFIG.threshold_send  = threshold_send;
	if(threshold_ws_request !=0) HIT_WORKER_CONFIG.threshold_ws_request = threshold_ws_request;
	if(threshold_ws_send !=0) HIT_WORKER_CONFIG.threshold_ws_send = threshold_ws_send;

	if(hit_Rank == 0){
		printf("======= Config =======\n");
		printf("Problem size: %dx%d\n",rows,cols);
		printf("Max iterations: %d\n",maxIter);
		printf("Buffer size: I:%d O:%d\n",HIT_WORKER_CONFIG.buffer_in,HIT_WORKER_CONFIG.buffer_out);
		printf("Default task weight: %f\n",( HIT_WORKER_CONFIG.threshold_send / (double) HIT_WORKER_CONFIG.buffer_out));
		printf("Threshold send: %f\n",HIT_WORKER_CONFIG.threshold_send);
		printf("Enable WS: %d\n",enable_ws);
		printf("Threshold ws request: %f\n",HIT_WORKER_CONFIG.threshold_ws_request);
		printf("Threshold ws send: %f\n",HIT_WORKER_CONFIG.threshold_ws_send);
		printf("======================\n");
	}



	int nWorkers = hit_NProcs - 2;

	if(hit_Rank == 0) printf("nWorkers = %d\n", nWorkers);

	xRatio     = width/cols;
	yRatio     = height/rows;


	 HitType HitComplex;
	 hit_comTypeStruct(&HitComplex,Complex,4,
			 real,1,HIT_DOUBLE,
			 imag,1,HIT_DOUBLE,
			 i,1,HIT_INT,
			 j,1,HIT_INT);

	 HitType HitResult;

	 hit_comTypeStruct(&HitResult,Result,3,
			 i,1,HIT_INT,
			 j,1,HIT_INT,
			 result,1,HIT_INT);

	HitWorker w1;
	hit_workerCreate(&w1, init,0,1,HitComplex);

	HitWorker w2 ;
	hit_workerCreate(&w2, iterate,1,1,HitComplex,HitResult);

	HitWorker farm;
	hit_workerCreateFarm(&farm,nWorkers,&w2);
	farm.enable_ws = enable_ws;

	HitWorker w3;
	hit_workerCreate(&w3, fin, 1,0, HitResult);

	HitWorker pipe;
	hit_workerOpSuccession(&pipe,3,&w1,&farm,&w3);
	hit_workerExecute(&pipe);

	MPI_Barrier(MPI_COMM_WORLD);
	t2 = MPI_Wtime();
	if(hit_Rank == 0) printf("time: %f\n",t2-t1);

	hit_comFinalize();

	return 0;
}


