/**
 * 
 * Hitmap functions to allocate memory. This is a set of macro-functions that
 * allocate memory and check the result. They print an error message if the
 * allocate routine fails.
 * 
 * @file hit_allocP.h
 * @version 1.0
 * @author Javier Fresno Bausela
 * @date Oct 2011
 * 
 */

/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2007-2011, Arturo Gonzalez Escribano
 *                          Javier Fresno Bausela
 *                          Carlos de Blas Carton
 *                          Yuri Torres de la Sierra
 * All Rights Reserved.
 */


#ifndef _HitAllocP_
#define _HitAllocP_


#include <stdlib.h>
#include <hit_error.h>



/**
 * Allocates an array of nmemb elements.
 * @fn void hit_calloc(void * ptr, size_t nmemb, size_t size)
 * This macros do not use cast because the stdlib malloc function should return
 * a void pointer.
 * 
 * @version 1.0
 * @author Javier Fresno Bausela
 * 
 * @param ptr   Pointer to the allocated memory.
 * @param nmemb Number of elements.
 * @param size  Size of each element.
 */
#ifdef __cplusplus

#define hit_calloc(ptr, nmemb, size, cast) \
{ \
	ptr = (cast) calloc(nmemb, size); \
	if(ptr == NULL) hit_errInternal(__func__,"Memory allocation request failed","",__FILE__,__LINE__) \
}
#else

#define hit_calloc(ptr, nmemb, size, cast) \
{ \
	ptr = calloc(nmemb, size); \
	if(ptr == NULL) hit_errInternal(__func__,"Memory allocation request failed","",__FILE__,__LINE__) \
}

#endif




/**
 * Allocates a memory block of size bytes.
 * @fn void hit_malloc(void * ptr, size_t size)
 * @version 1.0
 * @author Javier Fresno Bausela
 * 
 * @param ptr  Pointer to the allocated memory.
 * @param size Total size of the allocated memory.
 */
#ifdef __cplusplus

#define hit_malloc(ptr,size,cast) \
{ \
	ptr = (cast) malloc(size); \
	if(ptr == NULL) hit_errInternal(__func__,"Memory allocation request failed","",__FILE__,__LINE__) \
}

#else

#define hit_malloc(ptr,size,cast) \
{ \
	ptr = malloc(size); \
	if(ptr == NULL) hit_errInternal(__func__,"Memory allocation request failed","",__FILE__,__LINE__) \
}

#endif




/**
 * Frees a previous allocated memory.
 * @fn void hit_free(void *ptr)
 * @version 2.0
 * @author Javier Fresno Bausela, Arturo Gonzalez-Escribano
 * 
 * @param ptr Pointer to the allocated memory.
 */
#define hit_free(ptr) { free(ptr); ptr = NULL; }


/**
 * Changes the size of an allocated memory.
 * @fn void hit_realloc(void *ptr, size_t size)
 * @version 1.0
 * @author Javier Fresno Bausela
 * 
 * @param ptr  Pointer to the allocated memory.
 * @param size New size for the memory.
 */
#ifdef __cplusplus

#define hit_realloc(ptr,size, cast) \
{ \
	ptr = (cast) realloc(ptr, size); \
	if(ptr == NULL) hit_errInternal(__func__,"Memory reallocation request failed","",__FILE__,__LINE__) \
}

#else

#define hit_realloc(ptr,size, cast) \
{ \
	ptr = realloc(ptr, size); \
	if(ptr == NULL) hit_errInternal(__func__,"Memory reallocation request failed","",__FILE__,__LINE__) \
}

#endif






/* END OF HEADER FILE _HitAllocP_ */
#endif
