/**
 * Hierarchical tiling of multi-dimensional dense array.
 * Definitions and functions to manipulate HitTile types. 
 * This type defines a handler for dense arrays and sparse structures
 * with meta-information to create and manipulate hierarchical tiles. 
 *
 * @file hit_blockTile.h
 * @version 1.1
 * @author Arturo Gonzalez-Escribano
 * @date Feb 2012
 */

/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2012, Arturo Gonzalez Escribano
 * All Rights Reserved.
 */

#ifndef _HitBlockTile_
#define _HitBlockTile_

#include "hit_error.h"
#include "hit_shape.h"
#include "hit_tile.h"
#include "hit_dump.h"



// SHORTCUT FOR THE HIERARCHICAL TYPE
#define HitBlockTile	HitTile_HitTile

/* A. CONSTRUCTOR */
void hit_blockTileNew( HitBlockTile *container, void *originalVar, int blockSizes[ HIT_MAXDIMS ] );

/* B. ALLOCATION FUNCTIONS */
#define hit_blockTileAlloc( varP, baseType ) hit_blockTileAllocInternal( varP, sizeof(baseType), NULL )
#define hit_blockTileAllocFill( varP, baseType, fillValue ) hit_blockTileAllocInternal( varP, sizeof(baseType), fillValue )
void hit_blockTileAllocInternal( HitBlockTile *container, size_t baseSizeExtent, void *fillValue );

/* C. FREE */
void hit_blockTileFree( HitBlockTile container );

/* D. SELECTION SHORTCUTS */
#define hit_blockTileSelect( ... ) 	hit_tileSelect( __VA_ARGS__ );
#define hit_blockTileSelectNoBoundary( ... ) 	hit_tileSelectNoBoundary( __VA_ARGS__ );
#define hit_blockTileSelectArrayCoords( ... ) 	hit_tileSelectArrayCoords( __VA_ARGS__ );
#define hit_blockTileSelectArrayCoordsNoBoundary( ... ) 	hit_tileSelectArrayCoordsNoBoundary( __VA_ARGS__ );

/* E. DATA ACCESSING */
#define hit_blockTileUnpadElem( elem, baseType )	(((elem).unpadded != NULL) ? *(HitTile_##baseType *)((elem).unpadded) : *(HitTile_##baseType *)&(elem))

#define hit_blockTileElemAt( tile, baseType, ... )	hit_blockTileUnpadElem( hit_tileElemAt(tile, __VA_ARGS__ ), baseType )
#define hit_blockTileElemAtNS( tile, baseType, ... )	hit_blockTileUnpadElem( hit_tileElemAtNoStride(tile, __VA_ARGS__ ), baseType )

#define hit_blockTileElemAtArrayCoords( tile, baseType, ... )	hit_blockTileUnpadElem( hit_tileElemAtArrayCoords(tile, __VA_ARGS__ ), baseType )

#define hit_blockTileBufferAt( tile, baseType, ... )	( *(HitTile_##baseType *) & hit_tileElemAt(tile, __VA_ARGS__ ) )
#define hit_blockTileBufferAtNS( tile, baseType, ... )	( *(HitTile_##baseType *) & hit_tileElemAtNoStride(tile, __VA_ARGS__ ) )

#define hit_blockTileBufferAtArrayCoords( tile, baseType, ... )	( *(HitTile_##baseType *) & hit_tileElemAtArrayCoords(tile, __VA_ARGS__ ) )

/* F. UTILITIES */

// GET THE BLOCKING SIZES
#define hit_blockTileDimBlockSize( tile, dim )	( tile.childSize[ dim ] )

// FIND THE mbTile ANCESTOR
static inline HitBlockTile * hit_blockTileRoot( HitBlockTile * tile ) {
	while ( tile->ref != NULL && tile->ref->hierDepth > 0 ) tile = tile->ref;
	return tile;
}

/* G. DEBUG */
static inline void hit_blockTileDumpTiles( HitBlockTile container, const char *fileName ) {
	hit_dumpTileFile( container, fileName, "--- Level 1 ---" );		

	int k,l;
	hit_tileForDimDomain( container, 0, k ) {
		hit_tileForDimDomain( container, 1, l ) {
			HitTile third;
			third = hit_tileElemAt( container, 2, k, l );
			hit_dumpTileFile( third, fileName, "--- Level 2 ---" );	
			if ( third.unpadded != NULL ) {
				hit_dumpTileFile( *(third.unpadded), fileName, "--- Unpadded selection: " ); 
			}
		}
	}
}


/* END OF HEADER FILE _HitBlockTile_ */
#endif
