/**
 * Simple extra functions and operators.
 * Extra functions and operators to transform indexes and shape boundaries.
 * 
 * @file hit_funcop.h
 * @version 0.9.2
 * @author Arturo Gonzalez-Escribano
 * @author Javier Fresno Bausela
 * @author Carlos de Blas Carton
 * @date Jun 2010 
*/

/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2007-2011, Arturo Gonzalez Escribano
 *                          Javier Fresno Bausela
 *                          Carlos de Blas Carton
 *                          Yuri Torres de la Sierra
 * All Rights Reserved.
 */


#ifndef _HitFuncOp_
#define _HitFuncOp_

/* 1. INTEGER MAX,MIN FUNCTIONS */
/**
 * Use macro versions of operators
 */
#define HIT_USE_MACRO_OP

#ifdef HIT_USE_MACRO_OP
/**
 * Maximum of two numbers (macro version).
 * @param a Number of any type
 * @param b Number of any type
 */
#define max(a,b)	(((a)>(b)) ? (a) : (b))
/**
 * Minimum of two numbers (macro version).
 * @param a Number of any type
 * @param b Number of any type
 */
#define min(a,b)	(((a)<(b)) ? (a) : (b))
#else
/**
 * Maximum of two numbers (inlined function version).
 * @param a Number of any type
 * @param b Number of any type
 */
static inline int max(int a, int b) { return (a>b) ? a:b; }


/**
 * Minimum of two numbers (inlined function version).
 * @param a Number of any type
 * @param b Number of any type
 */
static inline int min(int a, int b) { return (a>b) ? b:a; }
#endif

/* 2. WRAP FUNCTION, GENERAL AND SPECIFIC VERSIONS FOR POSITIVE/NEGATIVE */
/**
 * Wrap index around a [0:card] domain.
 *
 * To transform index in a circular domain to the real [0:card] domain.
 * @param exp Expression that returns an index number
 * @param card Maximum cardinality of the domain.
 * @see hit_wrapPos
 * @see hit_wrapNeg
 */
#define	hit_wrap(exp,card)			(((exp)%(card)>=0) ? (exp)%(card) : (exp)%(card)+(card))
/**
 * Wrap index around a [0:card] domain for a positive index.
 *
 * The same as hit_wrap, but it only works for positive indexes. 
 * Use it for additions of positive integer expressions.
 * It is a shorter 
 * expression, cleaner, with less operations, and easier to be further optimized 
 * in later stages.
 * @param exp Expression that returns an index number
 * @param card Maximum cardinality of the domain.
 * @see hit_wrap
 * @see hit_wrapNeg
 */
#define	hit_wrapPos(exp,card)		((exp)%(card))
/**
 * Wrap index around a [0:card] domain for a negative index.
 *
 * The same as hit_wrap, but it only works for indexes inside the domain, or negative. 
 * Use it for subtraction of positive integer expressions.
 * It is a shorter 
 * expression, cleaner, with less operations, and easier to be further optimized 
 * in later stages.
 * @param exp Expression that returns an index number
 * @param card Maximum cardinality of the domain.
 * @see hit_wrap
 * @see hit_wrapPos
 */
#define	hit_wrapNeg(exp,card)		(((exp)%(card)==0) ? 0 : (exp)%(card)+(card))


/* END OF HEADER FILE _HitFuncOp_ */
#endif
