/**
 * Types and functions to manipulate domains.
 * The domains can be for multidimensional arrays or sparse domains.
 * Definition of the HitShape types for multidimensional
 * domain shapes. Constructors and functions to manipulate them.
 *
 * @file hit_shape.h
 * @version 1.1
 * @author Arturo Gonzalez-Escribano
 * @author Javier Fresno Bausela
 * @author Carlos de Blas Carton
 * @date Mar 2011
 */

/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2007-2011, Arturo Gonzalez Escribano
 *                          Javier Fresno Bausela
 *                          Carlos de Blas Carton
 *                          Yuri Torres de la Sierra
 * All Rights Reserved.
 */

#ifndef _HitShape_
#define _HitShape_

#include <stdarg.h>

#include "hit_error.h"
#include "hit_sig.h"


#ifndef INCLUDE_METIS__H
#define INCLUDE_METIS__H
#include "../metis/Lib/metis.h"
#endif




/* Hit SHAPES */
/* 1. DATA STRUCTURE */
/* 1.1 SigShape data structure */
/**
 * ADT for signatures of multidimensional domain.
 * A domain shape is tuple of signature triplets, one signature for each dimension.
 *
 * The cardinality of a shape is the number of all possible valid indexes combinations.
 *
 * @implements HitShape
 */
typedef struct {
	int numDims;				/**< Number of dimensions.**/
	HitSig	sig[HIT_MAXDIMS];	/**< HitSig objects to define each dimension.*/
} HitSigShape;


/* 1.2 CShape data structure */
typedef struct {
	int * names;		/**< Contains the list to translate between local and global names. */
	int nNames;			/**< Number of names. */
	int flagNames;		/**< Flag with information for the global to local translation. */
	int * invNames;		/**< Array with the global to local translation of vertices. */
	int nInvNames;		/**< Number of inv names. */
} HitNameList;


/**
 * Maximum number of dimensions in a sparse structure.
 * @note CSR only supports 2.
 * @note The sparse bitmap shape is implemented also for two dimensions.
 */
#define HIT_SPARSE_MAXDIMS 2


/**
 * ADT for sparse domains.
 * A sparse shape represent a graph or a sparse matrix in
 * CSR format (Compressed Sparse Row matrix).
 * A CSR can only represent a two dimensional space.
 *
 * @implements HitShape
 */
typedef struct{
	int cards[HIT_SPARSE_MAXDIMS];			/**< The cardinalities. */
	idxtype * xadj;							/**< Is an array with indices to adjncy to define the rows. */
	idxtype * adjncy;						/**< Contains the columns of each row element. */
	HitNameList names[HIT_SPARSE_MAXDIMS];	/**< Name list */
} HitCShape;


/* 1.3 BitmapShape data structure */

/** Bitmap datatype */
#define HIT_BITMAP_TYPE unsigned char
/** Number of bits per Bitmap datatype */
#define HIT_BITMAP_SIZE (sizeof(HIT_BITMAP_TYPE) * (size_t) 8)
/** Index of a bitmap element */
#define hit_bitmapShapeIndex(b) 	((b)/(int)HIT_BITMAP_SIZE)
/** Offset of a bitmap element */
#define hit_bitmapShapeOffset(b)	((b)%(int)HIT_BITMAP_SIZE)

/**
 * This struct represents a graph using a bitmap. It can represent a full graph
 * or a local selection of a global graph. The selections are supposed to be 
 * contiguous.
 * 
 * @implements HitShape
 */
typedef struct{
	int cards[HIT_SPARSE_MAXDIMS];			/**< The cardinalities. */
	int nz;									/**< Number of non-zero elements or edges of the graph. */
	HIT_BITMAP_TYPE * data;					/**< Array with the bitmap. */
	HitNameList names[HIT_SPARSE_MAXDIMS];	/**< Name list */
} HitBShape;





/* 1.4 Shape data structure */

/* Constants for signature and sparse shapes */
/**
 * Hit Signature Shape constant.
 */
#define HIT_SIG_SHAPE 0

/**
 * Hit CSR Sparse Shape constant.
 */
#define HIT_CSR_SHAPE 1

/**
 * Hit Bitmap Sparse Shape constant.
 */
#define HIT_BITMAP_SHAPE 2

/**
 * ADT for multidimensional signature, sparse or bitmap domains.
 *
 */
typedef struct {
	int type;						/**< Shape type. */

	/**
	 * Union with the different types of Shapes.
	 */
	union{
		HitSigShape	sig;	/**< Signature shape. */
		HitCShape	csr;	/**< Sparse shape. */
		HitBShape	bitmap;	/**< Bitmap shape. */
	} info;
} HitShape;


// CONSTANTS FOR THE INVERSE TRANSLATION LIST IN THE CSR AND BITMAP SHAPES.
/**
 * Constant to indicate that the cShape of bShape does not have a inverse
 * translation is of vertices from global to local.
 */
#define HIT_SHAPE_NAMES_NOARRAY	0
/**
 * Constant to indicate that the cShape of bShape has a inverse
 * translation is of vertices from global to local.
 */
#define HIT_SHAPE_NAMES_ARRAY	1
/**
 * Constant to indicate that the cShape of bShape does not need a inverse
 * translation is of vertices from global to local, because the vertices
 * are ordered from 0 to N.
 */
#define HIT_SHAPE_NAMES_ORDERED	2



/* 2. Hit SHAPE NULL CONSTANT */


/**
 * Null value for Name list.
 */
extern HitNameList HIT_NAMELIST_NULL;

/**
 * Null statoc value for Name list.
 */
#define HIT_NAMELIST_NULL_STATIC {NULL,0,HIT_SHAPE_NAMES_NOARRAY,NULL,0}

/**
 * Null internal value for multidimensional domain shapes.
 * The number of dimensions of this value is -1.
 */
#define	 HIT_SIG_SHAPE_INTERNAL_NULL_STATIC		{ -1, { HIT_SIG_NULL_STATIC, HIT_SIG_NULL_STATIC, HIT_SIG_NULL_STATIC, HIT_SIG_NULL_STATIC } }

/**
 * Null internal value for CSR sparse domain shapes.
 */
#define HIT_CSR_SHAPE_INTERNAL_NULL_STATIC { {0,0}, NULL, NULL, {HIT_NAMELIST_NULL_STATIC,HIT_NAMELIST_NULL_STATIC} }

/**
 *  Null internal value for bitmap sparse shapes.
 */
#define HIT_BITMAP_SHAPE_INTERNAL_NULL_STATIC { {0, 0}, 0, NULL, {HIT_NAMELIST_NULL_STATIC,HIT_NAMELIST_NULL_STATIC} }

/**
  Null value for sparse domain shapes.
 */
extern HitShape HIT_CSR_SHAPE_NULL;
/** Null static value for CSR sparse domain shapes. */
// @note @javfres C++ do not support this kind of struct initialization
// so the type is not fully initialized, this may cause troubles somewhere.
#ifdef __cplusplus
#define HIT_CSR_SHAPE_NULL_STATIC { HIT_CSR_SHAPE, }
#else
#define HIT_CSR_SHAPE_NULL_STATIC { HIT_CSR_SHAPE, {.csr = HIT_CSR_SHAPE_INTERNAL_NULL_STATIC} }
#endif

/**
 * Null value for signature shapes.
 * The number of dimensions of this value is -1.
 */
extern HitShape	HIT_SHAPE_NULL;
/** Null static value for signature shapes. */
#define HIT_SHAPE_NULL_STATIC { HIT_SIG_SHAPE, { HIT_SIG_SHAPE_INTERNAL_NULL_STATIC} }

/**
 * Null value for bitmap sparse shapes.
 */
extern HitShape HIT_BITMAP_SHAPE_NULL;
/** Null static value for bitmap sparse shapes. */
// @note @javfres C++ do not support this kind of struct initialization
// so the type is not fully initialized, this may cause troubles somewhere.
#ifdef __cplusplus
#define HIT_BITMAP_SHAPE_NULL_STATIC { HIT_BITMAP_SHAPE, }
#else
#define HIT_BITMAP_SHAPE_NULL_STATIC { HIT_BITMAP_SHAPE, { .bitmap = HIT_BITMAP_SHAPE_INTERNAL_NULL_STATIC } }
#endif



/* 3. Hit SHAPE WHOLE CONSTANT */
/**
 * Hit signature shape whole constant.
 * Used to indicate operations with the whole original shape of a tile.
 *
 * The number of dimensions of this value is 0.
 */
extern HitSigShape	HIT_SIG_SHAPE_WHOLE;
/** Hit signature shape whole static constant. */
#define	HIT_SIG_SHAPE_WHOLE_STATIC		{ 0, { HIT_SIG_NULL_STATIC, HIT_SIG_NULL_STATIC, HIT_SIG_NULL_STATIC, HIT_SIG_NULL_STATIC } }

/**
 * Hit shape whole constant.
 * Used to indicate operations with the whole original shape of a tile.
 * It encapsulates a Hit signature shape whole constant.
 */
extern HitShape	HIT_SHAPE_WHOLE;
/** Hit shape whole static constant.  */
#define	HIT_SHAPE_WHOLE_STATIC		{ HIT_SIG_SHAPE, { HIT_SIG_SHAPE_WHOLE_STATIC} }



/* 4. Hit SHAPE GENERATING FUNCTIONS */
/* @arturo Feb 2013
 * Hidden constructor, generating a shape from a internal HitSigShape structure */
static inline HitShape hit_shapeFromSigShape( HitSigShape s ) {
	HitShape a = { HIT_SIG_SHAPE, { s } };
	return a;
}


/* 4.1 Hit Signature Shape generating functions */
/**
 * Constructor of multidimensional domain shape values.
 *
 * The number of dimensions is limited by the maximum number of dimensions established
 * when compiling the library.
 * @see HIT_MAXDIMS
 * @param nd Number of dimensions.
 * @param ... A list of as many signature values (or calls to the signature constructor) 
 * 				as dimensions specified in "nd".
 * @return A new shape value.
 *
 * @def hit_shape(nd, ...)
 */
#define hit_shape(nd, ...)	hit_shape##nd(__VA_ARGS__)

/**
 * Hit Signature Shape constructor for one dimension.
 * @memberof HitSigShape
 */
static inline HitShape hit_shape1(HitSig s1) {
	HitShape a = { HIT_SIG_SHAPE, { { 1, { s1 } }} };
	return a;
}
/**
 * Hit Signature Shape constructor for two dimensions.
 * @memberof HitSigShape
 */
static inline HitShape hit_shape2(HitSig s1, HitSig s2) {
	HitShape a = { HIT_SIG_SHAPE, { { 2, { s1, s2 } }} };
	return a;
}
/**
 * Hit Signature Shape constructor for three dimensions.
 * @memberof HitSigShape
 */
static inline HitShape hit_shape3(HitSig s1, HitSig s2, HitSig s3) {
	HitShape a = { HIT_SIG_SHAPE, { { 3, { s1, s2, s3 } }} };
	return a;
}
/**
 * Hit Signature Shape constructor for four dimensions.
 * @memberof HitSigShape
 */
static inline HitShape hit_shape4(HitSig s1, HitSig s2, HitSig s3, HitSig s4) {
	HitShape a = { HIT_SIG_SHAPE, { { 4, { s1, s2, s3, s4 } }} };
	return a;
}



/**
 * Hit Signature Shape constructor for one dimension, select only one index element.
 * @memberof HitSigShape
 */
static inline HitShape hit_shapeIndex1( int idx ) {
	HitShape a = { HIT_SIG_SHAPE, { { 1, { hit_sig( idx, idx, 1 ) } }} };
	return a;
}

/**
 * Hit Signature Shape constructor for two dimensions, select only one index element per dimension.
 * @memberof HitSigShape
 */
static inline HitShape hit_shapeIndex2( int idx1, int idx2 ) {
	HitShape a = { HIT_SIG_SHAPE, { { 2, {
			hit_sig( idx1, idx1, 1 ),
			hit_sig( idx2, idx2, 1 ) 
			} }} };
	return a;
}

/**
 * Hit Signature Shape constructor for three dimensions, select only one index element per dimension.
 * @memberof HitSigShape
 */
static inline HitShape hit_shapeIndex3( int idx1, int idx2, int idx3 ) {
	HitShape a = { HIT_SIG_SHAPE, { { 3, {
			hit_sig( idx1, idx1, 1 ),
			hit_sig( idx2, idx2, 1 ),
			hit_sig( idx3, idx3, 1 ) 
			} }} };
	return a;
}

/**
 * Hit Signature Shape constructor for four dimensions, select only one index element per dimension.
 * @memberof HitSigShape
 */
static inline HitShape hit_shapeIndex4( int idx1, int idx2, int idx3, int idx4 ) {
	HitShape a = { HIT_SIG_SHAPE, { { 4, {
			hit_sig( idx1, idx1, 1 ),
			hit_sig( idx2, idx2, 1 ),
			hit_sig( idx3, idx3, 1 ),
			hit_sig( idx4, idx4, 1 ) 
			} }} };
	return a;
}


/**
 * Constructor of multidimensional domain shape values, with standard C array domains
 *
 * The number of dimensions is limited by the maximum number of dimensions established
 * when compiling the library.
 * @see HIT_MAXDIMS
 * @param nd Number of dimensions.
 * @param ... A list of as many integer values as dimensions specified in "nd".
 * @return A new shape value.
 *
 * @def hit_shapeStd(nd, ...)
 */
#define hit_shapeStd(nd, ...)	hit_shapeStd##nd(__VA_ARGS__)


/**
 * Hit Signature Shape constructor for one dimension, standard domain
 * @memberof HitSigShape
 */
static inline HitShape hit_shapeStd1( int size ) {
	HitShape a = { HIT_SIG_SHAPE, { { 1, { hit_sigStd( size ) } } } };
	return a;
}

/**
 * Hit Signature Shape constructor for two dimensions, standard domain
 * @memberof HitSigShape
 */
static inline HitShape hit_shapeStd2( int size1, int size2 ) {
	HitShape a = { HIT_SIG_SHAPE, { { 2, {
			hit_sigStd( size1 ),
			hit_sigStd( size2 ) 
			} }} };
	return a;
}

/**
 * Hit Signature Shape constructor for three dimensions, standard domain
 * @memberof HitSigShape
 */
static inline HitShape hit_shapeStd3( int size1, int size2, int size3 ) {
	HitShape a = { HIT_SIG_SHAPE, { { 3, {
			hit_sigStd( size1 ),
			hit_sigStd( size2 ),
			hit_sigStd( size3 ) 
			} }} };
	return a;
}

/**
 * Hit Signature Shape constructor for four dimensions, standard domain
 * @memberof HitSigShape
 */
static inline HitShape hit_shapeStd4( int size1, int size2, int size3, int size4 ) {
	HitShape a = { HIT_SIG_SHAPE, { { 4, {
			hit_sigStd( size1 ),
			hit_sigStd( size2 ),
			hit_sigStd( size3 ),
			hit_sigStd( size4 ) 
			} }} };
	return a;
}



/* 5. HIT SHAPE ACCESS MACROS */
/**
 * Access to the shape type.
 * @memberof HitShape
 */
#define hit_shapeType(s) ((s).type)
/**
 * Access to the sparse shape struct.
 * @memberof HitShape
 */
#define hit_cShapeAccess(s) ((s).info.csr)
/**
 * Access to the signature shape struct.
 * @memberof HitShape
 */
#define hit_sigShapeAccess(s) ((s).info.sig)

/**
 * Accesss to the bitmap shape struct
 * @param s A shape.
 * @memberof HitShape.
 * 
 * @return The Bitmap shape.
 */
#define hit_bShapeAccess(s) ((s).info.bitmap)

/**
 * Number of dimensions in a shape.
 * @memberof HitShape
 * @param shape A domain shape.
 * @return Number of dimensions of that shape.
 */
#define	hit_shapeDims(shape)	(((shape).type==HIT_SIG_SHAPE)?hit_shapeSigDims(shape):(1))



/* 5.1 Hit Signature Shape access macros */
/**
 * Number of dimensions in a shape.
 * @memberof HitSigShape
 * @param shape A signature domain shape.
 * @return Number of dimensions of that shape.
 */
#define hit_shapeSigDims(shape) (hit_sigShapeAccess(shape).numDims)

/**
 * Set the number of dimensions in a shape.
 * @memberof HitSigShape
 * @param shape A domain shape.
 * @param value The new value.
 * @return Number of dimensions of that shape.
 */
#define hit_shapeDimsSet(shape,value) (hit_shapeSigDims(shape) = (value))


/**
 * Dimension signature in a shape
 * @memberof HitSigShape
 * @param shape A domain shape.
 * @param dim A dimension,
 * @return The dimension signature of that shape.
 */
#define	hit_shapeSig(shape,dim)	((hit_sigShapeAccess((shape))).sig[(dim)])

/**
 * Cardinality of a dimension signature in a shape
 * @memberof HitSigShape
 * @param shape A domain shape.
 * @param dim A dimension,
 * @return The cardinality of the dimension signature of that shape.
 */
#define	hit_shapeSigCard(shape,dim)	hit_sigCard((hit_sigShapeAccess((shape))).sig[(dim)])


/**
 * Cardinality of a shape
 * @memberof HitSigShape
 * @param shape A domain shape.
 * @return The cardinality of that shape.
 */
static inline int hit_shapeCard( HitShape shape ) {
	int i;
	int result = 1;
	for ( i=0; i<hit_shapeDims( shape ); i++ )
		result = result * hit_shapeSigCard( shape, i );
	return result;
}




/* 6. HIT SHAPE FREE FUNTION */
/**
 * Free resources of shapes.
 * @memberof HitShape
 * @note It is not necessary for HitSigShape.
 * @param s shape.
 */
void hit_shapeFree(HitShape s);


/* 7. Hit SHAPE: COMPARISON OPERATOR */
/**
 * Compare two shape values.
 * @memberof HitShape
 * @warning It only works for HitSigShape,
 * it returns 0 with HitSparseShape.
 * @param sh1 Domain shape.
 * @param sh2 Domain shape.
 * @return Logical value, true if the two shapes are equal.
 */
int hit_shapeCmp(HitShape sh1, HitShape sh2);


/* 8. Hit INTERSECT SHAPES */
/**
 * Intersection of two shapes.
 * @memberof HitShape
 * @warning It only works for HitSigShape,
 * it returns HIT_SHAPE_NULL with HitSparseShape.
 * @param sh1 Domain shape.
 * @param sh2 Domain shape.
 * @return A new domain shape with the indexes combinations which
 * 			are valid in both shapes at the same time.
 */
HitShape hit_shapeIntersect(HitShape sh1, HitShape sh2);

/* 8.2. Hit SUBSELECT SHAPES */
/**
 * Subset or subselection of a shape according to local coordinates.
 * @memberof HitShape
 * @warning It only works for HitSigShape,
 * it returns HIT_SHAPE_NULL with HitSparseShape.
 * @param sh1 Domain shape.
 * @param sh2 Subselection shape.
 * @return A new domain shape with the indexes of the first shape that exist
 * 			in the second shape when renumbering the first shape in standard
 * 			C notation (starting at 0, stride 1).
 */
HitShape hit_shapeSubset(HitShape sh1, HitShape sh2);

/* 8.3. Hit SHAPE: TILE TO ARRAY */
/**
 * Transform tile coordinates to array coordinates
 * @memberof HitShape
 * @warning It only works for HitSigShape,
 * it returns HIT_SHAPE_NULL with HitSparseShape.
 * @param sh1 Domain shape.
 * @param sh2 Subshape in tile coordinates
 * @return A new domain shape with the indexes of the second shape transformed to
 * 			array coordinates in the reference system of the first shape.
 */
HitShape hit_shapeTileToArray(HitShape sh1, HitShape sh2);


/* 9. Hit Signature Shape operations */
/* 9.1 Hit SHAPE BORDER */
/* Values for hit_shapeBorder */
/**
 * Hit Shape Begin constant for the hit_shapeBorder function.
 * @see hit_shapeBorder
 */
#define HIT_SHAPE_BEGIN 0
/**
 * Hit Shape End constant for the hit_shapeBorder function.
 * @see hit_shapeBorder
 */
#define HIT_SHAPE_END	1

/**
 * Return another shape with the shape of a border in the given dimension.
 * @memberof HitSparseShape
 * @warning It only works for HitSigShape,
 * it returns HIT_SHAPE_NULL with HitSparseShape.
 * @param shape A HitShape.
 * @param dim The dimension.
 * @param position Begin or End.
 * @param offset A positive offset move the border outside the shape and a negative inside.
 */
HitShape hit_shapeBorder(HitShape shape, int dim, int position, int offset);

/* 9.2 Hit SHAPE EXPAND */
/**
 * Expand a shape adding a border by the given number of dimension.
 * @memberof HitSparseShape
 * @warning It only works for HitSigShape,
 * it returns HIT_SHAPE_NULL with HitSparseShape.
 * @param shape HitShape.
 * @param dims Number of dimension to expand.
 * @param border Border size.
 * @return The expanded shape.
 */
HitShape hit_shapeExpand(HitShape shape,int dims,int border);

/* 9.2 Hit SHAPE EXPAND DIM */
/**
 * Expand a shape changing the position of a border in a given dimension and direction (begin/end).
 * @memberof HitSparseShape
 * @warning It only works for HitSigShape,
 * it returns HIT_SHAPE_NULL with HitSparseShape.
 * @param shape HitShape.
 * @param dims Number of dimension to expand.
 * @param border Border size.
 * @return The expanded shape.
 */
HitShape hit_shapeDimExpand( HitShape shape, int dim, int position, int border );



/* 11. Hit Shape Iterators */

/* 11.1. Hit Sig Shape Iterators */
/**
 * hit_shapeIterator: element iterator.
 * @param var Loop variable
 * @param shape Signature shape
 * @param dim Dimension
 */
#define hit_shapeIterator(var,shape,dim) for(var=hit_shapeSig((shape),(dim)).begin;var<=hit_shapeSig((shape),(dim)).end;var++)

/**
 * hit_shapeIterator: element iterator.in tile coordinates
 * @param var Loop variable
 * @param shape Signature shape
 * @param dim Dimension
 */
#define hit_shapeIteratorTileCoord(var,shape,dim) for(var=0; var<=hit_sigCard(hit_shapeSig((shape),(dim)))-1;var++)


/* 12. Name list functionalities for sparse (csr and bitmap) shapes */
/**
 * Returns a name of the list
 * @param list The name list.
 * @param index The index.
 * @return The name of the index
 */
#define hit_nameListIndex2Name(list, index) (list.names[(index)])

/**
 * Returns the index of a name.
 * @param list The name list.
 * @param name The name.
 * @return The index.
 */
int hit_nameListName2Index(HitNameList list, int name);

/**
 * Creates a new name list.
 * @param list List pointer.
 * @param nelems Number of elements.
 */
void hit_nameListCreate(HitNameList * list, int nelems);

/**
 * Frees a name list.
 * @param list The list.
 */
void hit_nameListFree(HitNameList list);

/**
 * Adds a new element to the list
 * @param list A name list pointer.
 * @param x The name
 */
void hit_nameListAdd(HitNameList * list, int x);

/**
 * Create the inverse list name to make the translation.
 * @param list A name list pointer.
 */
void hit_nameListCreateInvNames(HitNameList * list);

/**
 * Clones a name list.
 * @param dst Destination list.
 * @param src Source list.
 */
void hit_nameListClone(HitNameList * dst, HitNameList * src);

/**
 * Vertex iterator in a bitmap shape
 * @param var Variable to iterate
 * @param shape Bitmap shape
 */
#define hit_bitmapShapeVertexIterator(var,shape) for(var=0; var<hit_bitmapShapeNvertices(shape); var++)

/* END OF HEADER FILE _HitShape_ */
#endif
