/**
* Definition of the HitSig types for domain signatures.
* Constructors and functions to manipulate them.
*
* @file hit_sig.h
* @version 1.0
* @author Arturo Gonzalez-Escribano
* @author Javier Fresno Bausela
* @author Carlos de Blas Carton
* @date Jun 2010
*/

/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2007-2011, Arturo Gonzalez Escribano
 *                          Javier Fresno Bausela
 *                          Carlos de Blas Carton
 *                          Yuri Torres de la Sierra
 * All Rights Reserved.
 */

#ifndef _HitSig_
#define _HitSig_

#include <stdarg.h>
#include "hit_error.h"

/* MAXIMUM NUMBER OF DIMENSIONS, FIXED AT COMPILATION TIME */
#ifndef HIT_MAXDIMS
/**
 * Maximum number of dimensions supported by the HitMap library.
 * Determined at compile time. This is the default value. It may be changed by the
 * programmer at any other place before compiling the library.
 */
#define HIT_MAXDIMS	4
#endif



/* Hit SIGNATURES */
/* 1. DATA STRUCTURE */
/**
 * ADT for signatures of a dimension domain.
 * A dimension domain signature is a triplet of three integer values [begin:end:stride]. 
 *
 * They identify the first and last valid indexes
 * in that domain (begin, end), and also a stride value which identifies the valid indexes in 
 * that domain, selected with a regular stride from the begin. An end value lesser than begin
 * indicates and empty domain.
 *
 * The cardinality of a signature is the number of valid indexes in that domain. 
 *
 * A signature defines a subdomain of the whole indexes domain, Z. The tile coordinates defined
 * by the signature are in the range [0:cardinality]. Each index expressed in the tile 
 * coordinate system may be translated to the corresponding array coordinate system.
 */
typedef struct {
	int begin;	/**< The begin index of the dimension */
	int end;	/**< The end index of the dimension */
	int stride; /**< The stride for regular sparse domains */
} HitSig;

/* 2.1. Hit SIGNATURE NULL CONSTANT */
/**
 * Null value for domain signatures.
 * The cardinality of this domain value is 0 elements. It starts at index -1, ends at -2,
 * and has stride 1.
 */
extern HitSig	HIT_SIG_NULL;
/**
 * Static Null value for domain signatures.
 */
#define	HIT_SIG_NULL_STATIC	{ -1, -2, 1 }

/* 2.2. Hit SIGNATURE WHOLE CONSTANT */
/**
 * Whole value for domain signatures. Used to indicate selections of whole dimensions
 * The cardinality of this domain value is 0 elements.
 */
extern HitSig	HIT_SIG_WHOLE;
/**
 * Static Whole value for domain signatures.
 */
#define	HIT_SIG_WHOLE_STATIC	{ -2, -3, 1 }

/* 3. Hit SIGNATURE GENERATING FUNCTION */
/**
 * Constructor (1) of domain signature values.
 * @memberof HitSig
 *
 * @param begin First valid index in the domain
 * @param end Last valid index in the domain
 * @param stride Stride for the selected valid indexes in the domain, starting at begin.
 * @return A new signature value.
 */
static inline HitSig hit_sig(int begin, int end, int stride) {
	HitSig a = { begin, end, stride };
	return a;
}

/**
 * Constructor (2) of domain signature values.
 * @memberof HitSig
 *
 * @param size Number of elements in the domain
 * @return A new signature value with the range [0,size-1] and no stride.
 */
static inline HitSig hit_sigStd(int size) {
	HitSig a = { 0, size-1, 1 };
	return a;
}

/**
 * Constructor (3) of domain signature values.
 * @memberof HitSig
 *
 * @param ind The integer index to select
 * @return A new signature value with a single index in its range [index,index].
 */
static inline HitSig hit_sigIndex(int ind) {
	HitSig a = { ind, ind, 1 };
	return a;
}


/* 4. Hit SIGNATURE BLENDING */
/**
 * Selection of a subdomain.
 * It computes the result of applying the second signature selection on the first domain
 * signature with tile coordinates (begin of the first signature is translated to 0).
 * @memberof HitSig
 * @param s1 Reference domain signature.
 * @param s2 Subselection signature.
 * @author Yo
 * @return The signature of the subselection in array coordinates.
 */
static inline HitSig hit_sigBlend(HitSig s1, HitSig s2) {
	HitSig newSig;
	newSig.begin = s1.begin + s2.begin * s1.stride;
	newSig.end = s1.begin + s2.end * s1.stride;
	newSig.stride = s1.stride * s2.stride;
	return newSig;
}

/* To expand the tile cardinalities cardinalities through two different signatures*/
static inline HitSig hit_sigExpand(HitSig s1, HitSig s2) {					//Yuri 
	HitSig newSig;
	newSig.begin = s1.begin + s2.begin;
	newSig.end = s1.end + s2.end;
	newSig.stride = s1.stride * s2.stride;
	return newSig;
}


/* 5. Hit SIGNATURE CARDINALITY */
/**
 * Compute the cardinality of a domain signature.
 * @fn int hit_sigCard(HitSig sig)

 * @memberof HitSig
 * @param sig Domain signature.
 * @return An integer with the number of valid indexes in that domain.
 */
#define	hit_sigCard(sig)	(((sig).end-(sig).begin)/(sig).stride+1)

/* 6. Hit APPLY A ONE-ELEMENT SELECTION THROUGH A SIGNATURE */
/**
 * Translate an index in tile coordinate system of a domain signature to array coordinates.
 * @fn int hit_sigTileToArray(HitSig sig, int i)
 * @memberof HitSig
 * @param sig Domain signature.
 * @param i Index value.
 * @return Index value in array coordinate system.
 */
#define	hit_sigTileToArray(sig,i)	((i)*(sig).stride+(sig).begin)

/**
 * Translate an index in array coordinate system of a domain signature to Tile coordinates.
 * @fn int hit_sigArrayToTile(HitSig sig, int i)
 * @memberof HitSig
 * @param sig Domain signature.
 * @param i Index value.
 * @return Index value in Tile coordinate system.
 */
#define	hit_sigArrayToTile(sig,i)	(   (int)(((i)-(sig).begin)/(sig).stride)   ) 

/* 7. Hit CONDITION: NUMBER IN THE DOMAIN OF A SIGNATURE */
/**
 * Determine if an index is a valid index in a domain.
 * @memberof HitSig
 * @param sig Domain signature.
 * @param a Index value.
 * @return Logical value, true if "a" is a valid index in the domain.
 */
#define hit_sigIn(sig,a)	(a>=(sig).begin && a<=(sig).end && ((a-(sig).begin)%(sig).stride == 0))

/* 8. Hit SIGNATURE: COMPARISON OPERATOR */
/**
 * Compare two signature values.
 * @memberof HitSig
 * @param a Domain signature.
 * @param b Domain signature.
 * @return Logical value, true if the two signatures are equal.
 */
#define	hit_sigCmp(a,b)	((a).begin == (b).begin && (a).end == (b).end && (a).stride == (b).stride)

/* 9. Hit SIGNATURE: INTERSECTION OPERATOR */
/**
 * Intersect two signature values.
 * @memberof HitSig
 * @param a Domain signature.
 * @param b Domain signature.
 * @return The signature representing the intersection. HIT_SIG_NULL if there is no intersection.
 */
HitSig hit_sigIntersect( HitSig s1, HitSig s2 );



/* END OF HEADER FILE _HitSig_ */
#endif
