

#ifndef _HitTaskContainers_
#define _HitTaskContainers_

#include <pthread.h>

#include <hit_task.h>
#include <hit_allocP.h>
#include <hit_mpi.h>


#define HIT_TASK_NCHANNELS_MAX 10


/**
 * A element of the task list.
 */
typedef struct HitTaskListElem {
	struct HitTaskListElem	*prev;	/**< Previous element */
	struct HitTaskListElem	*next;	/**< Next element */
	/** Array of tasks, one for each channel */
	HitTask	task[HIT_TASK_NCHANNELS_MAX];
} HitTaskListElem;


/**
 * A thread-safe list of tasks.
 */
typedef struct {
	int					size;				/**< Number of task in the list. */
	double				weight;				/**< Total weight of the tasks */
	int					exhausted_input;	/**< Flag to indicate that there is no more input tasks. */
	int					exhausted_ws;		/**< Flag to indicate that there is no more work stealing tasks. */
	pthread_mutex_t		mutex;				/**< Mutex to ensure thread mutual exclusion. */
	pthread_cond_t		cond;				/**< Condition to wake up threads waiting for tasks. */
	pthread_cond_t		cond_ws;			/**< Condition to wake up the ws thread to request more tasks. */
	HitTaskListElem 	*first;				/**< First element */
	HitTaskListElem 	*last;				/**< Last element */
	int					wakeup_ws;			/**< @new @javfres */
	int					n_channels;			/**< Number of input channels of the worker using this list */
} HitTaskList;


/**
 * Null constant for the task list.
 */
extern HitTaskList HIT_TASK_LIST_NULL;
#define HIT_TASK_LIST_NULL_STATIC {0,0,0,1,PTHREAD_MUTEX_INITIALIZER,PTHREAD_COND_INITIALIZER,PTHREAD_COND_INITIALIZER,NULL,NULL,0,1}



/**
 * Creates a new Task list with n channels
 * @param list Pointer to the struct
 * @param n_channels Number of channels
 */
void hit_taskListCreate(HitTaskList * list, int n_channels);

/**
 * Remove the list and deallocate the resources.
 * @param list A task list pointer.
 */
void hit_taskListFree(HitTaskList * list);


/**
 * Adds a new task (or several tasks) at the tail.
 * @param list The pointer to the list.
 * @param taskP The pointer to the task.
 */
void hit_taskListAddLast(HitTaskList * list, void * taskPArray);

/**
 * Removes an element (or several tasks) from the tail.
 * @param list The pointer to the list.
 * @param taskP The pointer to the task.
 * @return 0 if the task is null.
 */
int hit_taskListRemoveLast(HitTaskList * list, void * taskPArray);

/**
 * Adds a new task (or several tasks) at the head.
 * @param list The pointer to the list.
 * @param taskP The pointer to the task.

 */
void hit_taskListAddFirst(HitTaskList * list, void * taskPArray);

/**
 * Removes an element (or several tasks) from the head.
 * @param list The pointer to the list.
 * @param taskP The pointer to the task.
 * @return 0 if the task is null.
 */
int hit_taskListRemoveFirst(HitTaskList * list, void * taskPArray);


/**
 * This function waits until a tasks was been removed from the list.
 * This is used by the ws thread to ask for more tasks.
 * @param list The list pointer.
 * @return It returns 1 if if it could be called again. It return 0 if
 * there is no need to wait.
 */
int hit_taskListWaitRemove(HitTaskList * list);



/**
 * Return the size of the list.
 * @param list The pointer to the list.
 * @return Number of tasks in the list.
 */
int hit_taskListSize(HitTaskList * list);


/**
 * Return the weight of the list.
 * @param list The pointer to the list.
 * @return Total weight of the tasks in the list.
 */
double hit_taskListWeight(HitTaskList * list);


/**
 * Set the input stream task as exhausted.
 * @param list The pointer to the list.
 */
void hit_taskListSetExhausted_input(HitTaskList * list);

/**
 * Set the work stealing stream task as exhausted.
 * @param list The pointer to the list.
 */
void hit_taskListSetExhausted_ws(HitTaskList * list);

/**
 * Wakes up the work stealing thread waiting for a task removal.
 * @param list The list pointer.
 */
void hit_taskListWakeUp(HitTaskList * list);


// ----------------------------------------------------------- //


/**
 * Struct used to store the task info in the buffer.
 */
typedef struct {
	int n_elems;
	double weight;
} HitBufferInfo;

/**
 * HitBufferInfo mpi type variable
 */
extern MPI_Datatype HitBufferInfoType;

/** Flag for null task */
#define HIT_BUFFER_NULLTASK (0)

/** Flag for EOF */
#define HIT_BUFFER_EOF (-1)

/** Flag for WS request */
#define HIT_BUFFER_REQUEST (-2)


/**
 * Task buffer.
 */
typedef struct {

	int				n_types;	/**< Number of types of task in the buffer (> 1 for some WS buffers). */
	MPI_Datatype	type[HIT_TASK_NCHANNELS_MAX]; /**< Array with the task buffers */
	int				size[HIT_TASK_NCHANNELS_MAX]; /**< Array with the C size of the task to allocate them */

	void*			buffer; /**< Real buffer data */
	int				buffer_size; /**< Buffer size in bytes */
	int				position; /**< Position of the next element in the buffer */
	int				buffer_received; /**< Number of bytes received */

	MPI_Comm		comm; /**< MPI Communicator */
	int				tag; /**< Tag for this buffer */

	int				num_elem; /**< Max number of elements (Tasks can have several elements (Task arrays) */
	int				num_used; /**< Number of elements used (for Output buffers) */

	int				empty; /**< Empty flag (for Input buffers) */
	int				eof; /**< EOF flag */
	int 			request; /**< Request flag */

	double			weight; /**< Weight of the buffer to send it before its full (for Output buffer) */

	int				pending_recv; /**< Flag that indicates that there is an irecv operation */

	int				null_buffer; /**< Buffer that only returns null tasks */


	//MPI_Request		request;	/**< MPI Request for non blocking communications: MPI_REQUEST_NULL */
} HitTaskBuffer;


/**
 * Task buffer null constant.
 */
#define HIT_TASK_BUFFER_NULL_STATIC {	\
	0,{MPI_DATATYPE_NULL,},{0,},		\
	NULL,0,0,0,							\
	MPI_COMM_NULL,0,					\
	0,0,								\
	1,0,0,								\
	0,									\
	0,									\
	0									\
}


/**
 * Task buffer null variable.
 */
extern HitTaskBuffer HIT_TASK_BUFFER_NULL;

/**
 * Task buffer constructor
 * @param bufferP Pointer to the buffer
 * @param num_elem Number of elements in the buffer.
 * @param comm Communicator
 * @param tag Input tag
 * @param n_types Number of task types
 * @param ... List of hit data types.
 */
void hit_taskBufferCreate(HitTaskBuffer * bufferP, int num_elem, MPI_Comm comm, int tag, int n_types, ...);

/**
 * Frees the buffer memory.
 * @param bufferP Pointer to the buffer.
 */
void hit_taskBufferFree(HitTaskBuffer * bufferP);

/**
 * Adds a task (or several) to the buffer.
 * @param bufferP Pointer buffer.
 * @param taskArray Task array.
 */
void hit_taskBufferAdd(HitTaskBuffer * bufferP, void * taskArray);

/**
 * Adds a special task to the buffer (EOF, REQUEST).
 * @param bufferP Pointer buffer.
 * @param flag (HIT_BUFFER_EOF/HIT_BUFFER_REQUEST)
 */
void hit_taskBufferAddSpecial(HitTaskBuffer * bufferP, int flag);

/**
 * Send the content of the buffer.
 * @param bufferP Pointer to the buffer.
 * @param dest Rank of destination
 * @param tag Message tag
 */
void hit_taskBufferSend(HitTaskBuffer * bufferP, int dest, int tag);


/**
 * Recv task and stores them into the buffer.
 * @param bufferP Pointer to the buffer.
 * @param source Rank of source
 * @param comm Communicator
 */
void hit_taskBufferRecv(HitTaskBuffer * bufferP, int * source);


/**
 * Recv tasks and stores them into the buffer (non-blocking version).
 * @param bufferP Pointer to the buffer.
 * @param source Rank of source
 */
void hit_taskBufferIRecv(HitTaskBuffer * bufferP, MPI_Request * request);

/**
 * Finalizes a IRecv operation
 * @param bufferP Pointer to the buffer.
 * @param source Rank of source
 */
void hit_taskBufferIRecvEnd(HitTaskBuffer * bufferP, int * source, MPI_Status status);


/**
 * Creates a new task using the i buffer element.
 * @param bufferP Pointer to the buffer.
 * @param taskArray Pointer to the task.
 */
void hit_taskBufferCreateTask(HitTaskBuffer * bufferP, void * taskArray);






// @todo
#define hit_taskBufferWeight(bufferP) ((bufferP)->weight)




/**
 * Checks if the buffer is empty.
 * @param bufferP Pointer to the buffer.
 */
#define hit_taskBufferEmpty(bufferP) ((bufferP)->empty == 1)

/**
 * Checks if the buffer is waiting for a Irecv operation
 * @param bufferP Pointer to the buffer.
 */
#define hit_taskBufferPendingRecv(bufferP) ((bufferP)->pending_recv == 1)

/**
 * Checks if the buffer has rececived a EOF.
 * @param bufferP Pointer to the buffer.
 */
#define hit_taskBufferEOF(bufferP) ((bufferP)->eof == 1)

/**
 * Checks if the buffer has rececived a WS request.
 * @param bufferP Pointer to the buffer.
 */
#define hit_taskBufferRequest(bufferP) ((bufferP)->request == 1)







#endif

