/**
 * Hierarchical tiling of multi-dimensional dense array.
 * Definitions and functions to manipulate HitTile types. 
 * This type defines a handler for dense arrays and sparse structures
 * with meta-information to create and manipulate hierarchical tiles. 
 *
 * @file hit_tile.h
 * @version 1.6
 * @author Arturo Gonzalez-Escribano
 * @author Javier Fresno Bausela
 * @author Carlos de Blas Carton
 * @author Yuri Torres de la Sierra
 * @date Mar 2013
 */

/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2007-2013, Arturo Gonzalez Escribano
 *                          Javier Fresno Bausela
 *                          Carlos de Blas Carton
 *                          Yuri Torres de la Sierra
 * All Rights Reserved.
 */



#ifndef _HitTile_
#define _HitTile_

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "hit_error.h"
#include "hit_shape.h"
#include "hit_layout.h"
#include "hit_com.h"

/* Hit VARIABLES */
/**
 * @class HitTileInterface
 * Virtual interface for all the tile classes.
 */

/**
 * 
 * @class HitGCTile
 * @implements HitTileInterface
 * HitGCTile
 */

/**
 * @class HitGBTile
 * @implements HitTileInterface
 * HitGBTile
 */

/**
 * @class HitSCTile
 * @implements HitTileInterface
 * HitSCTile
 */

/**
 * @class HitSBTile
 * @implements HitTileInterface
 * HitSBTile
 */


/**
 * @class HitCTile
 * @implements HitTile
 * HitCTile
 */

/**
 * @class HitMTile
 * @implements HitTileInterface
 * HitMTile
 */

/* 0. TYPE CONSTANTS */
#define HIT_TILE	0
#define HIT_C_TILE	1
#define HIT_M_TILE	2
#define HIT_GC_TILE	3
#define HIT_GB_TILE	4
#define HIT_MC_TILE	5
#define HIT_MB_TILE	6




/* 1. ABSTRACT DATA-STRUCTURE DEFINITION */

/* 1.a. GENERIC ABSTRACT TYPE */
/**
 * Generic abstract data type for HitTile handler
 * @implements HitTileInterface
 */
typedef struct HitTile {
	int				type;							/**< Type of the HitTile: Tile, cTile, mTile, gcTile, gbTile, mcTile, mbTile. */
	size_t			baseExtent;						/**< Size of the variable type. */
	HitShape		shape;							/**< Shape of the Tile. */
	int				card[HIT_MAXDIMS];				/**< Dimension cardinalities. */
	int				acumCard;						/**< Flattened cardinality. */
	int				origAcumCard[HIT_MAXDIMS+1];	/**< Dimension accumulated cardinalities. */
	int				qstride[HIT_MAXDIMS];			/**< Memory stride for stride subselections */
	void			*data;							/**< Data pointer. It may be different to memPtr if it is a subselection tile. */
	void			*dataVertices;					/**< Data pointer to store the vertices in a graph tile. */
	void			*memPtr;						/**< Memory pointer. It keeps the original pointer. */
	void			*memPtrVertices;				/**< Memory pointer. It keeps the original pointer. */
	struct		 	HitTile	*ref;					/**< Pointer to the parent HitTile for hierarchical tiles. */
	char			memStatus;						/**< Memory status. */
/* ------------------------ MULTILEVEL TILES ---------------------------------------- */
	char			hierDepth;						/**< Hierarchical depth status. */
	struct		 	HitTile	*ancestor;				/**< Pointer to the ancestor in a multilevel hierarchy. */
	struct		 	HitTile	*unpadded;				/**< Pointer to a selection without padding. */
	int				childBegin[HIT_MAXDIMS];		/**< Starting array coordinate of first child in regular hierarchical structures, where all children have the same type and shape. */
	int				childSize[HIT_MAXDIMS];			/**< Normalized sizes of children in regular hierarchical structures, where all children have the same type and shape. */
	HitLayout Layout;
} HitTile, *HitPTile;

/* 1.b. MACRO TO GENERATE PARTICULAR POLYMORPHIC TYPES */
/**
 * hitTilNewType: Macro to generate particular polymorphic types of HitTile.
 *
 * @param newType Base type.
 *
 * Currently, only some C base types are supported by the communication library.
 * @see hit_mpi.h
 *
 * NOTE: This macro works properly with GNU gcc compiler or GNU cpp preprocessor. Other compilers
 * may have troubles to properly parse the ## concatenation operator.
 */
#define 	hit_tileNewType(newType) 					\
typedef struct HitTile_##newType {						\
	int				type;								\
	size_t			baseExtent;							\
	HitShape		shape;								\
	int				card[HIT_MAXDIMS];					\
	int				acumCard;							\
	int				origAcumCard[HIT_MAXDIMS+1];		\
	int				qstride[HIT_MAXDIMS];				\
	newType			*data;								\
	newType			*dataVertices;						\
	newType			*memPtr;							\
	newType			*memPtrVertices;					\
	struct 			HitTile_##newType			*ref;	\
	char			memStatus;							\
	char			hierDepth;							\
	struct HitTile	*ancestor;							\
	struct HitTile	*unpadded;							\
	int				childBegin[HIT_MAXDIMS];			\
	int				childSize[HIT_MAXDIMS];				\
        HitLayout Layout;								\
} HitTile_##newType, *HitPTile_##newType

/** Declare the type for the hierarchical Tiles */
hit_tileNewType(HitTile);

/* 2. CONSTANTS FOR MEMORY STATUS FIELD */
/**
 * Memory Status Constant
 */
#define	HIT_MS_NULL			0	/**< Reserved for HIT_TILE_NULL. */
#define	HIT_MS_NOMEM		1	/**< Root array with no memory allocated. */
#define	HIT_MS_OWNER		2	/**< Allocated array or tile. */
#define	HIT_MS_NOT_OWNER	3	/**< Tile selection pointing to the memory of another tile. */

/* 2b.1. CONSTANTS FOR HIERARCHICAL DEPTH FIELD */
/**
 * Constants to indicate Multilevel Hierarchical Tile property
 */
#define HIT_NONHIERARCHICAL		0	/**< Base elements are of native or programmer defined types. */
//#define HIT_HIERARCHICAL		1	/**< Base elements are of HitTile_x subtypes. */

/* 2b.2. CONSTANTS FOR MULTILEVEL CARDINALITIES */
#define HIT_MULTILEVEL_UNDEFINED_SIZES	{ 1, 1, 1, 1 }
#define HIT_MULTILEVEL_UNDEFINED_BEGINS	{ 0, 0, 0, 0 }

/* 2c. CONSTANTS FOR READING/WRITING TILES IN FILES */
/* 2.c.1. FILE FORMAT: BINARY vs. TEXT */
#define HIT_FILE_BINARY		0
#define HIT_FILE_TEXT		1

/* 2.c.2. MODE OF FILE OPERATIONS */
#define HIT_FILE_READ		0
#define HIT_FILE_WRITE		1

/* 2.c.3. TEXT FILES: ELEMENTS TYPE */
#define HIT_FILE_UNKNOWN	-1
#define HIT_FILE_INT		0
#define HIT_FILE_LONG		1
#define HIT_FILE_FLOAT		2
#define HIT_FILE_DOUBLE		3

/* 2.c.4. MODES FOR POSITION OF THE TILES IN THE FILE: COORDINATE SYSTEM */
#define HIT_FILE_TILE		0
#define HIT_FILE_ARRAY		1


/* 2d. CONSTANTS FOR MODE IN DATA UPDATE FROM/TO PARENT TILE */
#define HIT_UPDATE_FROM		0
#define HIT_UPDATE_TO		1


/* 3. NULL VARIABLES */
/**
 * Null value for HitTile derived types. 
 */
extern HitTile			HIT_TILE_NULL;
extern HitPTile			HIT_TILE_NULL_POINTER;
//#define	HIT_TILE_NULL_STATIC	{ 0, HIT_SHAPE_NULL_STATIC, { 0,0,0,0 }, 0, { 0,0,0,0,0 }, { 0,0,0,0 }, NULL, NULL, NULL, HIT_MS_NULL, HIT_NONHIERARCHICAL, HIT_SHAPE_NULL }
#define	HIT_TILE_NULL_STATIC	{ 0, 0, HIT_SHAPE_NULL_STATIC, { 0,0,0,0 }, 0, { 0,0,0,0,0 }, { 0,0,0,0 }, NULL, NULL, NULL, NULL, NULL, HIT_MS_NULL, HIT_NONHIERARCHICAL, NULL, NULL, HIT_MULTILEVEL_UNDEFINED_BEGINS, HIT_MULTILEVEL_UNDEFINED_SIZES }

/**
 * hit_tileIsNull: check if a variable is the Null value.
 * @param var A HitTile variable
 * @result int Boolean with the result
 * @see HIT_TILE_NULL
 */
#define hit_tileIsNull(var)	((var).memStatus==HIT_MS_NULL)


/* 4. VARIABLE MANAGEMENT */					//yuri
/* 4.a. DEBUG: DUMP VARIABLE STRUCTURE */
//#define hit_tileDumpTileFile(var,prefix,comment)	{ char name[16]; sprintf(name,"%s.%d",prefix,hit_Rank); FILE *fich; fich=fopen(name, "a"); fprintf(fich, comment); hit_tileDumpTileInternal(&(var), #var, fich); fclose(fich); }
//#define hit_tileDumpTile(var)	hit_tileDumpTileInternal(&(var), #var, stderr)
//void hit_tileDumpTileInternal(const void *var, const char * name, FILE *file);

/* 4.a. CREATE A TILE OVER A SINGLE VARIABLE OF ANY TYPE */
/* @arturo, Nov 2012 */
#define	hit_tileSingle( var, type )	hit_tileSingleInternal( &var, sizeof( type ) )
HitTile	hit_tileSingleInternal( void *var, size_t size );

/* 4.b. DECLARING VARIABLES WITHOUT MEMORY YET: ARRAY */
/**
 * hit_tileDomain: declare the shape and base type of a new multidimensional array.
 * @param newVarP	A pointer to a HitTile derived type variable.
 * @param baseExtent 	Size of the base type for the array elements.
 * @param numDims 	Number of dimensions of the array. Limited by HIT_MAXDIMS.
 * @param ... 	A list of integer positive numbers defining the cardinalities of each dimension.
 * @see hit_tileAlloc
 */
#define hit_tileDomain( newVarP, baseType, numDims, ...)	hit_tileDomainInternal( newVarP, sizeof(baseType), 0==strncmp("HitTile",#baseType,7), numDims, __VA_ARGS__)
void hit_tileDomainInternal(void *newVarP, size_t baseExtent, int hierDepth, int numDims, ...);

/**
 * hit_tileDomainShape: declare the shape and base type of a new multidimensional array.
 * @param newVarP	A pointer to a HitTile derived type variable.
 * @param baseExtent 	Size of the base type for the array elements.
 * @param shape		Shape of the new array.
 * @see hit_tileAlloc
 */
#define hit_tileDomainShape( newVarP, baseType, shape )	hit_tileDomainShapeInternal( newVarP, sizeof(baseType), 0==strncmp("HitTile",#baseType,7), shape )
void hit_tileDomainShapeInternal(void *newVarP, size_t baseExtent, int hierDepth, HitShape shape);



/**
 * Constant to allocate only the vertices in a CSR/Bitmap sparse shape.
 */
#define HIT_VERTICES 1

/*
 * Constant to allocate  edges in a CSR/Bitmap sparse shape.
 */
#define HIT_EDGES 2

/*
 * Constant to allocate vertices and edges in a CSR/Bitmap sparse shape.
 */
#define HIT_VERTICES_EDGES	(HIT_VERTICES | HIT_EDGES)




/* 4.c. INITIALIZING VARIABLES: ARRAY */
/**
 * hit_tileAlloc: Allocates memory for the elements of an array handler.
 * This variable should have been previously initialized and a shape declared.
 * @param var	A pointer to a HitTile derived type variable.
 * @see hit_tileDomain
 */
#define hit_tileAlloc( var )	hit_tileAllocInternal( var, #var, __FILE__, __LINE__);

void hit_tileAllocInternal(void *newVarP, const char *name, const char *file, int numLine);

/**
 * hit_tileDomainAlloc: declare the shape and base type of a new multidimensional array
 * and allocates memory for the elements.
 * @param var	A pointer to a HitTile derived type variable.
 * @param baseExtent 	Size of the base type for the array elements.
 * @param numDims 	Number of dimensions of the array. Limited by HIT_MAXDIMS.
 * @param ... 	A list of integer positive numbers defining the cardinalities of each dimension.
 * @see hit_tileDomain
 * @see hit_tileAlloc
 */
#define hit_tileDomainAlloc(var,baseType,numDims,...)	\
	hit_tileDomainInternal(var, sizeof(baseType), 0==strncmp("HitTile",#baseType,7), numDims, __VA_ARGS__);		\
	hit_tileAllocInternal( var, #var, __FILE__, __LINE__);
	

/**
 * hit_tileDomainShapeAlloc: declare the shape and base type of a new multidimensional array
 * and allocates memory for the elements.
 * @param var	A pointer to a HitTile derived type variable.
 * @param baseExtent 	Size of the base type for the array elements.
 * @param shape		Shape of the new array.
 * @see hit_tileDomain
 * @see hit_tileAlloc
 */
#define hit_tileDomainShapeAlloc(var,baseType,shape)	\
	hit_tileDomainShapeInternal(var, sizeof(baseType), 0==strncmp("HitTile",#baseType,7), shape); \
	hit_tileAllocInternal( var, #var, __FILE__, __LINE__);

/**
 * hit_tileFill: Initializes all positions of a variable with a value.
 * This variable should have been allocated.
 * @param var	A pointer to a HitTile derived type variable.
 * @param value Direction of the data which value has to been used in the initialization. 
 * Size of value should be equal to the elements of var.
 */
#define hit_tileFill( var, value )	hit_tileFillInternal( var, value, #var, __FILE__, __LINE__);

void hit_tileFillInternal(void * var, void * value, const char *name, const char *file, int numLine);


/**
 * hit_tileFileWrite: Write the data corresponding to the shape of the variable to a binary file.
 * If the file system support parallel writings this function can be called in parallel.
 * 
 * This variable should have been allocated, or it must be a selection of an allocated variable.
 * @param var	A pointer to a HitTile derived type variable.
 * @param coord	Coordinate system in the file: HIT_FILE_TILE for tile coordinates, HIT_FILE_ARRAY for array coordinates
 * @param file	Name of the file.
 * @param datatype  HIT_FILE_INT, HIT_FILE_LONG, HIT_FILE_FLOAT, or HIT_FILE_DOUBLE
 * @param s1  width for the format of the numbers
 * @param s2  precision for the format of the numbers (ignored for integers or longs)
 * @return 0 if ok, negative if variable has not memory
 * @see hit_error.h
 */
#define hit_tileFileWrite( var, file, coord )    \
	hit_tileFileInternal( var, HIT_FILE_BINARY, HIT_FILE_WRITE, coord, HIT_FILE_UNKNOWN, 0, 0, file, #var, __FILE__, __LINE__)

/**
 * hit_tileFileRead: Read the data corresponding to the shape of the variable from a binary file.
 * If the file system support parallel readings this function can be called in parallel.
 * 
 * This variable should have been allocated, or it must be a selection of an allocated variable.
 * @param var	A pointer to a HitTile derived type variable.
 * @param coord	Coordinate system in the file: HIT_FILE_TILE for tile coordinates, HIT_FILE_ARRAY for array coordinates
 * @param file	Name of the file.
 * @param datatype  HIT_FILE_INT, HIT_FILE_LONG, HIT_FILE_FLOAT, or HIT_FILE_DOUBLE
 * @param s1  width for the format of the numbers
 * @param s2  precision for the format of the numbers (ignored for integers or longs)
 * @return 0 if ok, negative if variable has not memory
 * @see hit_error.h
 */
#define hit_tileFileRead( var, file, coord )    \
	hit_tileFileInternal( var, HIT_FILE_BINARY, HIT_FILE_READ, coord, HIT_FILE_UNKNOWN, 0, 0, file, #var, __FILE__, __LINE__)


/**
 * hit_tileTextFileWrite: Write the data corresponding to the shape of the variable to a text file.
 * If the file system support parallel writings this function can be called in parallel.
 * 
 * This variable should have been allocated, or it must be a selection of an allocated variable.
 * @param var	A pointer to a HitTile derived type variable.
 * @param coord	Coordinate system in the file: HIT_FILE_TILE for tile coordinates, HIT_FILE_ARRAY for array coordinates
 * @param file	Name of the file.
 * @param datatype  HIT_FILE_INT, HIT_FILE_LONG, HIT_FILE_FLOAT, or HIT_FILE_DOUBLE
 * @param s1  width for the format of the numbers
 * @param s2  precision for the format of the numbers (ignored for integers or longs)
 * @return 0 if ok, negative if variable has not memory
 * @see hit_error.h
 */
#define hit_tileTextFileWrite( var, file, coord, datatype, s1, s2 )    \
	hit_tileFileInternal( var, HIT_FILE_TEXT, HIT_FILE_WRITE, coord, datatype, s1, s2, file, #var, __FILE__, __LINE__)

/**
 * hit_tileTextFileRead: Read the data corresponding to the shape of the variable from a text file.
 * If the file system support parallel readings this function can be called in parallel.
 * 
 * This variable should have been allocated, or it must be a selection of an allocated variable.
 * @param var	A pointer to a HitTile derived type variable.
 * @param coord	Coordinate system in the file: HIT_FILE_TILE for tile coordinates, HIT_FILE_ARRAY for array coordinates
 * @param file	Name of the file.
 * @param datatype  HIT_FILE_INT, HIT_FILE_LONG, HIT_FILE_FLOAT, or HIT_FILE_DOUBLE
 * @param s1  width for the format of the numbers
 * @param s2  precision for the format of the numbers (ignored for integers or longs)
 * @return 0 if ok, negative if variable has not memory
 * @see hit_error.h
 */
#define hit_tileTextFileRead( var, file, coord, datatype, s1, s2 )    \
	hit_tileFileInternal( var, HIT_FILE_TEXT, HIT_FILE_READ, coord, datatype, s1, s2, file, #var, __FILE__, __LINE__)


int hit_tileFileInternal( void * varP, const int fileFormat, const int fileMode, const int tileMode, const int type, const int formatSize1, const int formatSize2, const char *fileName, const char *debugVarName, const char *debugCodeFile, int debugCodeNumLine);
  	

/* 4.e. FREE VARIABLE (FREE INTERNAL DATA BUFFER) */
/**
 * hit_tileFree: free allocated memory for the elements of an array handler.
 * In case the handler has not allocated its own memory (like a selection) it is ignored.
 * @param var	A HitTile derived type variable.
 */
#define hit_tileFree(var)	{ \
	if ( (var).hierDepth ) { \
		hit_tileFreeRecInternal(&(var)); \
	} \
	if ((var).memStatus==HIT_MS_OWNER) {\
		free((var).memPtr); (var).memPtr=NULL; (var).data=NULL; (var).memStatus=HIT_MS_NOMEM;\
	} \
}

void hit_tileFreeRecInternal(void * var);


/* 4.f. DUPLICATE A WHOLE VARIABLE WITH ITS MEMORY AND DATA */
/**
 * hit_tileClone: clone the array. Duplicate the handler and the array elements.
 * @param newVar	A pointer to an empty HitTile derived type variable.
 * @param oldVar	A pointer to a initialized HitTile derived type variable.
 */
#define hit_tileClone( newVar, oldVar ) hit_tileCloneInternal(newVar, oldVar, #oldVar, __FILE__, __LINE__)
void hit_tileCloneInternal(void *newVar, const void *oldVar, const char *name, const char *file, int numLine);

/* 4.X. TESTING: REDUCE DIMENSIONS */
/**
 * hit_tileReduceDims: reduce the shape of a HitTile variable some number
 * of dimensions, only if the cardinality of the dimensions to reduce is 1.
 * The requested dimensions are removed and the variable.
 * For example, if a variable have four dimensions, and its cardinalities
 * are 1,1,2,4, we may request to reduce 2 dimensions. The result is the same
 * variable, with the same data, but having only 2 dimensions
 * with cardinalities 2,4.
 * @param newVar	A pointer to a HitTile derived type variable.
 * @param numReductions Number of dimension reductions.
 */
int hit_tileReduceDims(void *newVar, int numReductions);

/* 4.g. SELECTIONS */
/** Out of bounds flag. */
#define	HIT_OUTOFBOUNDS_CHECK		0
/** No nout of bounds flag.*/
#define	HIT_NO_OUTOFBOUNDS_CHECK	1


/**
 * Checks if the selection shape is out of bounds using tile coordinates.
 * @param tileP	A pointer to an initialized HitTile derived type variable.
 * @param sh	A subselection shape in coordinates relative to the original array domain.
 * @return out of bounds or no out of bounds flag.
 */
int hit_tileCheckBoundaryTileCoords(const void *tileP, HitShape sh);

/**
 * Checks if the selection shape is out of bounds using array coordinates.
 * @param tileP	A pointer to an initialized HitTile derived type variable.
 * @param sh	A subselection shape in coordinates relative to the original array domain.
 * @return out of bounds or no out of bounds flag.
 */
int hit_tileCheckBoundaryArrayCoords(const void *tileP, HitShape sh);


/* 4.g.1. SELECTING FROM VARIABLES TO RETURN ANOTHER VARIABLE (LOCAL COORDINATES) */
/**
 * Create a subselection handler using relative coordinates, and skipping to check if the 
 * subselection domain is out of bounds of the original variable.
 * Elements out of bounds can not be used is such selections. This functionality is
 * thought to create buffers in coordinates relative to a previous variable domain,
 * but allocate the selection before using it.
 * @param newVar	A pointer to a new HitTile derived type variable.
 * @param oldVar	A pointer to an initialized HitTile derived type variable.
 * @param shape		A subselection shape in coordinates relative to the original array domain.
 */
#define hit_tileSelectNoBoundary( newVar, oldVar, shape )	\
			hit_tileSelectInternal( newVar, oldVar, shape, HIT_NO_OUTOFBOUNDS_CHECK )

/**
 * Create a subselection handler using relative coordinates.
 * It checks if the subselection domain is out of bounds of the original variable and
 * returns HIT_NULL_VAR in that case.
 * @param newVar	A pointer to a new HitTile derived type variable.
 * @param oldVar	A pointer to an initialized HitTile derived type variable.
 * @param shape		A subselection shape in coordinates relative to the original array domain.
 * @return The subselection.
 */
#define hit_tileSelect( newVar, oldVar, shape )	\
			hit_tileSelectInternal( newVar, oldVar, shape, HIT_OUTOFBOUNDS_CHECK )

/**
 * hit_tileSelectInternal: Create a subselection handler using relative coordinates.
 * @param newVar	A pointer to a new HitTile derived type variable.
 * @param oldVar	A pointer to an initialized HitTile derived type variable.
 * @param shape		A subselection shape in coordinates relative to the original array domain.
 * @param out		Check if the subselection domain is out of bounds.
 * @return The subselection.
 */
int hit_tileSelectInternal(void *newVar, const void *oldVar, HitShape sh, int out);

/* 4.g.2. SELECTING FROM VARIABLES TO RETURN ANOTHER VARIABLE (GLOBAL COORDINATES) */
/**
 * Create a subselection handler using array coordinates, and skipping to check if the 
 * subselection domain is out of bounds of the original variable.
 * Elements out of bounds can not be used is such selections. This functionality is
 * thought to create buffers in coordinates relative to a previous variable domain,
 * but allocate the selection before using it.
 * @param newVar	A pointer to a new HitTile derived type variable.
 * @param oldVar	A pointer to an initialized HitTile derived type variable.
 * @param shape		A subselection shape in array coordinates.
 * @return The subselection.
 */
#define hit_tileSelectArrayCoordsNoBoundary( newVar, oldVar, shape )	\
			hit_tileSelectArrayCoordsInternal( newVar, oldVar, shape, HIT_NO_OUTOFBOUNDS_CHECK )

/**
 * Create a subselection handler using array coordinates.
 * It checks if the subselection domain is out of bounds of the original variable and
 * returns HIT_NULL_VAR in that case.
 * @param newVar	A pointer to a new HitTile derived type variable.
 * @param oldVar	A pointer to an initialized HitTile derived type variable.
 * @param shape		A subselection shape in array coordinates.
 * @return The subselection.
 */
#define hit_tileSelectArrayCoords( newVar, oldVar, shape )	\
			hit_tileSelectArrayCoordsInternal( newVar, oldVar, shape, HIT_OUTOFBOUNDS_CHECK )

/**
 * Create a subselection handler using array coordinates.
 * @param newVar	A pointer to a new HitTile derived type variable.
 * @param oldVar	A pointer to an initialized HitTile derived type variable.
 * @param shape		A subselection shape in array coordinates.
 * @param out		Check if the subselection domain is out of bounds.
 * @return The subselection.
 */
int hit_tileSelectArrayCoordsInternal(void *newVar, const void *oldVar, HitShape sh, int out);

/* 5. SHADOW COPY */

/* 5.a. UPDATE VALUES IN SHADOW COPY */
/**
 * Update the values of the elements of the reference variable (ancestor) of a shadow.
 *
 * A shadow is a selection of a reference variable whose own memory has
 * been allocated. This function considers the elements of the shadow to be
 * a copy of the elements of the original variable. It updates the original
 * elements with the values of the copy in the shadow.
 *
 * If the handler is not a shadow, it is ignored.
 * @param shadow	A pointer to a HitTile derived type variable.
 */
#define hit_tileUpdateToAncestor( shadow ) hit_tileUpdateFromToAncestorInternal( shadow, HIT_UPDATE_TO, #shadow, __FILE__, __LINE__ );

/**
 * Update the values of a shadow using the elements of the reference variable (ancestor).
 *
 * A shadow is a selection of a reference variable whose own memory has
 * been allocated. This function considers the elements of the shadow to be
 * a copy of the elements of the original variable. It updates the elements 
 * of the copy in the shadow with the values of the original elements.
 *
 * If the handler is not a shadow, it is ignored.
 * @param shadow	A pointer to a HitTile derived type variable.
 */
#define hit_tileUpdateFromAncestor( shadow ) hit_tileUpdateFromToAncestorInternal( shadow, HIT_UPDATE_FROM, #shadow, __FILE__, __LINE__ );

void hit_tileUpdateFromToAncestorInternal(void *shadow, int fromTo, const char *name, const char *file, int numLine);

/**
 * Find the root ancestor, the array tile from which this selection derives
 * @param tile	A tile
 * @return	Pointer to the root ancestor (to the same tile if it is not a selection)	
 */
static inline HitTile *	hit_tileRoot( void * tileP ) {
	HitTile *tile = (HitTile *)tileP;
	while( tile->ref != NULL ) tile = tile->ref;
	return tile;
}

/* 6. SELECTING IFOUT OF BOUNDS (TODO: Deprecated v1.0) */
/* AVOID EVALUATION OF THE SECOND ARGUMENT IF POSSIBLE */
/* GLOBAL TEMPORAL VARIABLE */
#define hit_tileSelectOut(select1,select2) { if (!select1) select2; }


/* 7. MERGE/FUSION OF TILES WITH A TOUCHING BOUNDARY WITH THE SAME GEOMETRY */
void hit_tileFusion( const void *tileInA, const void *tileInB, void *tileOut );


/* 
* 8. Hit MACROS FOR ACCESING THE VARIABLES
*/

/**
 * Returns the type of the tile.
 * @fn int hit_tileType(HitTile tile)
 * @memberof HitTile
 * @param tile A tile
 * @return The type of the tile.
 */
#define hit_tileType(tile) ((tile).type)


/* 8.1. BASIC SHAPE DATA */
/**
 * hit_tileDims: gets the number of dimensions of a HitTile variable.
 * @param var		A HitTile derived type variable.
 * @return 		Number of dimensions of the HitTile variable.
 */
#define	hit_tileDims(var)	(hit_shapeDims((var).shape))

/**
 * hit_tileShape: gets the shape of a HitTile variable.
 * @param var		A HitTile derived type variable.
 * @return 		Shape of the HitTile variable.
 */
#define	hit_tileShape(var)	((var).shape)

/**
 * @arturo Feb 2013
 * hit_tileShapeLocal: gets the shape of a HitTile variable in tile coordinates.
 * @param var		A HitTile derived type variable.
 * @return 		Shape of the HitTile variable in tile coordinates.
 */
HitShape hit_tileShapeLocal( const void *inTile );


/* 8.2. CARDINALITIES OF A GIVEN DIMENSION */
/* 8.2.1. CARDINALITY OF A GIVEN DIMENSION */
/**
 * hit_tileDimCard: gets the cardinality of a HitTile variable in the given dimension.
 * @param var		A HitTile derived type variable.
 * @param dim		Number of the dimension.
 * @return 		Cardinality in the specified dimension of the HitTile variable.
 */
#define	hit_tileDimCard(var,dim)	((var).card[dim])

/* 8.2.2. CARDINALITY OF THE WHOLE DOMAIN */
/**
 * hit_tileCard: gets the cardinality of a HitTile variable.
 * @param var		A HitTile derived type variable.
 * @return 		Cardinality of the HitTile variable.
 */
#define	hit_tileCard(var)	((var).acumCard)


/* 8.3. SIGNATURE OF A GIVEN DIMENSION */
/**
 * hit_tileDimSig: gets the signature of a specific dimension of a HitTile variable.
 * @param var		A HitTile derived type variable.
 * @param dim		Number of the dimension.
 * @return 		Signature in the specified dimension of the HitTile variable.
 */
#define	hit_tileDimSig(var,dim)		(hit_shapeSig(hit_tileShape((var)),dim))
/**
 * hit_tileDimBegin: gets the begin of the signature of a specific dimension of a HitTile variable.
 * @param var		A HitTile derived type variable.
 * @param dim		Number of the dimension.
 * @return 		Begin of the signature in the specified dimension of the HitTile variable.
 */
#define	hit_tileDimBegin(var,dim)		(hit_tileDimSig(var,dim).begin)
/**
 * hit_tileDimEnd: gets the end of the signature of a specific dimension of a HitTile variable.
 * @param var		A HitTile derived type variable.
 * @param dim		Number of the dimension.
 * @return 		Begin of the signature in the specified dimension of the HitTile variable.
 */
#define	hit_tileDimEnd(var,dim)		(hit_tileDimSig(var,dim).end)
/**
 * hit_tileDimStride: gets the stride of the signature of a specific dimension of a HitTile variable.
 * @param var		A HitTile derived type variable.
 * @param dim		Number of the dimension.
 * @return 		Stride of the signature in the specified dimension of the HitTile variable.
 */
#define	hit_tileDimStride(var,dim)		(hit_tileDimSig(var,dim).stride)
/**
 * hit_tileDimLast: gets the last tile-coordinate of a specific dimension of a HitTile variable.
 * @param var		A HitTile derived type variable.
 * @param dim		Number of the dimension.
 * @return 		Last tile-coordinate in the specified dimension of the HitTile variable.
 */
#define	hit_tileDimLast(var,dim)		(hit_tileDimCard(var,dim)-1)


/* 8.4. LOOPS FOR TRAVERSING COORDINATE SPACES */
/**
 * hit_tileForDimArrayDomain: Loop across the array coordinate space of a dimension of the tile
 * @param tile Tile.
 * @param dim Dimension.
 * @param index Variable index.
 */
#define	hit_tileForDimDomainArray(tile, dim, index)	for( index=hit_tileDimBegin(tile, dim);	\
												index<=hit_tileDimEnd(tile,dim);	\
												index=index+hit_tileDimStride(tile,dim) )

/**
 * hit_tileForDimDomain: Loop across the local coordinate space of a dimension of the tile
 * @param tile Tile.
 * @param dim Dimension.
 * @param index Variable index.
 */
#define	hit_tileForDimDomain(tile, dim, index)	for( index=0;	\
												index<hit_tileDimCard(tile,dim);	\
												index++ )



/* 8.5. ACCESS DATA ELEMENTS WITH TILE COORDINATES */
/** 
 * hit_tileGet: obtains an element of a HitTile.
 * @param var		A HitTile derived type variable.
 * @param ndims		Number of dimensions of the array. It must be known in compilation time. Limited by HIT_MAXDIMS.
 * @param ...		A list of integer positive numbers defining the coordinate in each dimension.
 * @return 		Element read.
 */
#define hit_tileGet(var,ndims,...)	hit_tileElemAt(var,ndims,__VA_ARGS__)

/** 
 * hit_tileSet: write a value in the indicate position of a HitTile.
 * @param var		A HitTile derived type variable.
 * @param ndims		Number of dimensions of the array. It must be known in compilation time. Limited by HIT_MAXDIMS.
 * @param val		Value to write.
 * @param ...		A list of integer positive numbers defining the coordinate in each dimension.
 */
#define hit_tileSet(var,ndims,val,...)	hit_tileElemAt(var,ndims,__VA_ARGS__) = (val)

/* 8.6. ACCESS DATA ELEMENTS WITH ARRAY COORDINATES */
/** 
 * hit_tileGetArrayCoords: obtains an element of a HitTile.
 * @param var		A HitTile derived type variable.
 * @param ndims		Number of dimensions of the array. It must be known in compilation time. Limited by HIT_MAXDIMS.
 * @param ...		A list of integer positive numbers defining the array coordinate in each dimension.
 * @return 		Element read.
 */
#define hit_tileGetArrayCoords(var,ndims,...)	hit_tileElemAtArrayCoords(var,ndims,__VA_ARGS__)

/** 
 * hit_tileSetArrayCoords: write a value in the indicate position of a HitTile.
 * @param var		A HitTile derived type variable.
 * @param ndims		Number of dimensions of the array. It must be known in compilation time. Limited by HIT_MAXDIMS.
 * @param val		Value to write.
 * @param ...		A list of integer positive numbers defining the array coordinate in each dimension.
 */
#define hit_tileSetArrayCoords(var,ndims,val,...)	hit_tileElemAtArrayCoords(var,ndims,__VA_ARGS__) = (val)


/* 8.7. ACCESS DATA ELEMENTS WITH TILE COORDINATES IN TILES WITH NO STRIDE */
/** 
 * hit_tileGetNoStride: obtains an element of a HitTile (with no stride in the selection).
 * @param var		A HitTile derived type variable.
 * @param ndims		Number of dimensions of the array. It must be known in compilation time. Limited by HIT_MAXDIMS.
 * @param ...		A list of integer positive numbers defining the coordinate in each dimension.
 * @return 		Element read.
 */
#define hit_tileGetNoStride(var,ndims,...)	hit_tileElemAtNoStride(var,ndims,__VA_ARGS__)

/** 
 * hit_tileSetNoStride: write a value in an element of a HitTile (with no stride in the selection).
 * @param var		A HitTile derived type variable.
 * @param ndims		Number of dimensions of the array. It must be known in compilation time. Limited by HIT_MAXDIMS.
 * @param val		Value to write.
 * @param ...		A list of integer positive numbers defining the coordinate in each dimension.
 */
#define hit_tileSetNoStride(var,ndims,val,...)	hit_tileElemAtNoStride(var,ndims,__VA_ARGS__) = (val)


/* 8.8. INTERNAL: MACROS TO OBTAIN THE POSITION OF AN ELEMENT */
/* 8.8.1. TILE COORDINATES */
#define hit_tileElemAt(var,ndims,...)	hit_tileElemAt##ndims(var,__VA_ARGS__)

#define hit_tileElemAt1(var, pos)	((var).data[(pos)*(var).qstride[0]])

#define hit_tileElemAt2(var, pos1, pos2)	((var).data[(pos1)*(var).qstride[0]*(var).origAcumCard[1]+(pos2)*(var).qstride[1]])

#define hit_tileElemAt3(var, pos1, pos2, pos3)	((var).data[(pos1)*(var).qstride[0]*(var).origAcumCard[1]+(pos2)*(var).qstride[1]*(var).origAcumCard[2]+(pos3)*(var).qstride[2]])

#define hit_tileElemAt4(var, pos1, pos2, pos3, pos4)	((var).data[(pos1)*(var).qstride[0]*(var).origAcumCard[1]+(pos2)*(var).qstride[1]*(var).origAcumCard[2]+(pos3)*(var).qstride[2]*(var).origAcumCard[3]+(pos4)*(var).qstride[3]])


/* 8.8.2. ARRAY COORDINATES */
#define hit_tileElemAtArrayCoords(var,ndims,...)	hit_tileElemAtArrayCoords##ndims(var,__VA_ARGS__)

#define hit_tileElemAtArrayCoords1(var, pos)	((var).data[ hit_tileArray2Tile(var,0,pos) ])

#define hit_tileElemAtArrayCoords2(var, pos1, pos2)	((var).data[ hit_tileArray2Tile((var),0,(pos1))*(var).origAcumCard[1] + hit_tileArray2Tile((var), 1, (pos2))])

#define hit_tileElemAtArrayCoords3(var, pos1, pos2, pos3)	((var).data[ hit_tileArray2Tile((var),0,(pos1))*(var).origAcumCard[1] + hit_tileArray2Tile((var), 1, (pos2))*(var).origAcumCard[2] + hit_tileArray2Tile((var), 2, (pos3))])

#define hit_tileElemAtArrayCoords4(var, pos1, pos2, pos3, pos4)	((var).data[ hit_tileArray2Tile((var),0,(pos1))*(var).origAcumCard[1] + hit_tileArray2Tile((var), 1, (pos2))*(var).origAcumCard[2] + hit_tileArray2Tile((var), 2, (pos3))*(var).origAcumCard[3] + hit_tileArray2Tile((var), 3, (pos3))])


/* 8.8.3. TILE COORDINATES IN TILES WITH NO STRIDE */
#define hit_tileElemAtNoStride(var,ndims,...)	hit_tileElemAtNoStride##ndims(var,__VA_ARGS__)

#define hit_tileElemAtNoStride1(var, pos)	((var).data[pos])

#define hit_tileElemAtNoStride2(var, pos1, pos2)	((var).data[(pos1)*(var).origAcumCard[1]+(pos2)])

#define hit_tileElemAtNoStride3(var, pos1, pos2, pos3)	((var).data[(pos1)*(var).origAcumCard[1]+(pos2)*(var).origAcumCard[2]+(pos3)])

#define hit_tileElemAtNoStride4(var, pos1, pos2, pos3, pos4)	((var).data[(pos1)*(var).origAcumCard[1]+(pos2)*(var).origAcumCard[2]+(pos3)*(var).origAcumCard[3]+(pos4)])



/* 8.9. TRANSFORM COORDINATES: TILE TO ARRAY, ARRAY TO TILE */
#define hit_tileTile2Array(var, dim, pos)	((pos)*hit_tileDimStride(var,dim) + hit_tileDimBegin(var,dim))
#define hit_tileArray2Tile(var, dim, pos)	(((pos)-hit_tileDimBegin(var,dim))/ hit_tileDimStride(var,dim))

/* 8.10. CHECK IF ARRAY COORDINATES INDEX IS IN TILE */
#define hit_tileDimHasArrayCoord(var, dim, pos)	\
	( pos>=hit_tileDimBegin(var,dim) && (pos<=hit_tileDimEnd(var,dim)) && \
	((pos-hit_tileDimBegin(var,dim))%hit_tileDimStride(var,dim) == 0) )

#define hit_tileHasArrayCoords(var, ndims, ...)	hit_tileHasArrayCoords##ndims(var, __VA_ARGS__)

#define hit_tileHasArrayCoords1(var, pos)	\
	( pos>=hit_tileDimBegin(var,0) && (pos<=hit_tileDimEnd(var,0)) && \
	((pos-hit_tileDimBegin(var,0))%hit_tileDimStride(var,0) == 0) )

#define hit_tileHasArrayCoords2(var, pos0, pos1) \
		( pos0>=hit_tileDimBegin(var,0) && (pos0<=hit_tileDimEnd(var,0)) && \
		pos1>=hit_tileDimBegin(var,1) && (pos1<=hit_tileDimEnd(var,1)) && \
		((pos0-hit_tileDimBegin(var,0))%hit_tileDimStride(var,0) == 0) && \
		((pos1-hit_tileDimBegin(var,1))%hit_tileDimStride(var,1) == 0) )

#define hit_tileHasArrayCoords3(var, pos0, pos1, pos2) \
		( pos0>=hit_tileDimBegin(var,0) && (pos0<=hit_tileDimEnd(var,0)) && \
		pos1>=hit_tileDimBegin(var,1) && (pos1<=hit_tileDimEnd(var,1)) && \
		pos2>=hit_tileDimBegin(var,2) && (pos2<=hit_tileDimEnd(var,2)) && \
		((pos0-hit_tileDimBegin(var,0))%hit_tileDimStride(var,0) == 0) && \
		((pos1-hit_tileDimBegin(var,1))%hit_tileDimStride(var,1) == 0) && \
		((pos2-hit_tileDimBegin(var,2))%hit_tileDimStride(var,2) == 0) )

#define hit_tileHasArrayCoords4(var, pos0, pos1, pos2, pos3) \
		( pos0>=hit_tileDimBegin(var,0) && (pos0<=hit_tileDimEnd(var,0)) && \
		pos1>=hit_tileDimBegin(var,1) && (pos1<=hit_tileDimEnd(var,1)) && \
		pos2>=hit_tileDimBegin(var,2) && (pos2<=hit_tileDimEnd(var,2)) && \
		pos3>=hit_tileDimBegin(var,3) && (pos3<=hit_tileDimEnd(var,3)) && \
		((pos0-hit_tileDimBegin(var,0))%hit_tileDimStride(var,0) == 0) && \
		((pos1-hit_tileDimBegin(var,1))%hit_tileDimStride(var,1) == 0) && \
		((pos2-hit_tileDimBegin(var,2))%hit_tileDimStride(var,2) == 0) && \
		((pos3-hit_tileDimBegin(var,3))%hit_tileDimStride(var,3) == 0) )

/**
 * hit_tileShapeTile2Array: transform a shape from tile coordinates to
 * array coordinates using the the shape fron a HitTile variable
 * @param var		A HitTile derived type variable.
 * @param sh		Input Shape
 * return		New shape in the new coordinates system
*/
HitShape hit_tileShapeTile2Array(void *var, HitShape sh);

/**
 * hit_tileShapeArray2Tile: transform a shape from array coordinates to
 * tile coordinates using the the shape fron a HitTile variable
 * @param var		A HitTile derived type variable.
 * @param sh		Input Shape
 * return		New shape in the new coordinates system
*/
HitShape hit_tileShapeArray2Tile(void *var, HitShape sh);



/* @arturo Feb 2013 */
/* HitTile GLUE TILES WITH TOUCHING BOUNDARIES OF THE SAME GEOMETRY */
/* TODO: Extent to more than one dimension */
/* To be deprecated: Not used in any example */
void hit_tileGlue( void *tileInA, void *tileInB, void *tileOut );





/* END OF HEADER FILE _HitTile_ */
#endif
