

#ifndef _HitWorker_
#define _HitWorker_

#include <hit_taskcontainers.h>
#include <hit_mpi.h>


// ----------------------------------------------------------- //

/**
 * Generic list of elements.
 */
typedef struct {
	size_t	size;		/**< Size of a single element. */
	int		nelems;		/**< Number of elements in the list. */
	void	*elements;	/**< Array with the elements. */
} HitAuxList;

/**
 * Auxiliary list constructor.
 */
#define hit_auxListNewType(newType)			\
typedef struct HitAuxList_##newType {		\
	size_t size;							\
	int nelems;								\
	newType	*elements;						\
} HitAuxList_##newType


#define hit_auxListInit(listP,type)		\
{										\
	(*(listP)).size = sizeof(type);		\
	(*(listP)).nelems = 0;				\
	(*(listP)).elements = NULL;			\
}


/**
 * Macro to add elements to the list.
 */
#define hit_auxListAdd(listP, value,cast)							\
{																	\
	(*(listP)).nelems++;											\
	size_t size = ((*(listP)).size) * ((size_t)(*(listP)).nelems);	\
	hit_realloc(((*(listP)).elements),size,cast);					\
	(*(listP)).elements[((*(listP)).nelems)-1] = (value);			\
}

/**
 * Macro to free the list.
 */
#define hit_auxListFree(listP) 		\
{									\
	(*listP).size = 0;				\
	hit_free((*listP).elements);	\
}

/**
 * Number of elements of the list
 */
#define hit_auxListNElems(listP) 	((*listP).nelems)

/**
 * Item i of the list
 */
#define hit_auxListElem(listP,i) 	(((*listP).elements)[(i)])

/**
 * Null constant for the list.
 */
#define HIT_AUX_LIST_NULL_STATIC {0,0,NULL}


// ----------------------------------------------------------- //


/**
 * Number of possible nested contexts.
 */
#define HIT_CONTEXT_NUMBER 5


/**
 * HitWorker context.
 */
typedef struct{
	int rank;	/**< Process ID. */
	int size;	/**< Number of processes. */
	int next;	/**< Next available. */
	int io;		/**< Process that will do the IO in the Pipe. */
	MPI_Comm comm;
} HitWorkerContext;


/**
 * Global variables used by the workers.
 */
typedef struct {

	int current_context; /**< Current context. */
	HitWorkerContext context[HIT_CONTEXT_NUMBER]; /**< Context list. */

} HitWorkerGlobal;

/**
 * Variable with the global worker information.
 */
extern HitWorkerGlobal HitWorkerGlobalVar;

/**
 * Initializes the contexts.
 */
void hit_workerContextInit();

/**
 * Maps a worker to a process.
 * @return The process rank.
 */
int hit_workerContextGetRank();

/**
 * Shortcut for the current context.
 */
#define hit_workerCurrentContext (HitWorkerGlobalVar.context[HitWorkerGlobalVar.current_context])


// ----------------------------------------------------------- //

/* Worker types */
#define HIT_WORKER_BASIC 1 /**< Basic worker */
#define HIT_WORKER_PIPE  2 /**< Pipe         */
#define HIT_WORKER_FARM  3 /**< Farm (Pool)  */

/* Worker threads */
#define HIT_WORKER_THREAD_INPUT			0	/**< Input thread. */
#define HIT_WORKER_THREAD_OUTPUT		1	/**< User function and output thread. */
#define HIT_WORKER_THREAD_WS_INPUT		2	/**< Input thread for workstealing farm. */
#define HIT_WORKER_THREAD_WS_OUTPUT		3	/**< Ouput thread for workstealing farm. */
#define HIT_WORKER_THREAD_NUM			4	/**< Number of threads. */


/**
 * Struct the the configuration of a worker
 */
typedef struct{

	int buffer_in; /**< Size of the input buffer (Number of tasks) */
	int buffer_out;/**< Size of the ouput buffer (Number of tasks) */

	double threshold_send; /**< Threshold to send the buffer before it is full (Total weight of tasks) */

	double threshold_ws_send; /**< Threshold to send a WS request (Total weight of task in the list) */
	double threshold_ws_request; /**< Minimum weight of the task list used to decide whether send tasks or not (Total weight of tasks) */

} HitWorkerConfig;


#define HIT_WORKER_CONFIG_STATIC {1000,1000,1.0,0.5,0.5}


/**
 * Worker config variable.
 */
extern HitWorkerConfig HIT_WORKER_CONFIG;



/**
 * Typedef for the HitWorker.
 */
typedef struct HitWorker HitWorker;

/**
 * Typedef for the HitWorker pointer.
 */
typedef HitWorker* HitWorkerP;

/**
 * Typedef for the User function.
 * @param
 */
typedef  void  (*HitWorkerFunc)  (HitWorker*);


/**
 * Define the integer list.
 */
hit_auxListNewType(int);

/**
 * Define the worker list.
 */
hit_auxListNewType(HitWorkerP);


/**
 * Struct to keep the successors.
 */
typedef struct {
	int proc;	/**< Processor id. */
	int flag;	/**< MPI flag to send the msg. */
} HitSuccessor;

// Create the list type for the successor
hit_auxListNewType(HitSuccessor);

/**
 * Worker struct for basic and aggregate skeletons.
 */
struct HitWorker {

	int 					type;				/**< Type of the worker (Basic, Pipe, Farm). */
	int						context;			/**< Context where the worker was created. */
	int						num_input_channels;	/**< Number of inputs. */
	int						num_output_channels; /**< Number of outputs. */

	/** Predecessors, workers that send task to this worker. */
	HitAuxList_int			predecessors[HIT_TASK_NCHANNELS_MAX];
	/** Entrances are the processors that should execute the worker. */
	HitAuxList_int			entrances;
	/** Successors for the task. */
	HitAuxList_HitSuccessor	successors[HIT_TASK_NCHANNELS_MAX];

	/** Successor where the last task has been sent. */
	int						last_successor[HIT_TASK_NCHANNELS_MAX];
	/*< Worker where the last work stealing request was sent. */
	int						last_ws;

	HitTaskList				list;								/**< Local task list. */
	HitTaskBuffer			inbuffer[HIT_TASK_NCHANNELS_MAX];	/**< Buffer for incoming tasks. */
	HitTaskBuffer			outbuffer[HIT_TASK_NCHANNELS_MAX];	/**< Buffer for outgoing tasks. */

	HitWorkerFunc			func;				/**< Function used by the worker. */
	HitAuxList_HitWorkerP	subworkers;			/**< Subworker for the pipe and farm. */

	pthread_mutex_t			mutex_ws;			/**< Mutex for workstealing. */
	int						pending_request;	/**< Flag for pending ws request. */
	int						finished_ws;		/**< Flag to mark that the ws is finished. */
	MPI_Comm				comm_ws;			/**< Workstealing communicator for the farm. */

	MPI_Datatype			type_input[HIT_TASK_NCHANNELS_MAX];			/**< Datatype for input tasks. */
	MPI_Datatype			type_output[HIT_TASK_NCHANNELS_MAX];		/**< Datatype for output tasks. */

	int 					enable_ws;			/**< Flag to enable or disable the ws */
	HitWorkerP				parent;				/**< Pointer to the parent for pipes and farms */

	HitWorkerConfig			config;				/**< Worker config */

#ifdef TODO_IO_PIPE

	/* Data for the IO pipe */
	// ---------------------
	void * data_in;
	int num_in;
	size_t size_in;

	void * data_out;
	int num_out;
	size_t size_out;
	// ---------------------
#endif

};


/**
 * Worker null constant.
 */
#define HIT_WORKER_NULL_STATIC {			\
	0,0,0,0,								\
	{},										\
	HIT_AUX_LIST_NULL_STATIC,				\
	{},										\
	{0},0,									\
	HIT_TASK_LIST_NULL_STATIC,				\
	{},										\
	{},										\
	NULL,									\
	HIT_AUX_LIST_NULL_STATIC,				\
	PTHREAD_MUTEX_INITIALIZER,0,0,			\
	MPI_COMM_NULL,							\
	{}, {},									\
	1,NULL,									\
	HIT_WORKER_CONFIG_STATIC				\
}


/**
 * Worker null variable.
 */
extern HitWorker HIT_WORKER_NULL;


/**
 * Worker constructor.
 * @param worker Pointer to the worker.
 * @param f_worker Function executed by the worker.
 * @param num_in Number of input types
 * @param num_out Number of output types
 * @param ... List of types
 */
void hit_workerCreate(HitWorker * worker, HitWorkerFunc f_worker, int num_in, int num_out, ...);

/**
 * Set w2 as successor of w1 and w1 as predecessor of w2.
 * @param w1 A worker pointer.
 * @param output Output channel.
 * @param w2 A worker pointer.
 * @param input Input channel.
 */
void hit_workerAddSuccessor(HitWorker * w1, int output, HitWorker * w2, int input);


/**
 * Combines several simple workers using the collaboration operator.
 * This is a generic farm where each worker can be different.
 * @param worker Pointer the the return worker
 * @param num Number of subworkers
 * @param w1 the first worker
 * @param ... The workers
 */
void hit_workerOpCollaboration(HitWorker * worker, int num, HitWorker * w1, ...);


/**
 * Creates a Pipeline worker combining several workers.
 * @param wpipe Pointer to the pipe worker
 * @param num Number of stages
 * @param ... Pointer to the worker of each stage.
 * @note This function only works for worker with one input and one output channel.
 */
void hit_workerOpSuccession(HitWorker * wpipe, int num, ...);

/**
 * Creates a Farm of workers cloning a basic worker that is given as
 * a parameter.
 * @param wfarm Pointer to the farm.
 * @param num Number of workers of the farm.
 * @param worker Pointer to a basic worker.
 */
void hit_workerCreateFarm(HitWorker * wfarm, int num, HitWorker * worker);


/**
 * Frees a worker
 * @param worker Pointer to the worker.
 */
void hit_workerFree(HitWorker * worker);


/**
 * Worker thread that process the incoming tasks.
 * @param workerP Pointer to the worker.
 */
void *thread_input(void * workerP);

/**
 * Worker thread that process the outgoing tasks.
 * @param workerP Pointer to the worker.
 */
void *thread_output(void * workerP);

/**
 * Gets a new task from the list, it could come
 * from some predecessor or from the WS pool.
 * @param worker The pointer to the worker.
 * @param taskP Pointer to the task.
 * @return True of false depending if there are more task.
 */
int hit_workerGetTask(HitWorker * worker, void * taskP);

/**
 * Send the task to any successor.
 * @param worker The pointer to the worker.
 * @param taskP Pointer to the task or task array.
 * @param output Output channel
 */
void hit_workerSendTask(HitWorker * worker, void * taskP, int output);

/**
 * Adds a new task to the local list.
 * @param worker The pointer to the worker.
 * @param taskP Pointer to the task.
 */
void hit_workerAddTask(HitWorker * worker, void * taskP);


/**
 * Executes the worker.
 * @param worker
 */
void hit_workerExecute(HitWorker * worker);


// ----------------------------------------------------------- //


#define hit_workerFlagSet(flag,mutex,value)		\
{												\
	pthread_mutex_lock(mutex);					\
	(flag) = (value);							\
	pthread_mutex_unlock(mutex);				\
}

#define hit_workerFlagGet(flag,mutex,var)		\
{												\
	pthread_mutex_lock(mutex);					\
	(var) = (flag);								\
	pthread_mutex_unlock(mutex);				\
}


#endif

