
/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 20011-2012,Yuri Torres de la Sierra 
 *                          Arturo Gonzalez Escribano
 *                          
 *                          
 * All Rights Reserved.
 */


#ifndef _HitAcel_
#define _HitAcel_

#include <hit_tile.h>


/*
 ***********************************************************************
 *
 *							DEFINES AND DATA STRUCTS 
 *  
***********************************************************************
 * */

/* The FERMI GPU has the number 0 in the device recognition*/
#define FERMI_GPU	0


/* GPUs dimension idexes*/
#define GPU_DIM_ROWS		1
#define GPU_DIM_COLUMS		0
#define GPU_DIM_DEPTH		2


/* Cuda clock init*/
#define hit_acelClockInit(var, var2) 	cudaEventCreate(&var);\
					cudaEventCreate(&var2);\
					cudaEventRecord(var, 0 )	
/* Cuda clock stop*/
#define hit_acelClockStop(var, var2, var3)	cudaEventRecord( var2, 0 );\
		cudaEventSynchronize( var2 );\
		cudaEventElapsedTime( &var3, var, var2 );\
		cudaEventDestroy( var );\
		cudaEventDestroy( var2 )

/* In CUDA the maximun number of block dimensions is 3*/
#define MAX_NUM_BLOCK_DIMENSIONS 3

/* In CUDA the maximun number of gRID dimensions is 2*/
#define MAX_NUM_GRID_DIMENSIONS 2

/* Maximum number of current GPU cards*/
#define MAX_NUM_CARDS 4	

/** Number of threads per block**/

	/* In FERMI 1024 is the max number of threads per block*/  
	#define MAX_THREAD_PER_BLOCK_FERMI 1024

	/* In PRE-FERMI 1024 is the max number of threads per block*/  
	#define MAX_THREAD_PER_BLOCK_PRE_FERMI 512

/** Number of threads per SM**/
	/* In FERMI the maximun number of threads per SM is 1536*/
	#define MAX_THREADS_PER_SM_FERMI 1536

	/* In pre-FERMI the maximun number of threads per SM is 1024*/
	#define MAX_THREADS_PER_SM_PRE_FERMI 1536

/** Number of threads per cores per SM**/

	/* In FERMI the number of cores per SM is 32*/
	#define NUMBER_OF_CORES_PER_SM_FERMI	32

	/* In pre-FERMI the number of cores per SM is 8*/	
	#define NUMBER_OF_CORES_PER_hit_acelSendDataInternalSM_PRE_FERMI 8


/* Cuda kernel features*/
	/*Global memory access pattern (coalesced or not coalesced)*/
#define COALESCING		1
#define NO_COALESCING		0

	/*Cuda kernel work-load, is a ration between global memory accesses and the number of arithmetic-logic operations*/
#define WORKLOAD		1
#define NO_WORKLOAD		0

	/*The data reutilization in the Cuda kernel*/
#define REUTILIZATION		1
#define NO_REUTILIZATION	0


/* To calculate the maximum number of threads per block for a given GPU*/
#define max_threads_per_block(gpu)  (HIT_GPUs_TOPOLOGY_INFO.GPUInf[(gpu)].maxThreadsPerBlock)


/* Struct that stores the GPUs features*/
typedef struct {
						
		int internalName; 			/*Internal GPU name (in the GPU system);
									* is a redundant field because de arrays position (in HitPTopology)
									* indicates the GPU name (is the same number)*/
		int numSMs;								//Number of SMs		
		int numCores;							//Number of GPU cores	
		int maxThreadsPerBlock;						//Maximum number of threads per block	 
		int maxThreadsPerSM; 						//Maximum number of threads per SM
		int maxThreadsPerDimension[MAX_NUM_BLOCK_DIMENSIONS];		//Maximun number of thread per Block dimension
		int maxBlocksPerDimension[MAX_NUM_GRID_DIMENSIONS];			//maximum number of block per Grid dimension
			int major;
			int minor;
} HitPGPUTopology;


/*extern HitPGPUsTopology HIT_GPU_TOPOLOGY_INFO;*/
#define	HIT_GPU_TOPOLOGY_NULL_STATIC {0, 0, 0, 0, 0, {0, 0, 0}, {0, 0}, 0, 0}


/* Struct to define the grid and threadBlock shape*/
typedef struct {

	HitShape blockShape;			//Normal thread-block shape
	HitShape gridShape;				//Normal Grid shape

	HitShape blockShapeFringe;		//Fringe thread-block shape
	HitShape gridShapeFringe;		//Fringe Grid shape
} HitGpuShape;

/* HitGpuShape NULL variables*/
extern HitGpuShape HIT_GPU_NULL;

#define HIT_GPUSHAPE_NULL_STATIC {HIT_SHAPE_NULL_STATIC, HIT_SHAPE_NULL_STATIC, HIT_SHAPE_NULL_STATIC, HIT_SHAPE_NULL_STATIC}


/* Physical topology data structure for GPUs*/
typedef struct {

	int numGPUs;			
		
	HitPGPUTopology		GPUInf[MAX_NUM_CARDS];	
	HitSig				GPUProcessLocation[MAX_NUM_CARDS];	
} HitPGPUsTopology;

#define	HIT_GPUs_TOPOLOGY_NULL_STATIC	{ 0, {HIT_GPU_TOPOLOGY_NULL_STATIC, HIT_GPU_TOPOLOGY_NULL_STATIC, HIT_GPU_TOPOLOGY_NULL_STATIC,HIT_GPU_TOPOLOGY_NULL_STATIC}}		//yuri


/* EXTERN VARIABLES */ 
extern HitPGPUsTopology	HIT_GPUs_TOPOLOGY_INFO;


/* Hit UPDATE MEMORY CARDINALITIES: INTERNAL */
	/* To update the origAcumCard when we have a selection variable without memory*/
	/* This function is part of  update funcion*/
static inline void hit_tileUpdateAcumCards(void *newVarP) {
	HitTile *newVar = (HitTile *)newVarP;
	int i;
	int cardinality;

	newVar->origAcumCard[hit_shapeDims(newVar->shape)] = 1;
	cardinality=1;

	for(i=hit_shapeDims(newVar->shape)-1; i>=0; i--) {
		cardinality = cardinality * newVar->card[i];
		newVar->origAcumCard[i] = cardinality;
	}
	newVar->acumCard = newVar->origAcumCard[0];
}//hit_tileUpdateAcumCards


/* To declare variables from tiles*/
/* For GPUs: hitAcel_##tile is the tile to be launched to GPU
*  hitAcelData_##tile is the data vector in the GPU device
*  hitAcelPadding_##tile is the tile with the information of the padding
*/
#define hit_acelDeclTile(tile,part, accessPattern,type)	HitTile * hitAcel_##tile;\
						void * hitAcelData_##tile;\
						HitTile_##type  hitAcelPadding_##tile;\
						HitShape extended_##tile= hit_partitionExtendedDataPaddingPatern(&tile,part);\
						hitAcelPadding_##tile=tile;\
						hit_acelDeclData(tile,type);\
//hit_tileDumpTileFile( hitAcelPadding_##tile, "pp" );\
						//hit_tileSelectNoBoundary( &hitAcelPadding_##tile, &tile, extended_##tile);\
						//hit_tileUpdateAcumCards(&hitAcelPadding_##tile);\

						//******TE TIENE QUE DAR UN AMPLIADO
						//HitShape extended_##tile= hit_partitionDataPaddingPatern(tile.shape, part, accessPattern);\
						//dataPadding_Pattern_##tile=hit_partitionMaxShape(part);\

/* Kernel definition depending with data pointer or with the Hit_tile struct*/
#ifdef POINTER_DATA

		/* kernelDefinition for CUDA (for tile data pointer)*/
		#define kernelDefinitionCUDA(type, name, ...)		__global__ void type##_##name( __VA_ARGS__ );\
								extern "C"\
								void INTERNAL_##name(HitPartition part, __VA_ARGS__)\
								{\
									acelKernelVariables();\
									GPUFER_##name<<<grid,block>>>( HIT_PARAM_KERNEL_LAUNCH_POINTER);\
									CUDA_SAFE_CALL(cudaThreadSynchronize());\
								}\
								__global__ void type##_##name( __VA_ARGS__ )


#else //POINTER_DATA

		/* kernelDefinition for CUDA (for hitTile pointer)*/
		#define kernelDefinitionCUDA(type, name, ...)		__global__ void type##_##name(HitShape dataShape , __VA_ARGS__ );\
									extern "C"\
									void INTERNAL_##name(HitPartition part, __VA_ARGS__)\
									{\
										acelKernelVariables();\
										GPUFER_##name<<<grid,block>>>(part.origDataPatternShape, HIT_PARAM_KERNEL_LAUNCH_##name);\
										CUDA_SAFE_CALL(cudaThreadSynchronize());\
									}\
									__global__ void type##_##name(HitShape dataShape, __VA_ARGS__ )

#endif //POINTER_DATA


/* kernelDefinition for CORE*/
#define kernelDefinitionCORE(type, name, ...)	void type##_##name(HitShape dataShape, __VA_ARGS__)

/* kernel parameters depending on the data pointer of hit_tile pointer*/
#ifdef POINTER_DATA

	/* Parameters to be used in intermediate kernel launch (size-variable parameters)*/
	#define hit_cudaParameters(n, ... )						hit_cudaParameters##n( __VA_ARGS__ )
	#define hit_cudaParameters1(tile1)						hitAcelData_##tile1
	#define hit_cudaParameters2(tile1,tile2)				hitAcelData_##tile1,hitAcelData_##tile2
	#define hit_cudaParameters3(tile1,tile2,tile3)			hitAcelData_##tile1,hitAcelData_##tile2,hitAcelData_##tile3
	#define hit_cudaParameters4(tile1,tile2,tile3,tile4)	hitAcelData_##tile1,hitAcelData_##tile2,hitAcelData_##tile3,hitAcelData_##tile4


#else /*POINTER_DATA*/


	/* Parameters to be used in intermediate kernel launch (size-variable parameters)*/
	#define hit_cudaParameters(n, ... )							hit_cudaParameters##n( __VA_ARGS__ )
	#define hit_cudaParameters1(tile1)							hitAcel_##tile1
	#define hit_cudaParameters2(tile1,tile2)					hitAcel_##tile1,hitAcel_##tile2
	#define hit_cudaParameters3(tile1,tile2,tile3)				hitAcel_##tile1,hitAcel_##tile2,hitAcel_##tile3
	#define hit_cudaParameters4(tile1,tile2,tile3,tile4)		hitAcel_##tile1,hitAcel_##tile2,hitAcel_##tile3,hitAcel_##tile4

	/* Macros to declare variables depending on the number of the variables*/
	#define hit_acelDeclTileALL(n,part,type, ... )				hit_acelDeclTileALL##n(part, type, __VA_ARGS__ )
	#define hit_acelDeclTileALL1(part,type,tile1)				hit_acelDeclTile(tile1,part,part.accessPattern[0],type)
	#define hit_acelDeclTileALL2(part,type,tile1,tile2)			hit_acelDeclTileALL1(part,type,tile1); hit_acelDeclTile(tile2,part,part.accessPattern[1],type)
	#define hit_acelDeclTileALL3(part,type,tile1,tile2,tile3)	hit_acelDeclTileALL2(part,type,tile1,tile2); hit_acelDeclTile(tile3,part,part.accessPattern[2],type)

	/* Macros to send data depending on the number of variables*/
	#define hit_acelSendDataALL(n, ... )						hit_acelSendDataALL##n( __VA_ARGS__ )
	#define hit_acelSendDataALL1(tile1)							if (hit_acelIN_OUTVector[0]==IN || hit_acelIN_OUTVector[0]==INOUT) {hit_acelSendData(tile1);}
	#define hit_acelSendDataALL2(tile1,tile2)					hit_acelSendDataALL1(tile1);if (hit_acelIN_OUTVector[1]==IN || hit_acelIN_OUTVector[1]==INOUT) {hit_acelSendData(tile2);}
	#define hit_acelSendDataALL3(tile1,tile2,tile3)				hit_acelSendDataALL2(tile1,tile2); if (hit_acelIN_OUTVector[2]==IN || hit_acelIN_OUTVector[2]==INOUT) {hit_acelSendData(tile3);}

	/* Macros to receive data depending on the number of variables*/
	#define hit_acelRecvDataALL(n, ... )						hit_acelRecvDataALL##n( __VA_ARGS__ )
	#define hit_acelRecvDataALL1(tile1)							if (hit_acelIN_OUTVector[0]==OUT || hit_acelIN_OUTVector[0]==INOUT) {hit_acelRecvData(tile1);}
	#define hit_acelRecvDataALL2(tile1,tile2)					hit_acelRecvDataALL1(tile1); if (hit_acelIN_OUTVector[1]==OUT || hit_acelIN_OUTVector[1]==INOUT) {hit_acelRecvData(tile2);}
	#define hit_acelRecvDataALL3(tile1,tile2,tile3)				hit_acelRecvDataALL2(tile1,tile2); if (hit_acelIN_OUTVector[2]==OUT || hit_acelIN_OUTVector[2]==INOUT) {hit_acelRecvData(tile3);}

	/* Macros to free data memory depending on the number of variables*/
	#define hit_acelFreeALL(n, ... )							hit_acelFreeALL##n( __VA_ARGS__ )
	#define hit_acelFreeALL1(tile1)								hit_acelFree(tile1)
	#define hit_acelFreeALL2(tile1,tile2)						hit_acelFreeALL1(tile1); hit_acelFree(tile2)
	#define hit_acelFreeALL3(tile1,tile2,tile3)					hit_acelFreeALL2(tile1,tile2); hit_acelFree(tile3)
 

	/* To initialize the inout variables*/
	#define hit_acelTileIN_OUT(n, ... )							n,hit_acelTileIN_OUTInternal(hit_acelIN_OUTVector,n,  __VA_ARGS__ ) //hit_acelIN_OUTInternal##n( __VA_ARGS__ )
	#define hit_acelIN_OUTInternal1(num1)						hit_acelIN_OUTVector[0]=num1
	#define hit_acelIN_OUTInternal2(num1,num2)					hit_acelIN_OUTInternal1((num1)); hit_acelIN_OUTVector[1]=(num2)
	#define hit_acelIN_OUTInternal3(num1,num2,num3)				hit_acelIN_OUTInternal2((num1),(num2)); hit_acelIN_OUTVector[2]=(num3)

	#define IN	0
	#define OUT	1
	#define INOUT 	2

	int hit_acelTileIN_OUTInternal(int *inout, int n, ...);
	
	
/*GPU global memory access pattern*/

	/* Neutral signature to calculate hit_partition (it does not have any effect on the partition)*/
	#define HIT_SIG_ZERO			HIT_SIG_ONE_POS


	/* ThreadBlock access pattern*/
	#define	HIT_ACEL_SIG_THIS { 0, 0, 1}
	#define	HIT_ACEL_SIG_ALL  { 0, 0,-1}
	#define	HIT_ACEL_SIG_E_1 { 1, -1, 1}
	#define	HIT_ACEL_SIG_E_1_M_1 { 1, 0, 1}

/* Funcitons to create signatures from a given threadBlock access pattern*/
	static inline HitSig hit_acelSigTHIS() {
		HitSig newSig=HIT_ACEL_SIG_THIS;
		return newSig;
	}

	static inline HitSig hit_acelSigALL() {
		HitSig newSig=HIT_ACEL_SIG_ALL;
		return newSig;
	}

	static inline HitSig hit_acelSigE1() {
		HitSig newSig=HIT_ACEL_SIG_E_1;
		return newSig;
	}

	static inline HitSig hit_acelSigE1M1() {
		HitSig newSig=HIT_ACEL_SIG_E_1_M_1;
		return newSig;
	}

	/* Macros to create the appropiate shape dependion on global memory access pattern*/
	#define hit_acelPatTHISTHIS			hit_shape2(hit_acelSigTHIS(), hit_acelSigTHIS())	
	#define hit_acelPatTHISALL			hit_shape2(hit_acelSigTHIS(), hit_acelSigALL())
	#define hit_acelPatALLTHIS			hit_shape2(hit_acelSigALL(), hit_acelSigTHIS())
	#define hit_acelPatALLALL			hit_shape2(hit_acelSigALL(), hit_acelSigALL())
	//#define hit_acelPatE1E1			hit_shape2(hit_acelSigE1(), hit_acelSigE1())
	#define hit_acelPatE1E1E1			hit_shape3(hit_acelSigE1(), hit_acelSigE1(), hit_acelSigE1())
	#define hit_acelPatE1M1E1M1E1M1		hit_shape3(hit_acelSigE1M1(), hit_acelSigE1M1(), hit_acelSigE1M1())


#endif //POINTER_DATA



/*
*****************************************************************************************************************
*
*			__CUDACC__    __CUDACC__    __CUDACC__    __CUDACC__    __CUDACC__    __CUDACC__    __CUDACC__
*
*****************************************************************************************************************
*/

/* If is defined the CUDACC compiler */
#ifdef  __CUDACC__

	/* kernel variables (Threads and block identification)*/	
	#define kernelVariables()  \
			int hitTidX		=	threadIdx.x;\
			int hitTidY		=	threadIdx.y;\
			int hitBidX		=	blockIdx.x;\
			int hitBidY		=	blockIdx.y;\
			int hitBDimX		=	blockDim.x;\
			int hitBDimY		=	blockDim.y;\
			int hitGDimX		=	gridDim.x;

#endif //__CUDACC__

/***********************************************************************************************************************************************
*
*		END	__CUDACC__  END  __CUDACC__  END  __CUDACC__  END  __CUDACC__  END  __CUDACC__  END  __CUDACC__  END  __CUDACC__  END
*
*************************************************************************************************************************************************/



/***********************************************************************************************************************************************
*
*						For each compilation mode.....
*
*************************************************************************************************************************************************/



/* If is defined CUDA compilation mode*/
#ifdef HIT_COMPILE_CUDA


	/* Include the NVIDIA acel driver*/
	//#include "../cutest/acelNVIDIA.h"

	/* If is defined CUDA and CORE compilation mode*/
	/*CUDA AND CORE******************************************************************************************************************/
	#ifdef HIT_COMPILE_CORE

	/* To know the work-load between CPU-GPU*/
	#include "../cutest/config.h"

	/* If is defined CUBLAS use then the library is oincluded*/
	#ifdef CUBLASUSE	

	#include <cublas.h>

	#endif //CUBLASUSE

	/* MPI TAG_SYNCHRONIZE as parameter in MPI send an recv functions (to synchronize)*/
	#define HIT_MPI_TAG_SYNCHRONIZE		32000

	/* Macro to have a mutual exclusion per each GPU independently; no use==0; use==1*/
	#define hit_acelMExclusionInit() \
			MPI_Status	status;\
			int acelMPIOk = 0;\
			int acelGlobalGPUUse[HIT_GPUs_TOPOLOGY_INFO.numGPUs];\
			\
			{int acelGlobalGPUUUseIndex;\
				for(acelGlobalGPUUUseIndex=0; acelGlobalGPUUUseIndex < HIT_GPUs_TOPOLOGY_INFO.numGPUs;acelGlobalGPUUUseIndex++)\
					acelGlobalGPUUse[acelGlobalGPUUUseIndex]=0;}

	/* Macro to start the mutual exclusion region*/
	#define hit_acelExclusionStart() { \
			if (HIT_TOPOLOGY_INFO.selfRank < acelGPUMpiThreads){\
				if (HIT_GPUs_TOPOLOGY_INFO.GPUInf[acelCurrentGPU].major>=1 ){\
					if ( HIT_TOPOLOGY_INFO.selfRank !=HIT_GPUs_TOPOLOGY_INFO.GPUProcessLocation[acelCurrentGPU].begin ){\
						acelMPIOk = MPI_Recv( NULL, 0, MPI_CHAR, HIT_TOPOLOGY_INFO.selfRank-1, HIT_MPI_TAG_SYNCHRONIZE, MPI_COMM_WORLD, &status );\
						hit_mpiTestError(acelMPIOk,"hit_acelExclusionStart");\
					}\
					if (HIT_TOPOLOGY_INFO.selfRank ==HIT_GPUs_TOPOLOGY_INFO.GPUProcessLocation[acelCurrentGPU].begin && acelGlobalGPUUse[acelCurrentGPU]>0 &&hit_sigCard(HIT_GPUs_TOPOLOGY_INFO.GPUProcessLocation[acelCurrentGPU])>1){\
						acelMPIOk = MPI_Recv( NULL, 0, MPI_CHAR,HIT_GPUs_TOPOLOGY_INFO.GPUProcessLocation[acelCurrentGPU].end, HIT_MPI_TAG_SYNCHRONIZE, MPI_COMM_WORLD, &status );\
						hit_mpiTestError(acelMPIOk,"hit_acelExclusionStart");\
					}\
					acelGlobalGPUUse[acelCurrentGPU]++;\
				}\
			}

	/* Macro to stop the mutual exclusion region*/ //el número de hilos por dimension //HIT_GPUs_TOPOLOGY_INFO.selfRank <= 4 &&
	#define hit_acelExclusionStop() \
			if (HIT_TOPOLOGY_INFO.selfRank < acelGPUMpiThreads){\
				if (HIT_GPUs_TOPOLOGY_INFO.GPUInf[acelCurrentGPU].major>=1 ){\
					if (  hit_sigCard(HIT_GPUs_TOPOLOGY_INFO.GPUProcessLocation[acelCurrentGPU])>1){\
						if (HIT_TOPOLOGY_INFO.selfRank == HIT_GPUs_TOPOLOGY_INFO.GPUProcessLocation[acelCurrentGPU].end)\
						acelMPIOk = MPI_Send( NULL, 0, MPI_CHAR, HIT_GPUs_TOPOLOGY_INFO.GPUProcessLocation[acelCurrentGPU].begin, HIT_MPI_TAG_SYNCHRONIZE, MPI_COMM_WORLD );		\
						if (HIT_TOPOLOGY_INFO.selfRank != HIT_GPUs_TOPOLOGY_INFO.GPUProcessLocation[acelCurrentGPU].end)\
						acelMPIOk = MPI_Send( NULL, 0, MPI_CHAR, (HIT_TOPOLOGY_INFO.selfRank+1), HIT_MPI_TAG_SYNCHRONIZE, MPI_COMM_WORLD );\
						hit_mpiTestError(acelMPIOk,"hit_acelExclusionStop");\
					}\
				}\
			}\
			}//It is to remove the internal variables declaracions

	/* Macro to free the GPU memory*/
	#define hit_acelFree(tile)	freeTileCuda(hitAcel_##tile,hitAcelData_##tile,hitAcelPadding_##tile);\

	/* Macro to liberate the GPU threads*/
	#define hit_acelThreadExit()	threadExitCuda()

	

	/* Load balancing between the hardware accelerators*/  //HIT_GPUs_TOPOLOGY_INFO.selfRank%2 //HIT_GPUs_TOPOLOGY_INFO.selfRank%HIT_GPUs_TOPOLOGY_INFO.numGPUs
	#define hit_acelFasterDeviceWorkLoad(percentage)        int acelGPUMpiThreads=  percentage;\
								int acelCurrentGPU=(HIT_TOPOLOGY_INFO.selfRank < firstGPUThreads)? 0 : 1;\
								if (firstGPUThreads>acelGPUMpiThreads){\
									fprintf(stderr,"Estas intentando asignar más procesos de los existentes a la GPU\n");\
									exit (1);\
								}\
								else{\
									HIT_GPUs_TOPOLOGY_INFO.GPUProcessLocation[0]= hit_sig(0,firstGPUThreads-1,1);\
									HIT_GPUs_TOPOLOGY_INFO.GPUProcessLocation[1]= hit_sig(firstGPUThreads,acelGPUMpiThreads-1,1);\
								}

	/* If CUBLAS use is defined (Kernel launch)*/
	#ifndef CUBLASUSE

		/* Kernel launch macro without CUBLAS*/
		#define hit_acelKernelLaunch(name,part,type,inout,...) 	{int hit_acelIN_OUTVector[4]; hit_acelKernelLaunchInternal(name,part,type,inout, __VA_ARGS__);}
		#define hit_acelKernelLaunchInternal(name,part,type,n,inout,...)	if (HIT_TOPOLOGY_INFO.selfRank < acelGPUMpiThreads){\
									if (HIT_GPUs_TOPOLOGY_INFO.GPUInf[acelCurrentGPU].major>=1 ){\
										int hit_acelNulo=inout;\
										hit_acelDeclTileALL(n, part, type, __VA_ARGS__);\
										hit_acelSendDataALL(n, __VA_ARGS__);\
											INTERNAL_##name(part,hit_cudaParameters(n, __VA_ARGS__) );\
										hit_acelRecvDataALL(n, __VA_ARGS__);\
										hit_acelFreeALL(n, __VA_ARGS__);\
									}\
								else CORE_##name(part.origDataPatternShape, __VA_ARGS__ );\								
							}\
							else CORE_##name(part.origDataPatternShape, __VA_ARGS__ )

	#else //CUBLASUSE

		/* Kernel launch macro for CUBLAS*/
		#define hit_acelKernelLaunch(name, part, n, ...)	if (HIT_TOPOLOGY_INFO.selfRank < acelGPUMpiThreads){\
									if (HIT_GPUs_TOPOLOGY_INFO.GPUInf[acelCurrentGPU].major>=1 ){\
									acelGPUCUBLAS( __VA_ARGS__,HIT_TOPOLOGY_INFO.selfRank );\
									}\
								else CORE_##name( __VA_ARGS__ );\
							}\
							else CORE_##name( __VA_ARGS__ )


	#endif //CUBLASUSE


	/* Macro to declare data variables (Target GPU selection and book data memory in memory of the same GPU)*/
	#define hit_acelDeclData(tile,type)	useDevice(acelCurrentGPU);\
						if(HIT_TOPOLOGY_INFO.selfRank < acelGPUMpiThreads){\
						if (HIT_GPUs_TOPOLOGY_INFO.GPUInf[acelCurrentGPU].major>=1 ){\
				hit_acelDeclDataInternal((HitTile*)& hitAcelPadding_##tile, (HitTile**)&hitAcel_##tile, (void**)&hitAcelData_##tile, sizeof(type*));\
							}\
						}
	/* hit_acelDeclDataInternal prototype from hit_acelDeclData(tile,type) macro*/
	void hit_acelDeclDataInternal(HitTile *tile, HitTile **acelTile, void **acelData, size_t typeSize);
		
	/* Macro to send data from the host to thje target GPU*/	
	#define hit_acelSendData(tile)	if(HIT_TOPOLOGY_INFO.selfRank < acelGPUMpiThreads){\
							if (HIT_GPUs_TOPOLOGY_INFO.GPUInf[acelCurrentGPU].major>=1 ){\
								acelSendRecvDataNVIDIA((HitTile*)&tile,(HitTile*)&hitAcelPadding_##tile,HIT_UPDATE_FROM, #tile);\
							}\
						}\
	
	//hit_acelSendDataInternal((HitTile*)&hitAcelPadding_##tile, hitAcelData_##tile);\
	//acelSendRecvDataNVIDIA((HitTile*)&tile,(HitTile*)&hitAcelPadding_##tile,HIT_UPDATE_FROM, #tile);\
	
	/* hit_acelSendDataInternal propotype function from hit_acelSendData(tile) macro*/
	void hit_acelSendDataInternal(HitTile *tile, void *acelData);


	/* If we CUBLASUSE is not defined (hit_acelRecvData(tile))*/
	#ifndef CUBLASUSE
	
		/* Send the data from GPU memory to Host global memory (we need to recive data)*/
		#define hit_acelRecvData(tile)		if(HIT_TOPOLOGY_INFO.selfRank < acelGPUMpiThreads){\
								if (HIT_GPUs_TOPOLOGY_INFO.GPUInf[acelCurrentGPU].major>=1 ){\
									acelSendRecvDataNVIDIA((HitTile*)&tile,(HitTile*)&hitAcelPadding_##tile,HIT_UPDATE_TO, #tile);\
								}\
							}\
							
		//hit_acelRecvDataInternal((HitTile*)&hitAcelPadding_##tile, hitAcelData_##tile);\
		//acelSendRecvDataNVIDIA((HitTile*)&tile,(HitTile*)&hitAcelPadding_##tile,HIT_UPDATE_TO, #tile);\
							
	#else //CUBLASUSE

		/* Send the data from GPU memory to Host global memory (we do not need to recive data)*/
		#define hit_acelRecvData(tile)

	#endif //CUBLASUSE

	/* hit_acelRecvDataInternal prototype function from hit_acelRecvData(tile) macro*/
	void hit_acelRecvDataInternal(HitTile *tile, void *accelData);

	/* ONLY CUDA******************************************************************************************************************/
	#else


	
	/* Load balancing between the hardware accelerators*/  //HIT_GPUs_TOPOLOGY_INFO.selfRank%2 //HIT_GPUs_TOPOLOGY_INFO.selfRank%HIT_GPUs_TOPOLOGY_INFO.numGPUs
	#define hit_acelFasterDeviceWorkLoad(percentage)        int acelGPUMpiThreads=  percentage;\
								int acelCurrentGPU=(HIT_TOPOLOGY_INFO.selfRank < firstGPUThreads)? 0 : 1;\
								if (firstGPUThreads>acelGPUMpiThreads){\
									fprintf(stderr,"Estas intentando asignar más procesos de los existentes a la GPU\n");\
									exit (1);\
								}\
								else{\
									HIT_GPUs_TOPOLOGY_INFO.GPUProcessLocation[0]= hit_sig(0,firstGPUThreads-1,1);\
									HIT_GPUs_TOPOLOGY_INFO.GPUProcessLocation[1]= hit_sig(firstGPUThreads,acelGPUMpiThreads-1,1);\
								}


	/* If CUBLAS use is defined (Kernel launch)*/
	#ifndef CUBLASUSE

		/* Kernel launch macro without CUBLAS*/
		#define hit_acelKernelLaunch(name,part,type,inout,...) 	{int hit_acelIN_OUTVector[4]; hit_acelKernelLaunchInternal(name,part,type,inout, __VA_ARGS__);}
		#define hit_acelKernelLaunchInternal(name,part,type,n,inout,...)	if (HIT_TOPOLOGY_INFO.selfRank < acelGPUMpiThreads){\
									if (HIT_GPUs_TOPOLOGY_INFO.GPUInf[acelCurrentGPU].major>=1 ){\
										int hit_acelNulo=inout;\
										hit_acelDeclTileALL(n, part, type, __VA_ARGS__);\
										hit_acelSendDataALL(n, __VA_ARGS__);\
											INTERNAL_##name(part,hit_cudaParameters(n, __VA_ARGS__) );\
										hit_acelRecvDataALL(n, __VA_ARGS__);\
										hit_acelFreeALL(n, __VA_ARGS__);\
									}\
								else {fprintf(stderr,"Hit ERROR: There is no NVIDIA device in proc/core %d\n",HIT_TOPOLOGY_INFO.selfRank);\
									exit (-1); }\
							}\
							else {fprintf(stderr,"Hit ERROR: There is no NVIDIA device in proc/core %d\n",HIT_TOPOLOGY_INFO.selfRank);\
								exit (-1);}

	#else //CUBLASUSE

		/* Kernel launch macro for CUBLAS*/
		#define hit_acelKernelLaunch(name, part, n, ...)	if (HIT_TOPOLOGY_INFO.selfRank < acelGPUMpiThreads){\
									if (HIT_GPUs_TOPOLOGY_INFO.GPUInf[acelCurrentGPU].major>=1 ){\
									acelGPUCUBLAS( __VA_ARGS__,HIT_TOPOLOGY_INFO.selfRank );\
									}\
								else {fprintf(stderr,"Hit ERROR: There is no NVIDIA device in proc/core %d\n",HIT_TOPOLOGY_INFO.selfRank);\
									exit (-1); }\
							}\
							else {fprintf(stderr,"Hit ERROR: There is no NVIDIA device in proc/core %d\n",HIT_TOPOLOGY_INFO.selfRank);\
								exit (-1);}



	#endif //CUBLASUSE





	/* Macro to free the GPU memory*/
	#define hit_acelFree(tile)	freeTileCuda(hitAcel_##tile,hitAcelData_##tile,hitAcelPadding_##tile);\

	/* Macro to declare data variables (Target GPU selection and book data memory in memory of the same GPU)*/
	#define hit_acelDeclData(tile,type)	useDevice(acelCurrentGPU);\
						if(HIT_TOPOLOGY_INFO.selfRank < acelGPUMpiThreads){\
						if (HIT_GPUs_TOPOLOGY_INFO.GPUInf[acelCurrentGPU].major>=1 ){\
				hit_acelDeclDataInternal((HitTile*)& hitAcelPadding_##tile, (HitTile**)&hitAcel_##tile, (void**)&hitAcelData_##tile, sizeof(type*));\
							}\
						}
	/* hit_acelDeclDataInternal prototype from hit_acelDeclData(tile,type) macro*/
	void hit_acelDeclDataInternal(HitTile *tile, HitTile **acelTile, void **acelData, size_t typeSize);
		
	/* Macro to send data from the host to thje target GPU*/	
	#define hit_acelSendData(tile)	if(HIT_TOPOLOGY_INFO.selfRank < acelGPUMpiThreads){\
							if (HIT_GPUs_TOPOLOGY_INFO.GPUInf[acelCurrentGPU].major>=1 ){\
								acelSendRecvDataNVIDIA((HitTile*)&tile,(HitTile*)&hitAcelPadding_##tile,HIT_UPDATE_FROM, #tile);\
							}\
						}\
	
	//hit_acelSendDataInternal((HitTile*)&hitAcelPadding_##tile, hitAcelData_##tile);\
	//acelSendRecvDataNVIDIA((HitTile*)&tile,(HitTile*)&hitAcelPadding_##tile,HIT_UPDATE_FROM, #tile);\
	
	/* hit_acelSendDataInternal propotype function from hit_acelSendData(tile) macro*/
	void hit_acelSendDataInternal(HitTile *tile, void *acelData);


	/* If we CUBLASUSE is not defined (hit_acelRecvData(tile))*/
	#ifndef CUBLASUSE
	
		/* Send the data from GPU memory to Host global memory (we need to recive data)*/
		#define hit_acelRecvData(tile)		if(HIT_TOPOLOGY_INFO.selfRank < acelGPUMpiThreads){\
								if (HIT_GPUs_TOPOLOGY_INFO.GPUInf[acelCurrentGPU].major>=1 ){\
									acelSendRecvDataNVIDIA((HitTile*)&tile,(HitTile*)&hitAcelPadding_##tile,HIT_UPDATE_TO, #tile);\
								}\
							}\
							
		//hit_acelRecvDataInternal((HitTile*)&hitAcelPadding_##tile, hitAcelData_##tile);\
		//acelSendRecvDataNVIDIA((HitTile*)&tile,(HitTile*)&hitAcelPadding_##tile,HIT_UPDATE_TO, #tile);\
							
	#else //CUBLASUSE

		/* Send the data from GPU memory to Host global memory (we do not need to recive data)*/
		#define hit_acelRecvData(tile)

	#endif //CUBLASUSE

	/* hit_acelRecvDataInternal prototype function from hit_acelRecvData(tile) macro*/
	void hit_acelRecvDataInternal(HitTile *tile, void *accelData);


	#endif //#ifdef HIT_COMPILE_CORE

#else //HIT_COMPILE_CUDA
	//ONLY CORE******************************************************************************************************************
	#ifdef HIT_COMPILE_CORE

	/* Empty macros*/
			
		/* Various macros*/
		#define hit_acelMExclusionInit()
		#define	hit_acelExclusionStart()
		#define hit_acelFree(tile)
		#define	hit_acelFree(tile)
		#define hit_acelThreadExit()
		#define hit_acelExclusionStop()

	/* Load balancing between the hardware accelerators*/
	#define hit_acelFasterDeviceWorkLoad(percentage)       

	/* Macro to send data from Host to device*/
	#define hit_acelSendData(tile)	

	/* Macro to receive data from device to Host */
	#define hit_acelRecvData(tile)	

	/* Kernel launch*/
	#define hit_acelKernelLaunch(name, part, n, ... )					CORE_##name( __VA_ARGS__)

	/* To free the hitTile variables on the core*/
	#define hit_acelFree(tile)  

	/* Macro to allocate memory on device*/
	#define hit_acelDeclData(tile,type)
	
	#else //HIT_COMPILE_CORE
	/*THERE IS NOT COMPILE MODE******************************************************************************************************************/

		//#error "Hit COMPILATION ERROR: No proc/core support defined for this aplication (HIT_COMPILE_CORE, HIT_COMPILE_CUDA, ...)"
	#endif //HIT_COMPILE_CORE

#endif //HIT_COMPILE_CUDA


void hit_cudaInit(int *pargc, char **pargv[]);

/***********************************************************************************************************************************************
*
*				END	END	For each compilation mode.....	END	END
*
*************************************************************************************************************************************************/


/* END OF HEADER FILE _HitGpu_ */
#endif	//_HitAcel_




