/**
 * Metis layout plug-in. This layout uses the Metis partition library to
 * map the elements of a sparse shape domain into the processors of a 
 * virtual topology.
 * 
 * @file hit_layoutMetis.c
 * @version 1.0
 * @author Javier Fresno Bausela
 * @date Jun 2011
 *
 */

/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2007-2011, Arturo Gonzalez Escribano
 *                          Javier Fresno Bausela
 *                          Carlos de Blas Carton
 *                          Yuri Torres de la Sierra
 * All Rights Reserved.
 */


#include <stdio.h>
#include <math.h>

#include <hit_layout.h>
#include <hit_layoutP.h>
#include <hit_funcop.h>
#include <hit_allocP.h>
#include <hit_com.h>

#include <hit_cshape.h>
#include <hit_bshape.h>




/**
 * Performs a Broadcast communication to share a sparse shape.
 */
void hit_sparseShapeBcastInternal(HitShape * shape, HitTopology topo){

	int ok;
	int sizes[2];
	// @arturo Mar 2013
	//MPI_Comm comm = *((MPI_Comm *)topo.pTopology.lowLevel);
	MPI_Comm comm = topo.pTopology->comm;
	
	if(topo.pTopology->selfRank == 0){

		int n = hit_cShapeNvertices(*shape);
		int m2 = hit_cShapeNedges(*shape);

		sizes[0] = n;
		sizes[1] = m2;

		ok = MPI_Bcast(sizes, 2, MPI_INT, 0, comm);
		hit_mpiTestError(ok,"Error in sparseShapeBcast");
		
		ok = MPI_Bcast(&(hit_cShapeNameList(*shape,0).flagNames),1,MPI_INT,0,comm);
		hit_mpiTestError(ok,"Error in sparseShapeBcast");
		ok = MPI_Bcast(hit_cShapeXadj(*shape), n+1, MPI_INT, 0, comm);
		hit_mpiTestError(ok,"Error in sparseShapeBcast");
		ok = MPI_Bcast(hit_cShapeAdjncy(*shape), m2, MPI_INT, 0, comm);
		hit_mpiTestError(ok,"Error in sparseShapeBcast");
		ok = MPI_Bcast(hit_cShapeNameList(*shape,0).names, n, MPI_INT, 0, comm);
		hit_mpiTestError(ok,"Error in sparseShapeBcast");

	} else {

		ok = MPI_Bcast(sizes, 2, MPI_INT, 0, comm);
		hit_mpiTestError(ok,"Error in sparseShapeBcast");
		int  n = sizes[0];
		int m2 = sizes[1];

		HitShape newShp = hit_csrShape(n,m2);

		ok = MPI_Bcast(&(hit_cShapeNameList(newShp,0).flagNames),1,MPI_INT,0,comm);
		hit_mpiTestError(ok,"Error in sparseShapeBcast");
		if(hit_cShapeNameList(newShp,0).flagNames == HIT_SHAPE_NAMES_ARRAY)
			hit_cShapeNameList(newShp,0).flagNames = HIT_SHAPE_NAMES_NOARRAY;
		
		ok = MPI_Bcast(hit_cShapeXadj(newShp), n+1, MPI_INT, 0, comm);
		hit_mpiTestError(ok,"Error in sparseShapeBcast");
		ok = MPI_Bcast(hit_cShapeAdjncy(newShp), m2, MPI_INT, 0, comm);
		hit_mpiTestError(ok,"Error in sparseShapeBcast");
		ok = MPI_Bcast(hit_cShapeNameList(newShp,0).names, n, MPI_INT, 0, comm);
		hit_mpiTestError(ok,"Error in sparseShapeBcast");

		hit_cShapeNameList(newShp,0).nNames = n;
		hit_cShapeNameList(newShp,1) = hit_cShapeNameList(newShp,0);

		*shape = newShp;
	}

}

//#define DEBUG_LAY_METIS

#ifdef DEBUG_LAY_METIS
#define debug(...) { if( hit_Rank == 0 ) { printf(__VA_ARGS__); fflush(stdout); }}
#define debugall(...) { printf(__VA_ARGS__);}
#endif




HitLayout hit_layout_plug_layMetis(HitTopology topo, HitShape * shapeP){


	// 0. Broadcast the sparse shape
	hit_sparseShapeBcastInternal(shapeP,topo);

	HitShape shape = *shapeP;
	

	int i;

	// 1. Create the layout
	HitLayout lay = HIT_LAYOUT_NULL;
	hit_layout_list_initGroups(&lay,hit_cShapeNvertices(shape));
	lay.topo = topo;
	lay.pTopology[0] = hit_ptopDup(topo.pTopology);
	//lay.origShape = shape;

	// 2. Obtain the number of partitions
	int numVertices = hit_cShapeNvertices(shape);
	int numProcessors = hit_topCard(topo);
	
	
	// @todo @javfres HitLayout metis doesn't work if there is more
	// processes than nodes, we can't call metis and we must split
	// the communicator and set non active processros.
	/*
	if(numProcessors <= numVertices){
		lay.active = 1;
	} else {
	
		if(topo.linearRank < numProcessors){
			lay.active = 1;
		} else {
			lay.active = 0;
		}
		
		// Split the communicator
		HitPTopology ptopo = HIT_PTOPOLOGY_NULL;
		MPI_Comm * new;
		hit_malloc(new,sizeof( MPI_Comm ));
		
		int result = MPI_Comm_split( *( (MPI_Comm *) (topo.pTopology.lowLevel)) ,lay.active, topo.linearRank,new);
		hit_mpiTestError(result,"Error using MPI_Comm_split");
	    MPI_Comm_rank(*new, &ptopo.selfRank);
	    MPI_Comm_size(*new, &ptopo.numProcs);
		ptopo.lowLevel = (void*) new;
		lay.pTopology[0] = ptopo;
		
	}
    */


	
	int numParts = min(numVertices,numProcessors);
	lay.numActives[0] = numParts;
	lay.type = HIT_LAYOUT_METIS;

	for(i=0;i<numParts;i++){
		hit_layout_list_addGroup(&lay,-1,numProcessors / numParts + ((numProcessors % numParts > i) ? 1 : 0 ));
	}

	int group = hit_lay_procGroup(lay,topo.linearRank);
	lay.group = group;


	// 3. Call the partition function
	int wgtflag = 0;
	int numflag = 0;
	int options[5] = {0,0,0,0,0};
	int edgecut;

	// We can use the assignedGroups generated in hit_layout_list_initGroups as the part array.
	int * part = lay.info.layoutList.assignedGroups;

	if(numProcessors > 1){

		// Call METIS function.
		METIS_PartGraphKway (&hit_cShapeNvertices(shape), hit_cShapeXadj(shape),
			hit_cShapeAdjncy(shape), NULL, NULL,
			&wgtflag, &numflag, &numParts, options, &edgecut, part);

	} else {

		bzero(part, (size_t) numVertices * sizeof(int));
	}


#ifdef DEBUG_LAY_METIS
	{int i; sleep(2*hit_Rank);
	debug("part: ");
	for(i=0;i<hit_cShapeNvertices(shape);i++){
		debug("%d",part[i]);
	}
	debug("\n");}
#endif

	// 4.a Get the number of vertices (n) from the partition.
	int n = 0;

	for(i=0;i<hit_cShapeNvertices(shape);i++){
		if(part[i] == group){
			n++;
		}
	}

	
	// 4.b Get the number of edges (m).
	int vertex;
	int m=0;

	for(vertex=0; vertex<hit_cShapeNvertices(shape); vertex++){

		// Get first and last links.
		int first = hit_cShapeXadj(shape)[vertex];
		int last = hit_cShapeXadj(shape)[vertex+1];
		int link;


		for(link=first; link<last; link++){

			//Get the neighbor.
			int neighbor = hit_cShapeAdjncy(shape)[link];
			if((part[vertex] == group && part[neighbor] == group) ){
					m++;
			}

		}
	
	}


#ifdef DEBUG_LAY_METIS
	{int i; 
	debugall("p(%d): ",group);
	debugall(" n=%d, m=%d \n",n,m);}
#endif

	
	// 5. Create the local shape.
	HitShape lshape = hit_csrShape(n,m);

	// 5.1 Calculate the global-local name translation.
	int current = 0;
	for(i=0; i<hit_cShapeNvertices(shape); i++){
		if( part[i] == group ){
			hit_cShapeNameList(lshape,0).names[current++] = hit_cShapeNameList(shape,0).names[i];
		}
	}

	// Create the inverse translation list to speedup the algorithm.
	hit_cShapeCreateInvNames(&lshape);

#ifdef DEBUG_LAY_METIS
	{int i; 
	debugall("globalName(%d): ",group);
	for(i=0;i<hit_cShapeNvertices(lshape);i++){
		debugall("%d",hit_cShapeNameList(lshape,0).names[i]);
	}
	debugall("\n");}
#endif

	// 5.2 Init xadj and adjncy
	hit_cShapeXadj(lshape)[0] = 0;
	int lvertex;

	
	// Create the local graph.
	for(lvertex=0; lvertex<hit_cShapeNvertices(lshape); lvertex++){

		
		
		hit_cShapeXadj(lshape)[lvertex+1] = hit_cShapeXadj(lshape)[lvertex];

		int gvertex = hit_cShapeNameList(lshape,0).names[lvertex];
		
		

		// Locate the vertex in the global shape.
		int g;
		for(g=0;g<hit_cShapeNvertices(shape); g++){
			if( gvertex == hit_cShapeNameList(shape,0).names[g]  ){
				vertex = g;
				break;
			}
		}


		
		// Get first and last links.
		int first = hit_cShapeXadj(shape)[vertex];
		int last = hit_cShapeXadj(shape)[vertex+1];
		int link;

		for(link=first; link<last; link++){

			//Get the neighbor.
			int neighbor = hit_cShapeAdjncy(shape)[link];
			int gneighbor = hit_cShapeNameList(shape,0).names[neighbor];
			int lneighbor = hit_cShapeVertexToLocal(lshape,gneighbor);
			
			if(lneighbor != -1){

				hit_cShapeAdjncy(lshape)[hit_cShapeXadj(lshape)[lvertex+1]] = lneighbor;
				hit_cShapeXadj(lshape)[lvertex+1] ++;
			}
		}
	}

#ifdef DEBUG_LAY_METIS
	{int i; 
	debugall("xadj(%d): ",group);
	for(i=0;i<hit_cShapeNvertices(lshape)+1;i++){
		debugall("%d",hit_cShapeXadj(lshape)[i]);
	}
	debugall("\n");}


	{int i;
	debugall("adjncy(%d): ",group);
	for(i=0;i<m;i++){
		debugall("%d",hit_cShapeAdjncy(lshape)[i]);
	}
	debugall("\n");}
#endif


	//
	hit_cShapeNameList(lshape,1) = hit_cShapeNameList(lshape,0);



	// 7. Fill the layout structure.
	lay.origShape = shape;
	hit_layShape(lay) = lshape;

	// 8. Create the assignedGroups array
	lay.info.layoutList.numElementsTotal = hit_cShapeNvertices(shape);


	// 10. Return the layout
	return lay;
}







