

#include <hit_worker.h>
#include <hit_com.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>

#include <string.h>



// ----------------------------------------------------------- //


HitWorkerConfig HIT_WORKER_CONFIG = HIT_WORKER_CONFIG_STATIC;


// ----------------------------------------------------------- //


/**
 * Variable with the global worker information.
 */
HitWorkerGlobal HitWorkerGlobalVar;


/**
 * Initializes the contexts.
 */
void hit_workerContextInit(){
	int rank, size;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &size);

	HitWorkerGlobalVar.context[0].rank = rank;
	HitWorkerGlobalVar.context[0].size = size;
	HitWorkerGlobalVar.context[0].next = 0;
	HitWorkerGlobalVar.context[0].comm = MPI_COMM_WORLD;
	HitWorkerGlobalVar.current_context = 0;
}


/**
 * Maps a worker to a process.
 * @return The process rank.
 */
int hit_workerContextGetRank(){

	int c = HitWorkerGlobalVar.current_context;
	int size = HitWorkerGlobalVar.context[c].size;
	int rank = HitWorkerGlobalVar.context[c].next;
	HitWorkerGlobalVar.context[c].next++;

	if(rank >= size){
		printf("Context %d: Not enough processes (%d)\n",c,size);
		exit(-1);
	}

	return rank;
}


// ----------------------------------------------------------- //


/**
 * Worker null variable.
 */
HitWorker HIT_WORKER_NULL = HIT_WORKER_NULL_STATIC;


/**
 * Worker constructor.
 * @param worker Pointer to the worker.
 * @param f_worker Function executed by the worker.
 * @param num_in Number of input types
 * @param num_out Number of output types
 * @param ... List of types
 */
void hit_workerCreate(HitWorker * worker, HitWorkerFunc f_worker, int num_in, int num_out, ...){

	// Set the worker to the default worker.
	*worker = HIT_WORKER_NULL;

	// Copy the worker config
	worker->config = HIT_WORKER_CONFIG;

	// Set the worker fields.
	worker->type = HIT_WORKER_BASIC;
	worker->context = HitWorkerGlobalVar.current_context;
	worker->func = f_worker;
	worker->num_input_channels = num_in;
	worker->num_output_channels = num_out;

	// Init the entrance list (list of process that execute the worker)
	hit_auxListInit(&worker->entrances,int);
	// Map the worker to a process.
	hit_auxListAdd(&worker->entrances,hit_workerContextGetRank(),int*);

	// Get the comm for our context
	MPI_Comm comm = HitWorkerGlobalVar.context[worker->context].comm;

	// The input and output types are in the variable arguments
	va_list arglist;
	va_start(arglist, num_out);

	// Init the input channels
	int i;
	for(i=0; i<num_in; i++){
		MPI_Datatype type = va_arg(arglist, MPI_Datatype);
		worker->type_input[i] = type;
		hit_auxListInit(&worker->predecessors[i],int);
		hit_taskBufferCreate(&worker->inbuffer[i],worker->config.buffer_in,comm,i,1,type);
	}

	// Ouput channels
	for(i=0; i<num_out; i++){
		MPI_Datatype type = va_arg(arglist, MPI_Datatype);
		worker->type_output[i] = type;
		hit_auxListInit(&worker->successors[i],HitSuccessor);
		hit_taskBufferCreate(&worker->outbuffer[i],worker->config.buffer_out,comm,i,1,type);
		worker->last_successor[i] = 0;
	}

	// End the list
	va_end(arglist);

	// Init the subworkers
	hit_auxListInit(&worker->subworkers,HitWorkerP);

	// Init the list
	hit_taskListCreate(&worker->list, num_in);


	/**
	 * @todo Other context.
	 */
}




/**
 * Creates a copy of the worker but assigning it to
 * a different processor.
 * @param source Pointer to the original worker.
 * @return Pointer to the clone.
 */
HitWorker * hit_workerBasicClone(HitWorker * source){

	if (source->type != HIT_WORKER_BASIC){
		printf("HitWorker: Clone not supported.\n");
	}

	HitWorker * worker;
	hit_malloc(worker,sizeof(HitWorker),HitWorker*);

	hit_workerCreate(worker, source->func, 0, 0);

	// Update the intputs and outputs
	worker->num_input_channels = source->num_input_channels;
	worker->num_output_channels = source->num_output_channels;

	// Get the comm for our context
	MPI_Comm comm = HitWorkerGlobalVar.context[worker->context].comm;

	// Init the input channels
	int i;
	for(i=0; i<worker->num_input_channels; i++){
		MPI_Datatype type = source->type_input[i];
		worker->type_input[i] = type;
		hit_auxListInit(&worker->predecessors[i],int);
		hit_taskBufferCreate(&worker->inbuffer[i],worker->config.buffer_in,comm,i,1,type);
	}

	// Output channels
	for(i=0; i<worker->num_output_channels; i++){
		MPI_Datatype type = source->type_output[i];
		worker->type_output[i] = type;
		hit_auxListInit(&worker->successors[i],HitSuccessor);
		hit_taskBufferCreate(&worker->outbuffer[i],worker->config.buffer_out,comm,i,1,type);
		worker->last_successor[i] = 0;
	}

	// Create a new list
	hit_taskListFree(&worker->list);
	hit_taskListCreate(&worker->list, worker->num_input_channels);


	worker->parent = source->parent;

	return worker;
}



/**
 * Set w2 as successor of w1 and w1 as predecessor of w2.
 * @param w1 A worker pointer.
 * @param output Output channel.
 * @param w2 A worker pointer.
 * @param input Input channel.
 */
void hit_workerAddSuccessor(HitWorker * w1, int output, HitWorker * w2, int input){

	// A. Recursive call for the first Worker
	if(w1->type == HIT_WORKER_PIPE){
		HitWorker * pipelast = hit_auxListElem(&w1->subworkers,hit_auxListNElems(&w1->subworkers)-1);

		hit_workerAddSuccessor(pipelast,output,w2,input);
		return;
	}

	if(w1->type == HIT_WORKER_FARM ){
		int i;
		for(i=0; i<hit_auxListNElems(&w1->subworkers); i++){
			HitWorker * subworker = hit_auxListElem(&w1->subworkers,i);
			hit_workerAddSuccessor(subworker, output, w2, input);
		}
		return;
	}

	// B. Recursive call for the second Worker
	if(w2->type == HIT_WORKER_PIPE){
		HitWorker * pipefirst = hit_auxListElem(&w2->subworkers,0);
		hit_workerAddSuccessor(w1,output,pipefirst,input);
		return;
	}

	if(w2->type == HIT_WORKER_FARM){
		int i;
		for(i=0; i<hit_auxListNElems(&w2->subworkers); i++){
			HitWorker * subworker = hit_auxListElem(&w2->subworkers,i);
			hit_workerAddSuccessor(w1,output, subworker, input);
		}
		return;
	}


	// Checks
	if( output >= w1->num_output_channels  ){
		hit_error_here("Adding successor, invalid output channel in the predecessor");
	}
	if( input >= w2->num_input_channels  ){
		printf("types %d[%d] -> %d[%d] \n",w1->num_output_channels,output,w2->num_input_channels,input);
		hit_error_here("Adding successor, invalid input channel in the successor");

	}

	// Set the successor
	HitSuccessor successor;
	successor.proc = hit_auxListElem(&w2->entrances,0);
	successor.flag = input;
	hit_auxListAdd(&w1->successors[output],successor,HitSuccessor*);

	// Set the predecessor
	int predecessor = hit_auxListElem(&w1->entrances,0);
	hit_auxListAdd(&(w2->predecessors[input]),predecessor,int*);

}






/**
 * Creates a Pipeline worker combining several workers.
 * @param wpipe Pointer to the pipe worker
 * @param num Number of stages
 * @param ... Pointer to the worker of each stage.
 */
void hit_workerOpSuccession(HitWorker * wpipe, int num, ...){

	int i;

	*wpipe = HIT_WORKER_NULL;

	wpipe->type = HIT_WORKER_PIPE;

	// Init the aux lists.
	//hit_auxListInit(&wpipe->predecessors,int);
	hit_auxListInit(&wpipe->entrances,int);
	//hit_auxListInit(&wpipe->successors,int);
	hit_auxListInit(&wpipe->subworkers,HitWorkerP);


	HitWorker * worker_p;
	HitWorker * worker_s;

	va_list list;
	va_start(list,num);

	worker_p =  va_arg(list,HitWorker*);
	hit_auxListAdd(&wpipe->subworkers, worker_p, HitWorker**);


	for (i=1; i<num; i++){
		worker_s = va_arg(list,HitWorker*);

		hit_auxListAdd(&wpipe->subworkers, worker_s, HitWorker**);

		hit_workerAddSuccessor(worker_p,0,worker_s,0);

		worker_p = worker_s;
	}
	va_end(list);

	// @todo @javfres Set the predecessors, entrances and successors lists.

}



/**
 * Combines several simple workers using the collaboration operator.
 * This is a generic farm where each worker can be different.
 * @param worker Pointer the the return worker
 * @param num Number of subworkes
 * @param ... The workers
 */
void hit_workerOpCollaboration(HitWorker * wmain, int num, HitWorker * w1, ...){

	if(num<1){
		printf("Farm error num workers == 0,  %d\n",num);
		return;
	}

	// No farm
	if(num == 1){
		w1->parent = NULL;
		*wmain = *w1;
		return;
	}

	int i;

	*wmain = HIT_WORKER_NULL;
	w1->parent = wmain;

	wmain->type = HIT_WORKER_FARM;

	// Init the aux lists.
	wmain->num_input_channels = w1->num_input_channels;
	wmain->num_output_channels = w1->num_output_channels;


	// @note @javfres Although it is technically possible to have farms with
	// several inputs (The buffers support sending different types of tasks)
	// if each worker receives a different number of task for each input buffer
	// the program can not continue.
	if(wmain->num_input_channels != 1){
		hit_error_here("Number of input channels must be exactly one for farms");
	}


	for(i=0; i<wmain->num_input_channels; i++){
		hit_auxListInit(&wmain->predecessors[i],int);
	}

	for(i=0; i<wmain->num_output_channels; i++){
		hit_auxListInit(&wmain->successors[i],int);
	}

	hit_auxListInit(&wmain->entrances,int);
	hit_auxListInit(&wmain->subworkers,HitWorkerP);

	// Comm and Group variables
	MPI_Comm ctxComm = HitWorkerGlobalVar.context[w1->context].comm;
	MPI_Comm newComm;

	MPI_Group ctxGroup;
	MPI_Group newGroup;

	// List with the ranks of the farm workers.
	int * ranks;
	hit_malloc(ranks,sizeof(int) * (size_t) num,int*);

	// Obtain the group of the context
	MPI_Comm_group(ctxComm, &ctxGroup);

	hit_auxListAdd(&wmain->subworkers, w1, HitWorker**);
	ranks[0] = hit_auxListElem(&w1->entrances,0);

	va_list list;
	va_start(list,w1);


	for(i=1; i<num; i++){
		HitWorker * worker_p =  va_arg(list,HitWorker*);
		hit_auxListAdd(&wmain->subworkers, worker_p, HitWorker**);
		ranks[i] = hit_auxListElem(&worker_p->entrances,0);

		// @note @javfres Types are not checked and it is impossible to do for
		// not standard ones, so the user may use this function wrong.
		if(	wmain->num_input_channels  != worker_p->num_input_channels ||
			wmain->num_output_channels != worker_p->num_output_channels){
			hit_error_here("Workers in a collaboration must have the same channels");
		}

		worker_p->parent = wmain;

	}


	va_end(list);

	MPI_Group_incl(ctxGroup, num, ranks, &newGroup);
	MPI_Comm_create(ctxComm, newGroup, &newComm);

	for(i=0; i<num; i++){
		HitWorker * subw = hit_auxListElem(&wmain->subworkers,i);
		subw->comm_ws = newComm;
		subw->list.exhausted_ws = 0;
	}



}



/**
 * Creates a Farm of workers cloning a basic worker that is given as
 * a parameter.
 * @param wfarm Pointer to the farm.
 * @param num Number of workers of the farm.
 * @param worker Pointer to a basic worker.
 */
void hit_workerCreateFarm(HitWorker * wfarm, int num, HitWorker * worker){

	if(num<1){
		printf("Farm error num workers == 0,  %d\n",num);
		return;
	}

	// No farm
	if(num == 1){
		worker->parent = NULL;
		*wfarm = *worker;
		return;
	}

	int i;

	*wfarm = HIT_WORKER_NULL;
	worker->parent = wfarm;

	wfarm->type = HIT_WORKER_FARM;

	// Init the aux lists.
	wfarm->num_input_channels = worker->num_input_channels;
	wfarm->num_output_channels = worker->num_output_channels;


	// @note @javfres Although it is technically possible to have farms with
	// several inputs (The buffers support sending different types of tasks)
	// if each worker receives a different number of task for each input buffer
	// the program can not continue.
	if(wfarm->num_input_channels != 1){
		hit_error_here("Number of input channels must be exactly one for farms");
	}


	for(i=0; i<wfarm->num_input_channels; i++){
		hit_auxListInit(&wfarm->predecessors[i],int);
	}

	for(i=0; i<wfarm->num_output_channels; i++){
		hit_auxListInit(&wfarm->successors[i],int);
	}


	hit_auxListInit(&wfarm->entrances,int);
	hit_auxListInit(&wfarm->subworkers,HitWorkerP);

	// Comm and Group variables
	MPI_Comm ctxComm = HitWorkerGlobalVar.context[worker->context].comm;
	MPI_Comm newComm;

	MPI_Group ctxGroup;
	MPI_Group newGroup;

	// List with the ranks of the farm workers.
	int * ranks;
	hit_malloc(ranks,sizeof(int) * (size_t) num,int*);

	// Obtain the group of the context
	MPI_Comm_group(ctxComm, &ctxGroup);

	hit_auxListAdd(&wfarm->subworkers, worker, HitWorker**);
	ranks[0] = hit_auxListElem(&worker->entrances,0);

	for(i=1; i<num; i++){
		HitWorker * wclone = hit_workerBasicClone(worker);
		hit_auxListAdd(&wfarm->subworkers, wclone, HitWorker**);
		ranks[i] = hit_auxListElem(&wclone->entrances,0);
	}


	MPI_Group_incl(ctxGroup, num, ranks, &newGroup);
	MPI_Comm_create(ctxComm, newGroup, &newComm);

	for(i=0; i<num; i++){
		HitWorker * subw = hit_auxListElem(&wfarm->subworkers,i);
		subw->comm_ws = newComm;
		subw->list.exhausted_ws = 0;
	}

}




#ifdef dafaf


/**
 * Frees a worker
 * @param worker Pointer to the worker.
 */
void hit_workerFree(HitWorker * worker){

	hit_auxListFree(&worker->predecessors);
	hit_auxListFree(&worker->entrances);
	hit_auxListFree(&worker->successors);
	hit_auxListFree(&worker->subworkers);

	if(worker->type_input != MPI_DATATYPE_NULL)
		hit_taskBufferFree(&worker->inbuffer);
	if(worker->type_output != MPI_DATATYPE_NULL)
		hit_taskBufferFree(&worker->outbuffer);

}

#endif


/**
 * Worker thread that process the incoming tasks.
 * @param workerP Pointer to the worker.
 */
void *hit_worker_thread_input(void * workerP){

	// Worker pointer
	HitWorker * worker = (HitWorker*) workerP;

	// MPI variables
	MPI_Request requests[HIT_TASK_NCHANNELS_MAX];
	MPI_Status status;

	// Count and init the predecessors
	int active_channels = 0; // worker->num_input_channels;
	int ii;
	for(ii=0; ii<worker->num_input_channels; ii++){
		if( hit_auxListNElems(&worker->predecessors[ii]) > 0){
			active_channels++;
		} else {
			// If a buffer has no predecessor, we mark it as
			// a null_buffer. It will not receive any tasks
			// when calling the Recv function but it always
			// will create null tasks.
			worker->inbuffer[ii].null_buffer = 1;
			worker->inbuffer[ii].empty = 0;
		}
	}


	int waiting_channels = active_channels;
	int active_predecessors[HIT_TASK_NCHANNELS_MAX];
	for(ii=0; ii<worker->num_input_channels; ii++){
		active_predecessors[ii] = hit_auxListNElems(&worker->predecessors[ii]);
		hit_taskBufferIRecv(&worker->inbuffer[ii],&requests[ii]);
	}

	// Reception buffer
	while(active_channels > 0){


		if( waiting_channels == 0 ) hit_error_here("There are pending tasks but no active buffers: Task are not balanced. "
													"Each channel must receive the same number of tasks form their predecessors.");
		// Wait for a msg
		int recv_idx;
		status.MPI_ERROR = MPI_SUCCESS; // MPI_Waitany do not change this field if there is no error.
		int ok = MPI_Waitany(worker->num_input_channels,requests,&recv_idx,&status);
		if(ok != MPI_SUCCESS) hit_error_here("MPI_WaitAny error\n");

		waiting_channels--;


		if(recv_idx < 0 || recv_idx >= worker->num_input_channels) hit_error_here("Impossible input channel");


		// Process the message
		int source;
		hit_taskBufferIRecvEnd(&worker->inbuffer[recv_idx],&source,status);


		while(1){

			// Check that all buffers have tasks
			int available_task = 0;
			int i;
			for(i=0; i<worker->num_input_channels; i++){

				if(!hit_taskBufferEmpty(&worker->inbuffer[i])){
					available_task++;
					//printf("Available tasks at %d (total %d)\n",i,available_task);
					continue;
				}

				if( !hit_taskBufferPendingRecv(&worker->inbuffer[i]) ){

					//printf("Buffer %d is empty and do not have pending tasks\n",i);

					if(active_predecessors[i] == 0){
						//printf("Is this possible?\n");
						continue; // @todo @javfres
					}

					if(hit_taskBufferEOF(&worker->inbuffer[i])){
						active_predecessors[i]--;
						if(active_predecessors[i] == 0){
							active_channels--;
							//printf("End of channel %d\n",i);
							continue; // @todo @javfres
						}
					}
					hit_taskBufferIRecv(&worker->inbuffer[i],&requests[i]);
					waiting_channels++;

					//printf("Start again channel %d\n",i);
					continue; // @todo @javfres
				}
			}

			if(available_task != worker->num_input_channels) break;

			//printf("All buffers have tasks\n");

			// Feed tasks to the list
			HitTask tasks[HIT_TASK_NCHANNELS_MAX];

			for(i=0; i<worker->num_input_channels; i++){
				hit_taskBufferCreateTask(&worker->inbuffer[i],&tasks[i]);
				//printf("Created task elems: %d, size_t: %d, weight: %f, data: %p[%d]\n",
				//		tasks[i].n_elems,tasks[i].baseExtend,tasks[i].weight,tasks[i].data,*((int*)tasks[i].data));

				//HitTaskBuffer * b = &worker->inbuffer[i];
				//printf("Buffer state [eof:%d,empty:%d,request:%d]\n",
				//		hit_taskBufferEOF(b),
				//		hit_taskBufferEmpty(b),
				//		hit_taskBufferRequest(b));
			}

			hit_taskListAddLast(&worker->list,tasks);

		} // while(1)

	}

	// No more incoming tasks.
	hit_taskListSetExhausted_input(&worker->list);

	pthread_exit(NULL);
	return NULL;
}


/**
 * Worker thread that process the outgoing tasks.
 * @param workerP Pointer to the worker.
 */
void *hit_worker_thread_output(void * workerP){

	HitWorker * worker = (HitWorker*) workerP;

	HitWorkerGlobalVar.current_context++;
	worker->func(worker);
	HitWorkerGlobalVar.current_context--;

	int i;
	for(i=0; i<worker->num_output_channels; i++){

		if(hit_auxListNElems(&worker->successors[i]) > 0){

			HitTaskBuffer * bufferP = &worker->outbuffer[i];

			int j;
			for(j=0;j<hit_auxListNElems(&worker->successors[i]);j++){

				hit_taskBufferAddSpecial(bufferP,HIT_BUFFER_EOF);

				HitSuccessor successor = hit_auxListElem(&worker->successors[i],j);
				hit_taskBufferSend(bufferP,successor.proc,successor.flag);
			}

		} // if


	} // for output channels


	pthread_exit(NULL);
	return NULL;
}



void *hit_worker_thread_ws_input(void * workerP){

	// Worker pointer
	HitWorker * worker = (HitWorker*) workerP;

	HitTaskBuffer inbuffer;
	HitTaskBuffer outbuffer;

	// type in and out are the same because in work stealing we only send or recv the
	// task of the internal worker list.
	// @note @javfres The worker can only share one type of tasks.
	hit_taskBufferCreate(&inbuffer,worker->config.buffer_in,worker->comm_ws,0,1,worker->type_input[0]);
	hit_taskBufferCreate(&outbuffer,worker->config.buffer_in,worker->comm_ws,0,1,worker->type_input[0]);

	int num_workers;
	MPI_Comm_size(worker->comm_ws,&num_workers);

	int remain_ws = num_workers-1;
	int tries_left = num_workers+1; // @todo

	while(remain_ws > 0){

		//printf_state(worker,2,"ws_input Waiting for something\n");

		int source = MPI_ANY_SOURCE;
		hit_taskBufferRecv(&inbuffer,&source);

		if(hit_taskBufferRequest(&inbuffer)){

			//printf("Received request from %d\n",source);

			int dest = source;

			HitTask task;

			// If we have enough task we send them.
			if(hit_taskListWeight(&worker->list) > worker->config.threshold_ws_send){

				while(outbuffer.num_used < outbuffer.num_elem){

					hit_taskListRemoveLast(&worker->list,&task);
					if(hit_taskIsNull(task)){
						break;
					}

					//printf("Worker: ADD task from WS\n");
					hit_taskBufferAdd(&outbuffer, &task);

					if(hit_taskBufferWeight(&outbuffer) > worker->config.threshold_ws_send) break;
				}

			// If we don't have them: Nothing.
			} else {

				//hit_taskBufferAddSpecial(&outbuffer,HIT_BUFFER_EOF);
			}

			hit_taskBufferSend(&outbuffer, dest, 0);
			continue;
		}


		if(hit_taskBufferEOF(&inbuffer)){

			//printf("Received end from %d\n",source);

			remain_ws--;

			if(remain_ws == 0){
				hit_workerFlagSet(worker->finished_ws,&worker->mutex_ws,1);
				hit_taskListWakeUp(&worker->list);
				break;
			}
			hit_taskListWakeUp(&worker->list);

			continue;
		}


		// ----------- Recv ----------- //

		if(hit_taskBufferEmpty(&inbuffer)){
			tries_left--;
			if(tries_left == 0){
				hit_workerFlagSet(worker->finished_ws,&worker->mutex_ws,1);
			}
		} else {
			tries_left = num_workers+1;

			// Process each input task
			while(!hit_taskBufferEmpty(&inbuffer)){

				HitTask task;
				hit_taskBufferCreateTask(&inbuffer,&task);

				// Add the task to the list
				hit_taskListAddLast(&worker->list, &task);
			}

			//printf("Received %d tasks from %d\n",ntask,source);
		}


		hit_workerFlagSet(worker->pending_request,&worker->mutex_ws,0);
		hit_taskListWakeUp(&worker->list);

		// ----------- Recv ----------- //


	} // while(remain_ws > 0)

	hit_taskListSetExhausted_ws(&worker->list);

	hit_taskBufferFree(&inbuffer);
	hit_taskBufferFree(&outbuffer);

	//printf("hit_worker_thread_ws_input end\n");

	pthread_exit(NULL);
	return NULL;
}



void *hit_worker_thread_ws_output(void * workerP){

	// Worker pointer
	HitWorker * worker = (HitWorker*) workerP;

	int rank_ws;
	int nprocs_ws;

	MPI_Comm_rank(worker->comm_ws, &rank_ws);
	MPI_Comm_size(worker->comm_ws, &nprocs_ws);

	HitTaskBuffer outbuffer;
	hit_taskBufferCreate(&outbuffer,1,worker->comm_ws,0,0); // A no-type buffer

	while( hit_taskListWaitRemove(&worker->list) ){

		// Check if the WS is over (hit_worker_thread_ws_input decide it).
		int flag;
		hit_workerFlagGet(worker->finished_ws,&worker->mutex_ws,flag);

		// If it is finished we have to send the all stop msg.
		if(flag){
			break;
		}

		// Check if we need tasks
		double weight = hit_taskListWeight(&worker->list);
		if(weight > worker->config.threshold_ws_request){
			continue;
		}

		// We need tasks
		//int me = hit_auxListElem(&worker->entrances,0);

		int flag2;
		// Check if there is a previous pending request
		hit_workerFlagGet(worker->pending_request,&worker->mutex_ws,flag2);

		// If there is a previous request we do nothing.
		if(flag2){
			continue;
		}

		// Send the request
		hit_workerFlagSet(worker->pending_request,&worker->mutex_ws,1);

		// @todo implement other methods to select dest
		worker->last_ws = (worker->last_ws + 1) % nprocs_ws;
		if(worker->last_ws == rank_ws){
			worker->last_ws = (worker->last_ws + 1) % nprocs_ws;
		}

		hit_taskBufferAddSpecial(&outbuffer,HIT_BUFFER_REQUEST);
		hit_taskBufferSend(&outbuffer,worker->last_ws,0);
		//printf("Sent request to %d\n",worker->last_ws);

	}

	// Send the all stop msg
	int i;

	for(i=0;i<nprocs_ws;i++){
		if(rank_ws == i) continue;

		hit_taskBufferAddSpecial(&outbuffer,HIT_BUFFER_EOF);
		hit_taskBufferSend(&outbuffer,i,0);


	}
	//printf("hit_worker_thread_ws_output end\n");
	hit_taskBufferFree(&outbuffer);

	pthread_exit(NULL);
	return NULL;
}



/**
 * Gets a new task from the list, it could come
 * from some predecessor or from the WS pool.
 * @param worker The pointer to the worker.
 * @param taskP Pointer to the task.
 * @return True of false depending if there are more task.
 */
int hit_workerGetTask(HitWorker * worker, void * taskP){

	//printf("GetTask number:%d, exahusted: %d %d\n",worker->list.size,worker->list.exhausted_input,worker->list.exhausted_ws);


	return hit_taskListRemoveFirst(&worker->list, taskP);

}




/**
 * Send the task to any successor.
 * @param worker The pointer to the worker.
 * @param taskP Pointer to the task or task array.
 * @param output Output channel
 */
void hit_workerSendTask(HitWorker * worker, void * taskP, int output){

	HitTaskBuffer * bufferP = &worker->outbuffer[output];

	HitTask * t = (HitTask*) taskP;
	if((bufferP->num_used+t->n_elems) > bufferP->num_elem){
		// @todo @javfres Change order, this is duplicate
		worker->last_successor[output] = (worker->last_successor[output] + 1) % hit_auxListNElems(&worker->successors[output]);
		HitSuccessor dest = hit_auxListElem(&worker->successors[output],worker->last_successor[output]);
		hit_taskBufferSend(bufferP,dest.proc,dest.flag);
	}


	hit_taskBufferAdd(bufferP, taskP);

	double weight = hit_taskBufferWeight(bufferP);


	if(weight > worker->config.threshold_send || bufferP->num_used >= bufferP->num_elem ){

		// @todo implement other methods to select successors
		worker->last_successor[output] = (worker->last_successor[output] + 1) % hit_auxListNElems(&worker->successors[output]);
		//worker->last_sucessor = 0;
		HitSuccessor dest = hit_auxListElem(&worker->successors[output],worker->last_successor[output]);

		//printf(">>>>>>>>> Sending pack of %d to %d\n",bufferP->num_used,dest);

		hit_taskBufferSend(bufferP,dest.proc,dest.flag);

	}
}

/**
 * Adds a new task to the local list.
 * @param worker The pointer to the worker.
 * @param taskP Pointer to the task.
 */
void hit_workerAddTask(HitWorker * worker, void * taskP){

	hit_taskListAddFirst(&worker->list, taskP);

}






























void hit_workerExecuteBasic(HitWorker * worker){


	// 1. Check if the worker should be executed
	int i;
	int should_execute = 0;
	for(i=0; i<hit_auxListNElems(&worker->entrances); i++){
		if(hit_workerCurrentContext.rank == hit_auxListElem(&worker->entrances,i)){
			should_execute = 1;
			break;
		}
	}
	if(!should_execute) return;


	pthread_t threads[HIT_WORKER_THREAD_NUM];


	pthread_create(&threads[HIT_WORKER_THREAD_INPUT], NULL, hit_worker_thread_input, (void*) worker);
	pthread_create(&threads[HIT_WORKER_THREAD_OUTPUT], NULL, hit_worker_thread_output, (void*) worker);


	if(worker->parent != NULL && worker->parent->enable_ws &&  worker->comm_ws != MPI_COMM_NULL){
		pthread_create(&threads[HIT_WORKER_THREAD_WS_INPUT], NULL, hit_worker_thread_ws_input, (void*) worker);
		pthread_create(&threads[HIT_WORKER_THREAD_WS_OUTPUT], NULL, hit_worker_thread_ws_output, (void*) worker);
	} else {
		hit_taskListSetExhausted_ws(&worker->list);

	}

	pthread_join(threads[HIT_WORKER_THREAD_INPUT],NULL);
	pthread_join(threads[HIT_WORKER_THREAD_OUTPUT],NULL);

	if(worker->parent != NULL && worker->parent->enable_ws && worker->comm_ws != MPI_COMM_NULL){
		pthread_join(threads[HIT_WORKER_THREAD_WS_INPUT],NULL);
		pthread_join(threads[HIT_WORKER_THREAD_WS_OUTPUT],NULL);
	}

}

/**
 * Executes the worker.
 * @param worker
 */
void hit_workerExecute(HitWorker * worker){

	// @todo
	//hit_workerResetStatus(worker);

	int i;

	switch(worker->type){

		case HIT_WORKER_BASIC:
			hit_workerExecuteBasic(worker);
			break;

		case HIT_WORKER_PIPE:
		case HIT_WORKER_FARM:
			for(i=0; i<hit_auxListNElems(&worker->subworkers); i++){
				hit_workerExecute(hit_auxListElem(&worker->subworkers,i));
			}

			break;

		default:
			break;

	}
}



