/*
 * Debug functions for test
 *
 * @file hit_test.c
 * @version 1.0
 * @version 1.1 13-11-2011 Added the dump for bitmap shape.
 * @date 13-11-2011
 * @author Javier Fresno Bausela
 */

/*
 * This software is provided to enhance knowledge and encourage progress
 * in the scientific community and are to be used only for research and
 * educational purposes. Any reproduction or use for commercial purpose
 * is prohibited without the prior express written permission.
 * 
 * This software is provided "as is" and without any express or implied
 * warranties, including, without limitation, the implied warranties of
 * merchantability and fitness for a particular purpose.
 * 
 * Copyright (c) 2007-2011, Javier Fresno Bausela.
 * All Rights Reserved.
 */



#include "hit_test.h"


// Dump a integer array
void dumpArray(int * array,int dims, const char * format, ...){

	char name[1024];

	va_list args;
	va_start( args, format );
	vsprintf( name, format, args );
	va_end( args );


	if(dims == 0){
		printf("%s [X]\n",name);
		return;
	}

	int i;
	printf("%s [%d",name,array[0]);
	for(i=1;i<dims;i++) printf(",%d",array[i]);
	printf("]");

}

// Dump a HitTopology
void dumpTopoCommon( HitTopology topo, const char * name ) {

	printf("\n Topology: %s\n",name);
	printf("---------------------------\n");
	printf("NumDims: %2d, ", topo.numDims );

	dumpArray(topo.card,topo.numDims,"Card");
	printf("\n");

}

void dumpTopoSelf( HitTopology topo) {

	printf("Lineal Rank: %2d", topo.linearRank );

	dumpArray(topo.self.rank,topo.numDims,", Self");


	printf(", Active: %d\n", topo.active );
}


void dumpListLayout(HitLayout lay, const char * name, double * weights){

	int group, i;

	printf("\n ListLayouty: %s\n",name);
	printf("---------------------------\n");
	printf("Processors: %d, Elements: %d -> Groups: %d\n",
			lay.topo.pTopology->numProcs,
			lay.info.layoutList.numElementsTotal,
			lay.info.layoutList.numGroups);
	printf("---------------------------\n");

	for(group=0;group<lay.info.layoutList.numGroups;group++){

		int leader = lay.info.layoutList.groups[group].leader;
		int nprocs = lay.info.layoutList.groups[group].numProcs;

		printf("Group %d\n   - Num Processors: %d",group,nprocs);

		printf("\n   - Processors: ");
		for(i=0;i<nprocs;i++){
			printf("%d",leader+i);
			if(i!=nprocs-1) printf(", ");
		}

		double weight = 0;
		printf("\n   - Elements: ");
		for(i=0;i<lay.info.layoutList.numElementsTotal;i++){
			if(lay.info.layoutList.assignedGroups[i] == group){
				printf("%d",i);
				printf(", ");
				if(weights != NULL) weight += weights[i];
			}

		}
		printf(" sum weights: %.1f\n",weight);


	}


}


void dumpShape(HitShape sh){

	int i;
	printf("[%d:%d:%d",hit_shapeSig(sh,0).begin,hit_shapeSig(sh,0).end,hit_shapeSig(sh,0).stride);
	for(i=1;i<hit_shapeDims(sh);i++) printf(",%d:%d:%d",hit_shapeSig(sh,i).begin,hit_shapeSig(sh,i).end,hit_shapeSig(sh,i).stride);
	printf("] \t cards: [%d",hit_sigCard(hit_shapeSig(sh,0)));
	for(i=1;i<hit_shapeDims(sh);i++) printf(",%d",hit_sigCard(hit_shapeSig(sh,i)));
	printf("]\n");
}


void dumpSignatureLayoutCommon( HitLayout lay , const char * name ) {

	printf("\n LayoutSig: %s\n",name);
	printf("---------------------------\n");
	dumpArray(lay.numActives,hit_shapeDims( hit_layShape( lay ) ),"Num Actives: ");
	printf(", Original shape: "); dumpShape(lay.origShape);
}


void dumpSignatureLayoutSelf( HitLayout lay ) {

	printf(", Active: %2d, ", lay.active );
	dumpArray(lay.topo.self.rank,lay.topo.numDims,"Topo ranks");
	printf(", Group: %2d, Leader: %2d",lay.group,lay.leader);
	printf(", My shape: "); dumpShape(hit_layShape(lay));

}




// All prints
void printfAllSec( const char* format, ... ) {

	int i = 0;

	for(i=0;i<HIT_TOPOLOGY_INFO->numProcs;i++){

		if(HIT_TOPOLOGY_INFO->selfRank == i){

			fprintf( stdout, "%2d: ",i);
			va_list args;
			va_start( args, format );
			vfprintf( stdout, format, args );
			va_end( args );

		}
		MPI_Barrier(HIT_TOPOLOGY_INFO->comm);
		fflush(stdout);
		struct timespec ts = {0,10000};
		nanosleep(&ts,&ts);
	}
 }

// All dumps a integer array
void dumpArrayAllSec(int * array,int dims, const char * format, ...){

	char name[1024];

	va_list args;
	va_start( args, format );
	vsprintf( name, format, args );
	va_end( args );

	char string[1024];
	char * s = string;

	if(dims == 0){
		s += sprintf(s,"%s [X]\n",name);
		return;
	}

	int i;
	s += sprintf(s,"%s [%d",name,array[0]);
	for(i=1;i<dims;i++) s += sprintf(s,",%d",array[i]);
	s += sprintf(s,"]\n");

	printfAllSec("%s",string);

}


void dumpNameList(HitNameList list, int dim){

	printf(" Names[%d] ",dim);


	if(list.flagNames == HIT_SHAPE_NAMES_ARRAY) printf("[array] :");
	if(list.flagNames == HIT_SHAPE_NAMES_ORDERED ) printf("[ordered] :");
	if(list.flagNames == HIT_SHAPE_NAMES_NOARRAY  ) printf("[noarray] :");

	int i;
	for(i=0;i<list.nNames;i++)
		printf(" %d ", list.names[i]);

	if(list.invNames != NULL){
		printf(" InvNames: ");
		for(i=0;i<list.nInvNames;i++)
			printf(" %d ", list.invNames[i]);
	} else {
		printf(" InvNames: NULL");
	}
	printf("\n");

}


void dumpCShape(HitShape shape){

	HitCShape s = shape.info.csr;

	int i,j;

	printf("Cards: %dx%d.   ",s.cards[0],s.cards[1]);

	for(i=0;i<s.cards[0];i++){

		int begin = s.xadj[i];
		int end = s.xadj[i+1];

		for(j=begin;j<end;j++){
			printf(" %d ", s.adjncy[j]);
		}
		printf("|");
	}
	printf("\n");
	dumpNameList(s.names[0],0);
	dumpNameList(s.names[1],1);

}


hit_tileNewType(double);

void print_matrix(HitTile_double matrix){

	HitShape shape = hit_tileShape(matrix);

	int n = hit_cShapeNameList(shape,0).nNames;
	int m = hit_cShapeNameList(shape,1).nNames;

	printf("Matrix [%d,%d]\n",n,m);

	int i,j;
	printf("     ");
	for(j=0;j<m;j++){
		int jg = hit_cShapeNameList(shape,1).names[j];
		printf("%4d|",jg);
	}
	printf("\n");



	for(i=0; i<n; i++){

		int ig = hit_cShapeNameList(shape,0).names[i];

		printf("%3d|",ig);

		for(j=0; j<m; j++){

			int jg = hit_cShapeNameList(shape,1).names[j];

			//printf("Print local [%d,%d], global [%d,%d]\n",i,j,ig,jg);

			if(hit_cShapeElemExists(shape,ig,jg)){
				printf("%5.1f",hit_mcTileGraphElemAt(matrix,ig,jg));
			} else {
				printf("     ");
			}
		}
		printf("\n");
	}
	printf("\n");
}


void dumpBShape(HitShape shape){

	HitBShape s = shape.info.bitmap;

	int nz = hit_bShapeNedges(shape);

	printf("NZ: %d\n",nz);

	int i,j;
	for(i=0;i<hit_cShapeCard(shape,0);i++){
		for(j=0;j<hit_cShapeCard(shape,1);j++){
			printf("%d",hit_bShapeGet(shape,i,j)!=0);
		}
		printf("\n");
	}

	dumpNameList(s.names[0],0);
	dumpNameList(s.names[1],1);

}
